# Copyright (C) 2001, 2003-2004, 2006, 2008-2009 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import string
import sys
import types

from urlparse import urljoin, urlparse

from UserString import UserString

from lxml.etree import XMLSyntaxError

import depot
import depotdb

from classutil import AbstractMeta, AbstractProperty, abstract
from misclib import VERBOSE_DEPOT
from xmlutil import XMLError

# constants
READ_CHUNK = 20480		# number of bytes to read at a time

# exceptions
class DepotError (Exception):
  """Base class for all exceptions raised by this module"""

  def __init__ (self, msg = None):
    self.msg = msg

  def __str__ (self):
    return self.msg

class DepotAccessDeniedError (DepotError):
  """access denied (authorization insufficient)"""

class DepotAccessError (DepotError):
  """unable to access depot"""

class DepotFileNotFoundError (DepotError):
  """file not found in depot"""

  def __init__ (self, file = None):
    self.file = file

class DepotGPGError (DepotError):
  """GPG signature for depot-db.xml failed"""

class DepotNoLocalDepot (DepotError):
  """no local depot specified with --local-depot"""

class DepotNoValidDepots (DepotError):
  """no valid depots"""

class DepotNotFoundError (DepotError):
  """depot not found"""

class DepotPackageVerNotFoundError (DepotError):
  """requested version of this package not found in depot"""

class DepotPermissionDeniedError (DepotError):
  """permission denied"""

class DepotProtocolUnsupportedError (DepotError):
  """protocol unsupported"""

class DepotProxyAuthenticationRequiredError (DepotError):
  """proxy authentication required"""

class DepotSftpAuthenticationError (DepotError):
  """authentication failure"""

  def __init__ (self, host):
    self.host = host

class DepotSftpKnownHostsFileNotFoundError (DepotError):
  """ssh_known_hosts file not found"""

class DepotSftpKnownHostsFileNotSpecifiedError (DepotError):
  """ssh_known_hosts file not specified"""

class DepotSftpKnownHostsPathNotFoundError (DepotError):
  """path specified for ssh_known_hosts not found"""

class DepotSftpInvalidKnownHostError (DepotError):
  """host unknown in ssh_known_hosts"""

class DepotSftpPrivateKeyEncryptedError (DepotError):
  """private key encrypted with passphrase"""

class DepotSftpPrivateKeyFileNotFoundError (DepotError):
  """private key file not found"""

class DepotSftpPrivateKeyFileNotSpecifiedError (DepotError):
  """private key file not specified"""

class DepotSSLError (DepotError):
  """SSL error"""

  def __init__ (self, msg):
    self.error = msg


# proxy authentication variables
class ProxyAuth:
  def __init__ (self):
    self.host = ''
    self.login = ''
    self.passwd = ''
    self.base64 = ''			# base64-encoded login/password

class ProtocolAuth:
  def __init__ (self):
    self.login = ''
    self.passwd = ''
    self.base64 = ''			# base64-encoded login/password

    self.sftp_known_hosts_path = None	# path to ssh_known_hosts
    self.sftp_private_key_path = None	# path to ssh private key


# version number and revision class
class DepotBase (UserString):
  """Version number and revision class.

  New style class method lookup resolution only finds the "stringy"
  methods if we inherit a "stringy" superclass.  Unfortunately, inheriting
  the builtin "str" makes our type into "str" - foiling the type switched
  code looking for "InstanceType"."""
  
  def __init__ (self, dep):
    if isinstance (dep, basestring):
      url = dep.url
    else:
      url = dep

    UserString.__init__ (self, url)
    
    self.archive = 0			# if depot is a zip archive

    self.url = url			# parent url
    self.pkg_url = url			# where to find packages
    self.type = urlparse (url).scheme   # URL scheme ('http', ...)
    self.local_path = None

    self.proxy = ProxyAuth ()
    self.auth = ProtocolAuth ()

    self.dtd_dirs = None        	# DTD search directories
    self.dist = None			# distribution name
    self.packages = []			# packages contained in depot

  # output depot path
  def __str__ (self):
    return self.url

  def _db_exists (self, base_name, suffix):
    if not os.path.exists (os.path.join (self.local_path,
                                         base_name + suffix)):
      return False
    return True

  def fetch (self, pkg):
    pass

  def open (self, path, mode):
    if self.local_path and os.path.exists (self.local_path + '/' + path):
      return open (self.local_path + '/' + path, 'r')
    else:
      return (self.urlopen (urljoin (self.pkg_url, path)))

  def set_auth (self):
    pass

  def set_proxy (self):
    pass

# list of repositories
class DepotsBase (object):
  # Raise an exception at class instantiation time if subclasses
  # contain any (unoverridden) AbstractProperty's or @abstract methods:
  __metaclass__ = AbstractMeta

  # Set these appropriately in subclass
  XML_FILE = AbstractProperty ('XML_FILE')
  DB_SUFFIX = AbstractProperty ('DB_SUFFIX')

  def __init__ (self, add_package, assign_depot_vars,
                local_depot, msg_callback, package_list,
                package_names, dist, systype_path):
    self.add_package = add_package
    self.assign_depot_vars = assign_depot_vars
    self.local_depot = local_depot
    self.msg_callback = msg_callback
    self.package_list = package_list
    self.package_names = package_names
    self.dist = dist
    self.systype_path = systype_path

    self.depots = []
    
  @abstract
  def archive_init (self, path):
    """Initialise an archive object in the concrete class

    @param path: File path to the archive"""
    pass
  
  @abstract
  def read_db_xml (self, archive, dep, basename):
    """Read an XML file

    @param archive: As returned by self.archive_init method
    @param dep: A depot object
    @param basename: The name of the file to find and read

    @returns A parsed XML DB entry"""
    pass

  def set (self, depot_names):
    valid_depots = []			# depots we can use
    search_depots = 0

    search_package_names = []		# packages to search for
    for package_name in self.package_names:
      if os.path.isfile (package_name['name']):
        if os.path.isabs (package_name['name']):
          path = os.path.normpath (package_name['name'])
        else:
          path = os.path.normpath (os.path.join (os.getcwd (),
                                   package_name['name']))

        dep = init (None, self.local_depot, os.path.dirname (path))

        apply (self.assign_depot_vars, [dep])

        apply (self.msg_callback,
               ['begin', 'reading ' +
                urljoin (dep.pkg_url, os.path.basename (path)),
                0, VERBOSE_DEPOT])

        try:
          archive = self.archive_init (path)
          if archive:
            # set depot path to the archive file
            dep.archive = archive
            dep.local_path = path
            dep.url = 'file://' + path

            if archive.exists (self.XML_FILE):
              db_entry = self.read_db_xml (archive, dep, self.XML_FILE)
            elif archive.exists (os.path.basename (path)):
              db_entry = self.read_db_xml (archive, dep,
                                            os.path.basename (path))
            else:
              apply (self.msg_callback, ['end', 'failed', 0, VERBOSE_DEPOT])
              apply (self.msg_callback, ['error', 'an input archive \
was specified but did not contain either a "%s" or "%s".  Archives \
must contain one of these files. skipping.' % (self.XML_FILE,
                                               os.path.basename (path))])
              continue
          else:
            db_entry = self.read_db_xml (dep, dep, path)

          if not db_entry:
            apply (self.msg_callback, ['end', 'failed', 0, VERBOSE_DEPOT])
            apply (self.msg_callback, ['error', 'unable to parse the \
XML file. please check "%s". it does not appear to follow \
%s(4). skipping.' % (path, self.XML_FILE)])
            continue

        except XMLError, msg:
          apply (self.msg_callback, ['end', 'failed', 0, VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg)])
          continue

        self.depots.append (dep)
        valid_depots.append (dep)

        apply (self.msg_callback, ['end', 'done', 0, VERBOSE_DEPOT])

        apply (self.add_package, [db_entry.prog_name + '-' +
                                  db_entry.prog_ver, archive])
      else:
        apply (self.add_package, [package_name['name']])
        search_package_names.append (package_name['name'])
        search_depots = 1		# search through list of
					# depots for package

    if search_depots:
      for depot_name in depot_names:
        try:
          dep = init (self.dist, self.local_depot, depot_name)
        except DepotProtocolUnsupportedError, msg:
          apply (self.msg_callback, ['end', 'failed', 0, VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg) + ' removing ' +
                                     depot_name + ' from list of depots.'])
          continue

        if dep.is_remote () and not self.local_depot:
          raise DepotNoLocalDepot ('remote depots were specified but a \
--local-depot option was not given. please rerun specifying a local \
local path where packages from remote repositories can be copied to.')

        try:
          apply (self.assign_depot_vars, [dep])
        except (DepotSftpKnownHostsFileNotFoundError,
                DepotSftpKnownHostsFileNotSpecifiedError,
                DepotSftpPrivateKeyFileNotFoundError,
                DepotSftpPrivateKeyFileNotSpecifiedError), msg:
          apply (self.msg_callback, ['end', 'failed', 0, VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg)])
          continue

        self.depots.append (dep)

      apply (self.msg_callback, ['end', 'reading depot databases', 0,
                                 VERBOSE_DEPOT])

      for dep in self.depots:
        depot_db_xml = 0		# if depot-db.xml exists
        db_xml = 0			# if xx-db.xml exists

        # if parent depot contains more depots, add to list
        apply (self.msg_callback,
               ['begin', urljoin (dep.pkg_url, 'depot-db.xml'),
                2, VERBOSE_DEPOT])

        try:
          if dep.exists ('depot-db.xml'):
            depot_db_xml = 1

            dep.fetch ('depot-db.xml')	# make copy in local depot

            sub_depots = depotdb.read_depotdb_xml ('depot-db.xml', dep)

            apply (self.msg_callback, ['end', 'done', 2, VERBOSE_DEPOT])

            # create new depots
            for sub_depot in sub_depots:
              # only interested in directories with source
              if sub_depot.category:
                if sub_depot.category != 'src':
                  continue
                if not sub_depot.auto_include:
                  continue

              # make sure distribution matches one we are interested in
              if sub_depot.dist_ver:
                if sub_depot.dist_ver not in dep.dist_ver:
                  continue

              # if package name specified in depot add to list of
              # depots only if package name in list of packages
              # we're building
              if sub_depot.package:
                for package in self.package_list:
                  if sub_depot.package == package.name and \
                  package.name in search_package_names:
                    break
                else:
                  continue

              apply (self.msg_callback,
                     ['end', 'adding ' +
                      urljoin (dep.pkg_url, sub_depot.name),
                      4, VERBOSE_DEPOT])

              if sub_depot.name:
                new_dep = init (self.dist, self.local_depot, dep,
                                sub_depot.name)
                apply (self.assign_depot_vars, [new_dep])
                if sub_depot.package and \
                not sub_depot.package in new_dep.packages:
                  new_dep.packages.append (sub_depot.package)
                self.depots.append (new_dep)
              else:
                dep.packages.append (sub_depot.package)
          else:
            apply (self.msg_callback, ['end', 'failed (not found)', 2,
                                       VERBOSE_DEPOT])
        except DepotAccessDeniedError:
          apply (self.msg_callback, ['end', 'failed (access denied)', 0,
                                     VERBOSE_DEPOT])
          continue
        except DepotPermissionDeniedError:
          apply (self.msg_callback, ['end', 'failed (permission denied)',
                                     2, VERBOSE_DEPOT])
          continue
        except DepotProxyAuthenticationRequiredError:
          apply (self.msg_callback, ['error', 'access denied to proxy \
server (' + dep.proxy.host + '). either the proxy login id or \
password was incorrect or not specified. removing ' + dep.url +
                                     ' from list of depots.'])
          continue
        except DepotSftpAuthenticationError, msg:
          apply (self.msg_callback, ['error', 'unable to authenticate to ' +
                                     msg.host + ' using the private key \
specified. please either specify the correct private key or ensure \
the remote host has the contents of "' +
                                     dep.auth.sftp_private_key_path +
                                     '.pub" in ~' + dep.auth.login +
                                     '/.ssh/authorized_keys. removing ' +
                                     dep.url + ' from list of depots.'])
          continue
        except (DepotSftpInvalidKnownHostError,
                DepotSftpPrivateKeyEncryptedError,
                XMLError), msg:
          apply (self.msg_callback, ['error', str (msg) + ' removing ' +
                                     dep.url + ' from list of depots.'])
          continue

        # depot has suffixed file. Either:
        #   the package name + self.DB_SUFFIX
        #   or just self._xml_name
        db_filename = ''
        if dep.exists (self.XML_FILE):
          db_xml = 1
          db_filename = self.XML_FILE
        else:
          for package in dep.packages:
            if dep.exists (package + self.DB_SUFFIX):
              db_xml = 1
              db_filename = package + self.DB_SUFFIX
              break
          else:
            if depot_db_xml:
              valid_depots.append (dep)
            continue

        valid_depots.append (dep)

        apply (self.msg_callback,
               ['begin', 'reading ' +
                urljoin (dep.pkg_url, db_filename), 2, VERBOSE_DEPOT])

        dep.fetch (db_filename)		# copy to local depot

        try:
          self.read_db_xml (dep, dep, db_filename)
          apply (self.msg_callback, ['end', 'done', 2, VERBOSE_DEPOT])

        except XMLError, msg:
          # Finish the callback output, and reraise the exception
          apply (self.msg_callback, ['end', 'failed', 2, VERBOSE_DEPOT])
          raise XMLError (str (msg))

      self.depots = valid_depots

    # separate depot output and package installation
    apply (self.msg_callback, ['end', '', 0, VERBOSE_DEPOT])

    # must have at least one valid depot
    if not self.depots:
      if depot_names:
        raise DepotNoValidDepots ('while some depot paths were given (' + \
        string.join (depot_names, ',') + '), none specified a path we \
could use. The repository must contain either a "depot-db.xml" file \
specifying further subdirectories to walk, a "%s" file specifying a \
package definition, or "[package]%s" files specifying package \
definitions for the input package names.' % (self.XML_FILE, self.DB_SUFFIX))
      else:
        raise DepotNoValidDepots ('no depots given. please specify \
at least one depot with the --depot option or the path of a \
%s(4) file or a %s archive file.' % (self.XML_FILE, self.DB_SUFFIX))

  def append (self, dep):
    self.depots.append (dep)

  def __getitem__ (self, key):
    return self.depots[key]

  def __iter__ (self):
    return self.depots.__iter__ ()

  def __nonzero__ (self):
    return len (self.depots)

def init (dist, local_depot, dep, sub_depot = None):
  # dep is either a URL or an instance of the Depot class. if the
  # latter, dep is the parent URL of sub_depot.
  if isinstance (dep, basestring):
    url = dep.url
  else:
    url = dep

  if url.startswith ('ftp://'):
    import depot.ftp
    return (depot.ftp.ftp (dist, local_depot, dep, sub_depot))
  elif url.startswith ('sftp://'):
    try:
      import depot.sftp
      return (depot.sftp.sftp (dist, local_depot, dep, sub_depot))
    except ImportError:
      raise DepotProtocolUnsupportedError ('sftp protocol not \
supported on this platform. one or more of the dependencies \
required to support it are unavailable.')
  elif url.startswith ('http://'):
    import depot.http
    return (depot.http.http (dist, local_depot, dep, sub_depot))
  elif url.startswith ('https://'):
    import depot.https
    return (depot.https.https (dist, local_depot, dep, sub_depot))
  elif url.startswith ('file://'):
    import depot.file
    return (depot.file.file (dist, dep, sub_depot))
  else:
    import depot.file
    return (depot.file.file (dist, 'file://' + dep, sub_depot))
