# Configuration file parser
#
# Copyright (C) 2002-2003, 2005-2006 The Written Word, Inc.
# Copyright (C) 2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import shlex
import string

from fnmatch import fnmatch
from misclib import systype_match

# exceptions
class ConfigError (Exception):
  """base class for all exceptions raised by this module"""

  def __init__ (self, msg, lineno):
    self.msg = msg
    self.lineno = lineno

  def __str__ (self):
    return self.msg

class ConfigFileNotFoundError (Exception):
  """configuration file not found"""

class ConfigSectionParseError (Exception):
  """error parsing a section"""

  def __init__ (self, section, msg = None):
    self.section = section
    self.msg = msg

# expected types for variables
_var_types = {
  'check-sig' : 'boolean',
  'depots' : 'array',
  'destdir' : 'string',
  'dist-ver' : 'array',
  'dtd-dirs' : 'array',
  'egd-socket' : 'string',
  'gpg-keyring-path' : 'string',
  'global-pkg-db' : 'string',
  'install-base' : 'string',
  'local-depot' : 'string',
  'login' : 'string',
  'npkg-db' : 'string',
  'npkgadd-path' : 'string',
  'npkginfo-path' : 'string',
  'npkgrm-path' : 'string',
  'password' : 'string',
  'pkg-db' : 'string',
  'pkg-type' : 'string',
  'pkgmgr-args' : 'array',
  'pkgmgr-args-inst' : 'array',
  'pkgname-prefix' : 'string',
  'proxy-host' : 'string',
  'proxy-login' : 'string',
  'proxy-password' : 'string',
  'root' : 'string',
  'rpm-arch' : 'string',
  'rpm-db' : 'string',
  'rpm-path' : 'string',
  'sftp-known-hosts-path' : 'string',
  'sftp-private-key-path' : 'string',
  'subpkgs' : 'array',
  'tmpbasedir' : 'string',
  'update-type' : 'array'
}

# valid variables in depot section
SECTION_DEPOT_VARS = [
  'dist-ver', 'login', 'password', 'proxy-host', 'proxy-login',
  'proxy-password', 'subpkgs', 'update-type'
]

# valid variables in dist-opts section
SECTION_DIST_OPTS_VARS = [
  'check-sig', 'depots', 'destdir', 'dist-ver', 'install-base',
  'local-depot', 'login', 'npkg-db', 'password', 'pkg-db',
  'pkg-type', 'pkgmgr-args', 'pkgmgr-args-inst', 'pkgname-prefix',
  'root', 'rpm-arch', 'rpm-db', 'rpm-path', 'subpkgs', 'update-type'
]

# valid variables in systype section
SECTION_SYSTYPE_VARS = SECTION_DIST_OPTS_VARS

def parse (path, systype):
  conf = {}

  if not os.path.exists (path):
    raise ConfigFileNotFoundError

  stream = shlex.shlex (open (path), path)
  while 1:
    var = None
    val = None

    # get variable name or section identifier
    stream.wordchars = string.letters + string.digits + '-_'
    token = stream.get_token ()
    if not token:
      break
    if len (token) == 1 and not token in stream.wordchars:
      raise ConfigError ('invalid character in variable', stream.lineno)

    if token == 'depot':
      if not conf.has_key (token):
        conf[token] = {}
      parse_section_depot (stream, conf)
    elif token == 'dist-opts':
      if not conf.has_key (token):
        conf[token] = {}
      parse_section_dist_opts (stream, conf, systype)
    elif token == 'systype':
      parse_section_systype (stream, conf, systype)
    else:
      if not _var_types.has_key (token):
        raise ConfigError ('invalid variable name %s' % token, stream.lineno)
      conf[token] = parse_assignment (token, stream)

  stream.instream.close ()

  return conf

# get remainder of variable assignment
def parse_assignment (var, stream):
  stream.wordchars = '='
  token = stream.get_token ()
  if not token:
    raise ConfigError ('no assignment for variable', stream.lineno)
  if token != '=':
    raise ConfigError ('= must follow variable assignment', stream.lineno)

  # get variable value
  stream.wordchars = string.letters + string.digits + string.punctuation
  token = stream.get_token ()
  if not token:
    raise ConfigError ('no assignment for variable', stream.lineno)

  if _var_types[var] == 'array':
    # array values must begin with '['
    if token != '[':
      raise ConfigError ('"%s" accepts an array value. therefore, \
its value must be enclosed by brackets ("[]").' % var, stream.lineno)

    # read list
    val = []
    quoted_str = ''				# value between quotes
    while 1:
      token = stream.get_token ()
      if not token:
        raise ConfigError ('unterminated list assignment', stream.lineno)
      if token == ']':
        if quoted_str:
          raise ConfigError ('unterminated quoted string', stream.lineno)
        else:
          break

      # treat words between single or double quotes as single word
      if quoted_str:
        if token[-1] == '"' or token[-1] == "'":
          token = quoted_str + ' ' + token[:len (token) - 1]
        else:
          quoted_str += ' ' + token
          continue
      else:
        if token[0] == '"':
          if token[-1] == '"':
            token = token[1:len (token) - 1]	# strip double quotes
          else:
            quoted_str = token[1:]
            continue
        if token[0] == "'":
          if token[-1] == "'":
            token = token[1:len (token) - 1]	# strip single quotes
          else:
            quoted_str = token[1:]
            continue

      val.append (token)
      quoted_str = ''
  elif _var_types[var] == 'string':
    if token[0] == '"' and token[-1] == '"':
      token = token[1:len (token) - 1]		# strip double quotes
    if token[0] == "'" and token[-1] == "'":
      token = token[1:len (token) - 1]		# strip single quotes
    val = token
  elif _var_types[var] == 'boolean':
    if token != 'false' and token != 'true':
      raise ConfigError ('%s must be a boolean "true" or "false"' % var,
      stream.lineno)
    if token == 'false':
      val = 0
    if token == 'true':
      val = 1

  return val

# parse depot section
def parse_section_depot (stream, conf):
  while 1:
    # argument will be a URL. allow anything but whitespace in the
    # URL. The caller should be smart enough to handle an invalid name
    stream.wordchars = string.letters + string.digits + string.punctuation
    url = stream.get_token ()
    if not url:
      raise ConfigError ('end of depot section unexpected',
                         stream.lineno)

    # make sure next token is section start symbol '{'
    stream.wordchars = '{'
    token = stream.get_token ()
    if not token:
      raise ConfigError ('end of depot section unexpected',
                         stream.lineno)
    if token != '{':
      raise ConfigError ('sections begin with {', stream.lineno)

    conf['depot'][url] = {}
    while 1:
      # get variable name or section end token
      stream.wordchars = string.letters + string.digits + '-_'
      token = stream.get_token ()
      if not token:
        raise ConfigError ('end of depot section unexpected',
                           stream.lineno)
      if token == '}':
        return
      if len (token) == 1 and not token in stream.wordchars:
        raise ConfigError ('invalid character in variable',
                           stream.lineno)

      if not _var_types.has_key (token):
        raise ConfigError ('invalid variable name "%s"' % token,
                           stream.lineno)
      if not token in SECTION_DEPOT_VARS:
        raise ConfigError ('invalid variable "%s" in dist-opts section' \
                           % token, stream.lineno)
      conf['depot'][url][token] = parse_assignment (token, stream)

# parse dist-opts section
def parse_section_dist_opts (stream, conf, systype):
  while 1:
    # argument will be a distribution. allow anything but whitespace.
    stream.wordchars = string.letters + string.digits + string.punctuation
    dist = stream.get_token ()
    if not dist:
      raise ConfigError ('end of dist-opts section unexpected', stream.lineno)

    # make sure next token is section start symbol '{'
    stream.wordchars = '{'
    token = stream.get_token ()
    if not token:
      raise ConfigError ('end of dist-opts section unexpected', stream.lineno)
    if token != '{':
      raise ConfigError ('sections begin with {', stream.lineno)

    conf['dist-opts'][dist] = {}
    while 1:
      # get variable name or section end token
      stream.wordchars = string.letters + string.digits + '-_'
      token = stream.get_token ()
      if not token:
        raise ConfigError ('end of dist-opts section unexpected', stream.lineno)
      if token == '}':
        return
      if len (token) == 1 and not token in stream.wordchars:
        raise ConfigError ('invalid character in variable', stream.lineno)

      if token == 'systype':
        parse_section_systype (stream, conf, systype, dist)
        continue

      if not _var_types.has_key (token):
        raise ConfigError ('invalid variable name "%s"' % token, stream.lineno)
      if not token in SECTION_DIST_OPTS_VARS:
        raise ConfigError ('invalid variable "%s" in dist-opts section' \
        % token, stream.lineno)
      conf['dist-opts'][dist][token] = parse_assignment (token, stream)

# parse systype section
def parse_section_systype (stream, conf, systype, dist = None):
  while 1:
    # argument will be a system type. allow anything but whitespace.
    stream.wordchars = string.letters + string.digits + string.punctuation
    config_systype = stream.get_token ()
    if not config_systype:
      raise ConfigError ('end of systype section unexpected', stream.lineno)

    # make sure next token is section start symbol '{'
    stream.wordchars = '{'
    token = stream.get_token ()
    if not token:
      raise ConfigError ('end of systype section unexpected', stream.lineno)
    if token != '{':
      raise ConfigError ('sections begin with {', stream.lineno)

    while 1:
      # get variable name or section end token
      stream.wordchars = string.letters + string.digits + '-_'
      token = stream.get_token ()
      if not token:
        raise ConfigError ('end of systype section unexpected', stream.lineno)
      if token == '}':
        return
      if len (token) == 1 and not token in stream.wordchars:
        raise ConfigError ('invalid character in variable', stream.lineno)

      _systype_matched = 0
      if config_systype[0] == '!':
        if not systype_match (config_systype[1:], systype):
          _systype_matched = 1
      else:
        if systype_match (config_systype, systype):
          _systype_matched = 1

      if not _var_types.has_key (token):
        raise ConfigError ('invalid variable name "%s"' % token,
                           stream.lineno)
      if not token in SECTION_SYSTYPE_VARS:
        raise ConfigError ('invalid variable "%s" in systype section' \
        % token, stream.lineno)
      if _systype_matched:
        if dist:
          conf['dist-opts'][dist][token] = parse_assignment (token, stream)
        else:
          conf[token] = parse_assignment (token, stream)
      else:
        parse_assignment (token, stream)
