# Copyright (C) 2000-2006, 2009-2010, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import stat
import string
import sys
import types

from urlparse import urlparse

import depot
import depotdb
import misclib
import myio
import mysig
import pkgdb
import pkginst
import pkginst.archive
import version

from misclib import sort_depot_package, urljoin
from packagenames import PackageName, PackageNames
from pkgdb import PkgDBVersion
from version import Version

# constants
READ_CHUNK = 20480		# number of bytes to read at a time

# options for depots.set()
SEARCH_ACCEPT_NO_DEPOTS = 1	# caller doesn't require a depot

# number of bytes in a checksum file
CHECKSUM_FILESIZE = 33          # pycurl can't determine this for all protocols

# IO Handler
io = myio.MyIO ()

# exceptions
class DepotError (Exception):
  """Base class for all exceptions raised by this module"""

  def __init__ (self, arg = None):
    self.msg = arg

  def __str__ (self):
    return self.msg

class DepotAccessDeniedError (DepotError):
  """access denied (authorization insufficient)"""

class DepotAccessError (DepotError):
  """unable to access depot"""

class DepotChecksumsOff (DepotError):
  """can't check local depot is up-to-date with checksums off"""

class DepotFileNotFoundError (DepotError):
  """file not found in depot"""

  def __init__ (self, file = None):
    self.file = file

  def __str__ (self):
    return self.file

class DepotGPGError (DepotError):
  """GPG signature for depot-db.xml failed"""

class DepotNoLocalDepot (DepotError):
  """no local depot specified with --local-depot"""

class DepotNoValidDepots (DepotError):
  """no valid depots"""

class DepotNotFoundError (DepotError):
  """depot not found"""

class DepotPackageVerNotFoundError (DepotError):
  """requested version of this package not found in depot"""

class DepotPermissionDeniedError (DepotError):
  """permission denied"""

class DepotProtocolUnsupportedError (DepotError):
  """protocol unsupported"""

class DepotProxyAuthenticationRequiredError (DepotError):
  """proxy authentication required"""

class DepotSftpAuthenticationError (DepotError):
  """authentication failure"""

  def __init__ (self, host):
    self.host = host

class DepotSftpKnownHostsFileNotFoundError (DepotError):
  """ssh_known_hosts file not found"""

class DepotSftpKnownHostsFileNotSpecifiedError (DepotError):
  """ssh_known_hosts file not specified"""

class DepotSftpKnownHostsPathNotFoundError (DepotError):
  """path specified for ssh_known_hosts not found"""

class DepotSftpInvalidKnownHostError (DepotError):
  """host unknown in ssh_known_hosts"""

class DepotSftpPrivateKeyEncryptedError (DepotError):
  """private key encrypted with passphrase"""

class DepotSftpPrivateKeyFileNotFoundError (DepotError):
  """private key file not found"""

class DepotSftpPrivateKeyFileNotSpecifiedError (DepotError):
  """private key file not specified"""

class DepotSSLError (DepotError):
  """SSL error"""

  def __init__ (self, msg):
    self.error = msg

class DepotTempdirCreationError (DepotError):
  """error creating temporary directory for any-url package retrieval"""

class ProgressAdaptor (object):
  """Adaptor to allow use of `TextProgressMeter` with pycurl.

  pycurls PROGRESSFUNCTION repeatedly calls the named function between
  downloaded chunks, but pkgutils potentially wants to show progress
  of several related files along a single `TextProgressMeter`. A
  `ProgressAdaptor` maintains the book-keeping to make that possible,
  while providing a `progress_cb` method to pass as the PROGRESSFUNCTION.
  """

  def __init__(self, progressmeter, fetched_bytes=0, additional_bytes=0):
    """Return an adaptor wrapped around `progressmeter`.

    :Parameters:
        `progressmeter`: TextProgressMeter
            Re-use `progressmeter` for fetching related files.
        `fetched_bytes`: integer
            Number of bytes already fetched as previous files using
            this `progressmeter`.
        `additional_bytes`: integer
            Additional bytes yet to be downloaded with further files
            using this `progressmeter`.
    """

    self.progressmeter = progressmeter
    self.fetched_bytes = fetched_bytes
    self.additional_bytes = additional_bytes

    self.running_total = 0

  def progress_cb (self, dltotal, dlnow, ultotal, ulnow):
    """Callback function for updating `progressmeter`.

    :Parameters:
        `dltotal`: float
            Total number of bytes to download for the current file.
        `dlnow`: float
            Number of bytes downloaded so far for the current file.
        `ultotal`: not used
        `ulnow`: not used

    Set pycurl.PROGRESSFUNCTION to call this function to redraw
    `progressmeter` taking into account other related files already
    fetched, and not yet started.
    """

    # pycurl tells us the total number of bytes that need to be fetched
    # for just the current file in `dltotal`, but when we call the start
    # method for `progressmeter` before any bytes have been fetched for
    # any files in the current group, we need to tell it the aggregated
    # size for the entire group:
    if dlnow == 0 or self.fetched_bytes == 0 and self.running_total == 0:
      self.progressmeter.start (dltotal + self.additional_bytes)

    self.running_total = int (dltotal) + self.fetched_bytes
    self.progressmeter.update (int (dlnow) + self.fetched_bytes)

  def get_running_total (self):
    """Return the number of bytes downloaded so far for the whole group."""
    return self.running_total


# proxy authentication variables
class ProxyAuth:
  def __init__ (self):
    self.host = ''
    self.login = ''
    self.passwd = ''

class ProtocolAuth:
  def __init__ (self):
    self.login = ''
    self.passwd = ''

    self.sftp_known_hosts_path = None	# path to ssh_known_hosts
    self.sftp_private_key_path = None	# path to ssh private key

# version number and revision class
class Depot:
  def __init__ (self, dep):
    if not isinstance (dep, basestring):
      url = dep.url
    else:
      url = dep

    self.archive = 0			# if depot is a zip archive

    self.url = url			# parent url
    self.pkg_url = url			# where to find packages
    self.type = urlparse (url).scheme	# URL scheme ('http', ...)
    self.local_path = None

    self.proxy = ProxyAuth ()
    self.auth = ProtocolAuth ()

    self.pkginst = None
    self.pkg_type = None
    self.info = None			# package information database
    self.update_types = []		# update types allowed
    self.update_type = None		# update type
    self.dist_ver = []			# distribution versions
    self.pkgdbs = []			# list of pkg-db.xml entries

    self.verify_checksum = True
    self.indent_progressmeter = 4

  # output depot path
  def __str__ (self):
    return self.url

  def set_proxy (self):
    pass

  def set_auth (self):
    pass

  # if package exists in local depot
  def pkg_exists (self, pkginst):
    if not os.path.exists (os.path.join (self.local_path, pkginst)):
      return False
    return True

  def open (self, path, mode):
    if self.local_path and os.path.exists (self.local_path + '/' + path):
      return open (self.local_path + '/' + path, 'r')
    else:
      return (self.urlopen (urljoin (self.pkg_url, path)))

  def fetch (self, pkg):
    pass

  def has_pkg (self, prog_name, prog_ver = None, pkg_rev = Version (0)):
    for pkgdb_entry in self.pkgdbs:
      if pkgdb_entry.prog_name == prog_name:
        if prog_ver:
          for pkgdb_ver in pkgdb_entry.version.keys ():
            if pkgdb_ver.prog_ver == prog_ver:
              if pkg_rev:
                if pkgdb_ver.pkg_rev == Version (pkg_rev):
                  return 1
              else:
                return 1
        else:
          return 1
    return 0

  def get_pkgdb_entry (self, prog_name):
    for pkgdb_entry in self.pkgdbs:
      if pkgdb_entry.prog_name == prog_name:
        return pkgdb_entry
    return None

class Depots:
  def __init__ (self, local_depot = '', msg_callback = None,
                package_names = PackageNames (), var = {}, check_sig = 0,
                gpg_keyring_path = '', pkg_type = '',
                tmpbasedir = '', add_package = None,
                assign_depot_vars = None, dtd_dirs = None):
    self.local_depot = local_depot		# local path to save
						# files to
    self.temp_package_dir = None	# where to store
                                                # any-url packages
    self.msg_callback = msg_callback
    self.package_names = package_names
    self.var = var
    self.check_sig = check_sig
    self.gpg_keyring_path = gpg_keyring_path
    self.pkg_type = pkg_type
    self.tmpbasedir = tmpbasedir
    self.add_package = add_package
    self.assign_depot_vars = assign_depot_vars
    self.dtd_dirs = dtd_dirs

    self.invalid_archive_package_callback = None
    self.local_depot_is_temporary = False	# if local path is a
						# temporary path because
						# a permanent path was
						# not specified
    self.depots = []

  # remove local repository
  def remove_local_depot (self, local_depot, msg_callback = None):
    if not local_depot and self.local_depot:
      if msg_callback:
        apply (msg_callback, ['begin', 'removing temporary \
local depot directory', 2])
      misclib.rmdir (self.local_depot)
      if msg_callback:
        apply (msg_callback, ['end', 'done', 2])

      mysig.unregister (self.remove_local_depot)

  # remove temporary package directory
  def remove_temp_package_dir (self, msg_callback = None):
    if self.temp_package_dir:
      if msg_callback:
        apply (msg_callback, ['begin', 'removing temporary \
package directory', 2])
      misclib.rmdir (self.temp_package_dir)
      if msg_callback:
        apply (msg_callback, ['end', 'done', 2])

      mysig.unregister (self.remove_temp_package_dir)
      self.temp_package_dir = None

  # walk list of depots for pb-db.xml entries
  def set (self, depot_names, pkgmgr, auth, options = []):
    self.depots = []
    valid_depots = []			# depots we can use
    search_for_packages_in_depots = False
    packages_to_search_for = []		# packages to search for

    # make copy as invalid_archive_package_callback might modify
    for packagename in self.package_names[:]:
      if not packagename.isfile ():
        # Unless we already have a local file, try to fetch the file from
        # (remote) url into a temporary local depot for processing later on.
        try:
          self.temp_package_dir = misclib.mkstempdir ()
        except misclib.MisclibTempdirCreationError, msg:
          raise DepotTempdirCreationError (str (msg))

        mysig.register (self.remove_temp_package_dir)

        dep = init (None, self.temp_package_dir,
                    packagename.remote_depot ())
        dep.auth = auth
        dep.indent_progressmeter = 2
        dep.verify_checksum = False

        # To minimize chatter, only admit that we created a temporary
        # directory if we actually need it...
        if dep.exists (packagename.remote_path ()):
          apply (self.msg_callback, ['begin', 'creating temporary \
directory', 0, misclib.VERBOSE_LOCAL_DEPOT])
          apply (self.msg_callback, ['end', self.temp_package_dir,
                                     0, misclib.VERBOSE_LOCAL_DEPOT])
          io.stdout (0, '', 'retrieving ' + \
                     packagename.geturl () + ' ...\n')
        elif packagename.getscheme () != 'file':
          raise DepotFileNotFoundError ('%s: File not found.' %
                                        packagename.geturl ())
        else:
          # ...otherwise, quietly remove it and pretend it was never
          # here:
          self.remove_temp_package_dir ()

        apply (self.assign_depot_vars, [dep])

        # If we successfully fetched a local copy of the remote file,
        # replace the remote packagename object with a file:// scheme
	# packagename:
        pkgfile = packagename.basename ()
        if dep.exists (pkgfile):
          dep.get (pkgfile, show_progress = True)
          packagename = \
            PackageName (os.path.join (self.temp_package_dir,
                                       pkgfile))
        
      if packagename.isfile ():
        path = packagename.normpath ()

        dep = init (None, self.local_depot, os.path.dirname (path))

        apply (self.assign_depot_vars, [dep])

        apply (self.msg_callback,
               ['begin', 'reading ' + 
                urljoin (dep.pkg_url, os.path.basename (path)),
                0, misclib.VERBOSE_DEPOT])

        try:
          archive = pkginst.archive.init (path)
        except pkginst.PkginstArchiveError, msg:
          if self.invalid_archive_package_callback:
            apply (self.invalid_archive_package_callback, [packagename.name])

          apply (self.msg_callback, ['end', 'failed', 0,
                                     misclib.VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg)])
          continue
        
        if archive:
          archive.path = path

          # set depot path to the archive file
          dep.local_path = path
          dep.url = packagename.getscheme () + "://" + path
          dep.archive = archive

          if archive.exists ('pkg-db.xml'):
            dep.pkgdbs = pkgdb.read_pkgdb_xml (archive, 'pkg-db.xml',
                                               self.pkg_type, pkgmgr,
                                               self.dtd_dirs)
          else:
            if self.invalid_archive_package_callback:
              apply (self.invalid_archive_package_callback, [packagename.name])

            apply (self.msg_callback, ['end', 'failed', 0,
                                       misclib.VERBOSE_DEPOT])
            apply (self.msg_callback, ['error', path + 'was specified as \
a pkg-inst archive but did not contain the "pkg-db.xml" member. \
archives must this contain this file. skipping.'])
            continue
        else:
          dep.pkgdbs = pkgdb.read_pkgdb_xml (archive, 'pkg-db.xml',
                                             self.pkg_type, pkgmgr,
                                             self.dtd_dirs)

        if not dep.pkgdbs:
          apply (self.msg_callback, ['end', 'failed', 0,
                                     misclib.VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', 'unable to parse the \
XML file. please check "' + path + '". it does not appear to follow \
pkg-db.xml(4). skipping.'])
          continue

        self.depots.append (dep)
        valid_depots.append (dep)

        apply (self.msg_callback, ['end', 'done', 0, misclib.VERBOSE_DEPOT])

        for pkgdb_entry in dep.pkgdbs:
          for pkgdb_ver in pkgdb_entry.version.keys ():
            pkgdb_entry_ver = pkgdb_entry.version[pkgdb_ver]

            package = apply (self.add_package, [pkgdb_entry.prog_name,
                                                pkgdb_ver.prog_ver,
                                                pkgdb_entry_ver.version,
                                                pkgdb_entry_ver.revision,
                                                archive])
            package.dep = dep
      else:
        packages_to_search_for.append (packagename.name)
        search_for_packages_in_depots = True
    else:
      # if packages specified, and at least one valid package
      # name remains, search repositories looking for it
      if self.package_names and packages_to_search_for:
        search_for_packages_in_depots = True

      # also search repositories if no packages given
      if not self.package_names:
        search_for_packages_in_depots = True

    # initialize depots
    if search_for_packages_in_depots and depot_names:
      for depot_name in depot_names:
        # create temp directory for local depot
        if is_remote (depot_name) and not self.local_depot:
          apply (self.msg_callback, ['begin', 'creating temporary \
local depot directory', 0, misclib.VERBOSE_LOCAL_DEPOT])

          try:
            self.local_depot = misclib.mkstempdir (self.tmpbasedir)
            self.local_depot_is_temporary = True
          except misclib.MisclibTempdirCreationError, msg:
            raise DepotNoLocalDepot (str (msg))

          mysig.register (self.remove_local_depot, None)
          apply (self.msg_callback, ['end', self.local_depot, 0,
                                     misclib.VERBOSE_LOCAL_DEPOT])

        try:
          dep = init (self.var, self.local_depot, depot_name)
        except DepotProtocolUnsupportedError, msg:
          apply (self.msg_callback, ['end', 'failed', 0,
                                     misclib.VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg) + ' removing ' +
                                     depot_name + ' from list of depots.'])
          continue

        try:
          apply (self.assign_depot_vars, [dep])
        except (DepotSftpKnownHostsFileNotFoundError,
                DepotSftpKnownHostsFileNotSpecifiedError,
                DepotSftpPrivateKeyFileNotFoundError,
                DepotSftpPrivateKeyFileNotSpecifiedError), msg:
          apply (self.msg_callback, ['end', 'failed', 0,
                                     misclib.VERBOSE_DEPOT])
          apply (self.msg_callback, ['error', str (msg)])
          continue

        self.depots.append (dep)

      # verify pkg-db.xml file in software depots
      new_depots = []
      for dep in self.depots:
        if dep.archive:
          new_depots.append (dep)
          continue

        try:
          if not dep.exists ('pkg-db.xml') and \
          not dep.exists ('depot-db.xml'):
            apply (self.msg_callback, ['error', 'no depot-db.xml or \
pkg-db.xml file found in ' + dep.url + '. removing from list of \
depots. the distribution name specified with --dist (default "' +
                                  self.var['dist'] + '") is \
appended to the repository path.'])
          else:
            new_depots.append (dep)
        except DepotSSLError, msg:
          if not egd_socket:
            apply (self.msg_callback, ['error', 'SSL error: ' +
                                       str (msg.error) + '. Maybe \
you need to specify the path to the socket EGD/PRNGD is listening \
on with --egd-socket?'])
          else:
            apply (self.msg_callback, ['error', 'SSL error: ' +
                                       str (msg.error) + '. Maybe \
the path you specified for --egd-socket is incorrect?'])
        except DepotAccessError, msg:
          apply (self.msg_callback, ['error', 'unable to access ' +
                                     str (dep) + '. the error \
received was: ' + str (msg) + '. removing from list of depots.'])
        except DepotAccessDeniedError:
          apply (self.msg_callback, ['error', 'access denied to ' +
                                     str (dep) + '. maybe you did \
not provide a correct login/password?'])
        except DepotProxyAuthenticationRequiredError:
          apply (self.msg_callback, ['error', 'access denied to proxy \
server (' + dep.proxy.host + '). either the proxy login id or \
password was incorrect or not specified. removing ' + dep.url +
                                ' from list of depots.'])
        except DepotPermissionDeniedError:
          apply (self.msg_callback, ['error', 'permission denied for ' +
                                     str (dep) + '. maybe the \
permissions on the file are incorrect?'])
        except DepotSftpAuthenticationError, msg:
          apply (self.msg_callback, ['error', 'unable to authenticate to ' +
                                     msg.host + ' using the private key \
specified. please either specify the correct private key or ensure \
the remote host has the contents of "' +
                                dep.auth.sftp_private_key_path +
                                '.pub" in ~' + dep.auth.login +
                                '/.ssh/authorized_keys. removing ' +
                                dep.url + ' from list of depots.'])
        except (DepotSftpInvalidKnownHostError,
                DepotSftpPrivateKeyEncryptedError), msg:
          apply (self.msg_callback, ['error', str (msg) + ' removing ' +
                                     dep.url + ' from list of depots.'])

      if not new_depots:
        if SEARCH_ACCEPT_NO_DEPOTS in options:
          self.depots = new_depots
        else:
          raise DepotNoValidDepots ('no depot to search for packages')
      else:
        self.depots = new_depots

      apply (self.msg_callback, ['end', 'reading depot databases',
                                 0, misclib.VERBOSE_DEPOT])

      for dep in self.depots:
        depot_db_xml = 0		# if depot-db.xml exists
        pb_db_xml = 0			# if pb-db.xml exists

        if dep.archive:			# skip archives
          continue

        # if parent depot contains more depots, add to list
        apply (self.msg_callback,
               ['begin', urljoin (dep.pkg_url, 'depot-db.xml'),
                2, misclib.VERBOSE_DEPOT])

        try:
          if dep.exists ('depot-db.xml'):
            depot_db_xml = 1

            dep.fetch ('depot-db.xml')	# make copy in local depot
            apply (self.msg_callback, ['end', 'done', 2,
                                       misclib.VERBOSE_DEPOT])

            # check depot-db.xml signature
            if self.check_sig:
              apply (self.msg_callback,
                     ['begin', urljoin (dep.pkg_url, 'depot-db.xml.gpg'),
                      2, misclib.VERBOSE_DEPOT])
              try:
                dep.fetch ('depot-db.xml.gpg')
                apply (self.msg_callback, ['end', 'done', 2,
                                           misclib.VERBOSE_DEPOT])

                apply (self.msg_callback,
                       ['begin', 'verifying GPG signature of ' +
                        urljoin (dep.pkg_url, 'depot-db.xml'), 2,
                        misclib.VERBOSE_DEPOT])
                try:
                  if misclib.verify_gpg_sig (os.path.join \
                                             (dep.local_path,
                                             'depot-db.xml'),
                                             os.path.join \
                                             (dep.local_path,
                                             'depot-db.xml.gpg'),
                                             self.gpg_keyring_path):
                    apply (self.msg_callback, ['end', 'ok', 2,
                                               misclib.VERBOSE_DEPOT])
                  else:
                    apply (self.msg_callback, ['end', 'failed', 2,
                                               misclib.VERBOSE_DEPOT])
                    raise DepotGPGError ('failed verifying GPG \
signature of ' + urljoin (dep.pkg_url, 'depot-db.xml'))
                except misclib.MisclibGPGProgNotFoundError, msg:
                  raise DepotGPGError (str (msg))
              except DepotFileNotFoundError:
                apply (self.msg_callback, ['end', 'failed (file not found)',
                                           2, misclib.VERBOSE_DEPOT])

            sub_depots = depotdb.read_depotdb_xml ('depot-db.xml', dep)

            # create new depots
            for sub_depot in sub_depots:
              depot_category = sub_depot.category

              # make sure distribution matches one we are interested in
              if sub_depot.dist_ver:
                if sub_depot.dist_ver not in dep.dist_ver:
                  continue

              # if architecture depot, we are only interested in
              # architecture we will perform install for
              if sub_depot.systype:
                if sub_depot.systype != self.var['systype']:
                  continue

              # if category associated with depot, use only if
              # it matches a category we are interested in
              if depot_category:
                # ignore src category
                if depot_category == 'src':
                  continue

                if dep.update_types:
                  if depot_category not in dep.update_types:
                    continue
                else:
                  # if no --update-type specified, add category only
                  # if not marked to be excluded (auto-include = "0")
                  if sub_depot.auto_include == '0':
                    continue

              apply (self.msg_callback,
                     ['end', 'adding ' +
                      urljoin (dep.pkg_url, sub_depot.name),
                      4, misclib.VERBOSE_DEPOT])

              new_dep = init (None, self.local_depot, dep, sub_depot.name)
              apply (self.assign_depot_vars, [new_dep])
              self.depots.append (new_dep)

              # record update type
              if not depot_category:
                new_dep.update_type = dep.update_type
              else:
                new_dep.update_type = depot_category
          else:
            apply (self.msg_callback, ['end', 'failed (not found)',
                                       2, misclib.VERBOSE_DEPOT])
        except DepotAccessDeniedError:
          apply (self.msg_callback, ['end', 'failed (access denied)',
                                     2, misclib.VERBOSE_DEPOT])
          continue
        except DepotPermissionDeniedError:
          apply (self.msg_callback, ['end', 'failed (permission denied)',
                                     2, misclib.VERBOSE_DEPOT])
          continue

        # depot has depot-db.xml, pkg-db.xml, or both
        try:
          apply (self.msg_callback,
                 ['begin', urljoin (dep.pkg_url, 'pkg-db.xml'),
                  2, misclib.VERBOSE_DEPOT])

          if not dep.exists ('pkg-db.xml'):
            apply (self.msg_callback, ['end', 'failed (not found)', 2,
                                       misclib.VERBOSE_DEPOT])
            continue
        except DepotPermissionDeniedError:
          apply (self.msg_callback, ['end', 'failed (permission denied)',
                                2, misclib.VERBOSE_DEPOT])
          continue

        valid_depots.append (dep)

        if not dep.local_uptodate ('pkg-db.xml'):
          dep.fetch ('pkg-db.xml')	# make copy in local depot
        apply (self.msg_callback, ['end', 'done', 2, misclib.VERBOSE_DEPOT])

        # check pkg-db.xml signature
        if self.check_sig:
          apply (self.msg_callback,
                 ['begin', urljoin (dep.pkg_url, 'pkg-db.xml.gpg'),
                  2, misclib.VERBOSE_DEPOT])
          try:
            dep.fetch ('pkg-db.xml.gpg')
            apply (self.msg_callback, ['end', 'done', 2,
                                       misclib.VERBOSE_DEPOT])

            apply (self.msg_callback,
                   ['begin', 'verifying GPG signature of ' +
                    urljoin (dep.pkg_url, 'pkg-db.xml'),
                    2, misclib.VERBOSE_DEPOT])
            if misclib.verify_gpg_sig (os.path.join (dep.local_path,
                                                     'pkg-db.xml'),
                                       os.path.join (dep.local_path,
                                                     'pkg-db.xml.gpg'),
                                       self.gpg_keyring_path):
              apply (self.msg_callback, ['end', 'ok', 2,
                                         misclib.VERBOSE_DEPOT])
            else:
              apply (self.msg_callback, ['end', 'failed', 0,
                                         misclib.VERBOSE_DEPOT])
              raise DepotGPGError ('failed verifying GPG signature \
of ' + urljoin (dep.pkg_url, 'pkg-db.xml'))
          except DepotFileNotFoundError:
            apply (self.msg_callback, ['end', 'failed (file not found)',
                                       2, misclib.VERBOSE_DEPOT])

        # check file signature
        if self.check_sig:
          apply (self.msg_callback,
                 ['begin', 'reading ' +
                  urljoin (dep.pkg_url, 'pkg-db.xml.gpg'),
                  2, misclib.VERBOSE_DEPOT])
          try:
            dep.fetch ('pkg-db.xml.gpg')
            apply (self.msg_callback, ['end', 'done', 2,
                                       misclib.VERBOSE_DEPOT])
          except DepotFileNotFoundError:
            apply (self.msg_callback, ['end', 'failed (file not found)',
                                       2, misclib.VERBOSE_DEPOT])

        dep.pkgdbs = pkgdb.read_pkgdb_xml (dep, 'pkg-db.xml',
                                           self.pkg_type, pkgmgr,
                                           self.dtd_dirs)

    self.depots = valid_depots

    # separate depot output and package installation
    if self.depots:
      apply (self.msg_callback, ['end', '', 0, misclib.VERBOSE_DEPOT])

    # must have at least one valid depot
    if not self.depots:
      if depot_names:
        if self.package_names and not packages_to_search_for:
          raise DepotNoValidDepots ('of the package names specified on \
the command-line, none were valid archive packages.')
        else:
          raise DepotNoValidDepots ('while some depot paths were given (' + \
          string.join (depot_names, ',') + '), none specified a path we \
could use. when a path is specified, it must be accessible and \
it must contain either a "depot-db.xml" file specifying \
subdirectories to descend into or a "pkg-db.xml" file specifying the \
list of packages in the directory.')
      else:
        if not SEARCH_ACCEPT_NO_DEPOTS in options:
          raise DepotNoValidDepots ('no depots given. please specify \
at least one depot with the --depot option or the path of a \
.pkg-inst archive file.')

  def append (self, dep):
    self.depots.append (dep)

  def __getitem__ (self, key):
    return self.depots[key]

  def __iter__ (self):
    return self.depots.__iter__ ()

  def __nonzero__ (self):
    return len (self.depots)

# test if URL is remote
def is_remote (url):
  if url.startswith ('ftp://'):
    return True
  elif url.startswith ('sftp://'):
    return True
  elif url.startswith ('http://'):
    return True
  elif url.startswith ('https://'):
    return True
  else:
    return False

def init (var, local_depot, dep, sub_depot = None):
  # dep is either a URL or an instance of the Depot class. if the
  # latter, dep is the parent URL of sub_depot.
  if not isinstance (dep, basestring):
    url = dep.url
  else:
    url = dep

  if url.startswith ('ftp://'):
    import depot.ftp
    return (depot.ftp.ftp (var, local_depot, dep, sub_depot))
  elif url.startswith ('sftp://'):
    try:
      import depot.sftp
      return (depot.sftp.sftp (var, local_depot, dep, sub_depot))
    except ImportError:
      raise DepotProtocolUnsupportedError ('sftp protocol not \
supported on this platform. one or more of the dependencies \
required to support it are unavailable.')
  elif url.startswith ('http://'):
    import depot.http
    return (depot.http.http (var, local_depot, dep, sub_depot))
  elif url.startswith ('https://'):
    import depot.https
    return (depot.https.https (var, local_depot, dep, sub_depot))
  elif url.startswith ('file://'):
    import depot.file
    return (depot.file.file (var, dep, sub_depot))
  else:
    import depot.file
    return (depot.file.file (var, 'file://' + dep, sub_depot))

# locate depot containing version of package. if no version specified,
# return depot with most recent version and release of package.
def find (depots, package, versions = [], pkg_dver = None,
pkg_drev = None):
  pkg_depot = None		# depot containing package
  prog_ver = None
  pkg_ver = None		# version of package in depot (if
				# package.prog_ver = 0)
  pkg_rev = None		# release of package in depot
  pkg_subpkg = None		# subpkgs available

  for dep in depots:
    if not dep.has_pkg (package.name):
      continue

    # if trying to match version in depot to satisfy a dependency,
    # test version against dependency criteria
    if pkg_dver or pkg_drev:
      pkgdb_entry = dep.get_pkgdb_entry (package.name)

      for dep_ver in pkgdb_entry.version.keys ():
        if pkgdb_entry.version[dep_ver].pkgname_base != \
        package.pkgname_base:
          continue

        ver = pkgdb_entry.version[dep_ver].version
        rev = pkgdb_entry.version[dep_ver].revision

        if version.met ([pkg_dver, pkg_drev],
                        package.make_version (ver, rev),
                        package.make_revision (ver, rev)) \
        and (not pkg_ver or ver > pkg_ver or \
        (ver == pkg_ver and rev > pkg_rev)):
          pkg_depot = dep
          prog_ver = dep_ver.prog_ver
          pkg_ver = ver
          pkg_rev = rev
          pkg_subpkg = pkgdb_entry.version[dep_ver].subpkg
    elif not package.prog_ver:
      [latest_prog_ver, latest_pkg_ver, latest_pkg_rev,
      latest_pkg_subpkg] = version.latest (dep, package.name,
      package.pkgname_base)
      if not latest_prog_ver:
        continue

      if not pkg_ver or latest_pkg_ver > pkg_ver or \
      (latest_pkg_ver == pkg_ver and latest_pkg_rev > pkg_rev):
        pkg_depot = dep
        pkg_ver = latest_pkg_ver
        pkg_rev = latest_pkg_rev
        pkg_subpkg = latest_pkg_subpkg
        prog_ver = latest_prog_ver
    else:
      # even though more than one depot might contain the version
      # for the package we are interested in, we return the depot
      # with the most recent version and release
      pkgdb_entry = dep.get_pkgdb_entry (package.name)
      for pkgdb_ver in pkgdb_entry.version.keys ():
        if pkgdb_ver.prog_ver != package.prog_ver:
          # if version not found, build list of versions we know about
          ver_string = pkgdb_ver.prog_ver + '-' + pkgdb_ver.pkg_rev
          if not ver_string in versions:
            versions.append (ver_string)
          continue

        if package.pkg_rev:
          pkgdb_ver = PkgDBVersion (package.prog_ver, package.pkg_rev)
          if not pkgdb_entry.version.has_key (pkgdb_ver) or \
          pkgdb_entry.version[pkgdb_ver].revision != package.pkg_rev:
            # if revision not found, build list of versions we know about
            ver_string = pkgdb_ver.prog_ver + '-' + pkgdb_ver.pkg_rev
            if not ver_string in versions:
              versions.append (ver_string)
            continue

        ver = pkgdb_entry.version[pkgdb_ver].version
        rev = pkgdb_entry.version[pkgdb_ver].revision
        subpkg = pkgdb_entry.version[pkgdb_ver].subpkg

        if not pkg_ver:
          pkg_ver = ver
        if not pkg_rev:
          pkg_rev = rev
        if not pkg_depot:
          pkg_depot = dep
        if not pkg_subpkg:
          pkg_subpkg = subpkg
        if not prog_ver:
          prog_ver = package.prog_ver

        if ver > pkg_ver or (ver == pkg_ver and rev > pkg_rev):
          pkg_depot = dep
          pkg_ver = ver
          pkg_rev = rev
          pkg_subpkg = subpkg
          prog_ver = package.prog_ver

  if package.prog_ver and not pkg_ver:
    raise DepotPackageVerNotFoundError
  if not pkg_depot:
    raise DepotNotFoundError

  return [pkg_depot, prog_ver, pkg_ver, pkg_rev, pkg_subpkg]

# locate depots containing all versions of package.
def find_all (depots, package_name, prog_vers = []):
  depot_versions = []

  for dep in depots:
    if not dep.has_pkg (package_name):
      continue

    pkgdb_entry = dep.get_pkgdb_entry (package_name)
    if not pkgdb_entry:
      continue

    versions = []
    for pkgdb_ver in pkgdb_entry.version.keys ():
      if prog_vers and not pkgdb_ver.prog_ver in prog_vers:
        continue
      versions.append ([pkgdb_ver.prog_ver,
      pkgdb_entry.version[pkgdb_ver].version, pkgdb_ver.pkg_rev])
    versions.sort (lambda x, y: cmp (x[1], y[1]))

    if not versions:
      continue

    for [prog_ver, pkg_ver, pkg_rev] in versions:
      # no duplicate versions. if duplicate exists, first version "wins".
      for depot_version in depot_versions:
        if depot_version['pkg ver'] == pkg_ver and \
        depot_version['pkg rev'] == pkg_rev:
          break
      else:
        depot_version = {}
        depot_version['depot'] = dep
        depot_version['prog ver'] = prog_ver
        depot_version['pkg ver'] = pkg_ver
        depot_version['pkg rev'] = pkg_rev

        if not depot_version in depot_versions:
          depot_versions.append (depot_version)

  depot_versions.sort (sort_depot_package, reverse = True)

  return depot_versions
