# Copyright (C) 2002-2003, 2005-2006, 2009, 2011,
#               2013 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import cgi
import os
import re
import string
import sys
import types

from lxml import etree, objectify
from copy import copy, deepcopy
from types import IntType, SliceType, StringType, UnicodeType

import pkgsys

from version import Version
from xmlutil import XMLDTDError, XMLParseError, DTD, dtd_search

# constants
XML_VERSION = "1.0"
XML_DTD = "pkg-db.dtd"
datadir = '/opt/TWWfsw/pkgutils16/share'

# if entry exists in package database
def exists (package, pkgmgr, pkgname, pkg_db, dtd_dirs):
  pkgdbs = []
  filename = package.name + '-' + package.prog_ver
  pkgdb_path = os.path.join (pkg_db, filename)
  pkgdb_entry_ver = package.pkgdb_entry_ver

  # read existing database entry
  if os.path.exists (pkgdb_path):
    pkgdbs = read_pkg_db (pkgdb_path, package.pkg_type, pkgmgr, dtd_dirs)
    pkgdb_entry = pkgdbs[0]

    pkgdb_ver = pkgdb_entry.version.keys ()[0]
    old_pkgdb_entry_ver = pkgdb_entry.version[pkgdb_ver]
  else:
    return 0

  # false if package has version/revision different than
  # what is installed
  if pkgdb_entry_ver.version != old_pkgdb_entry_ver.version:
    return 0
  if pkgdb_entry_ver.revision != old_pkgdb_entry_ver.revision:
    return 0

  # write out new subpkg entry for installed package but only
  # if not in existing entry
  if pkgdbs:
    subpkgs = old_pkgdb_entry_ver.subpkg
    for subpkg_type in subpkgs.keys ():
      subpkg = subpkgs[subpkg_type]
      for pkgname_ext in subpkg['names']:
        if pkgname[len (pkgdb_entry_ver.pkgname_base):] == pkgname_ext:
          if pkgdb_entry_ver.version != subpkg['version']:
            return 0
          elif pkgdb_entry_ver.revision != subpkg['revision']:
            return 0
          else:
            return 1

  return 0

# add entry to package database
def install (package, pkgmgr, pkgname, dry_run, pkg_db, dtd_dirs, output):
  pkgdbs = []
  filename = package.name + '-' + package.prog_ver
  pkgdb_path = os.path.join (pkg_db, filename)
  pkgdb_entry_ver = package.pkgdb_entry_ver

  apply (output, ['begin', 'updating ' +
  package.get_subpkg_type (pkgname) + ' fileset package database entry'])
  if dry_run:
    apply (output, ['end', 'done'])
    return

  # read existing database entry
  if os.path.exists (pkgdb_path):
    pkgdbs = read_pkg_db (pkgdb_path, package.pkg_type, pkgmgr, dtd_dirs)
    pkgdb_entry = pkgdbs[0]

    pkgdb_ver = pkgdb_entry.version.keys ()[0]
    old_pkgdb_entry_ver = pkgdb_entry.version[pkgdb_ver]

  fh = open (pkgdb_path, 'w')
  fh.write ('<?xml version="%s"?>\n' % XML_VERSION)
  fh.write ('<!DOCTYPE packages SYSTEM "%s">\n' % XML_DTD)
  fh.write ('\
<packages>\n\
  <package name="' + package.name + '" version="' + package.prog_ver + '">\n')

  if pkgdb_entry_ver.description:
    fh.write ('\
    <description')
    if pkgdb_entry_ver.description['attribution']:
      fh.write (' attribution="' + 
      pkgdb_entry_ver.description['attribution'] + '">\n')
    else:
      fh.write ('>\n')
    fh.write (cgi.escape (pkgdb_entry_ver.description['text']) + '\n\
    </description>\n\n')

  # write <licenses> entry
  write_licenses_entry (fh, pkgdb_entry_ver.licenses[:])

  # write <notes> entry
  write_notes_entry (fh, pkgdb_entry_ver.notes[:])

  version = package.make_version (package.pkg_ver, package.pkg_rev)

  fh.write ('\
    <package-manager name="' + package.pkg_type + '">\n')
  if pkgdb_entry_ver.install_name:
    fh.write ('\
      <install-name>' + pkgdb_entry_ver.install_name + '</install-name>\n')
  else:
    fh.write ('\
      <install-name/>\n')
  fh.write ('\
      <pkgname-base>' + pkgdb_entry_ver.pkgname_base + '</pkgname-base>\n\
      <version>' + version + '</version>\n\
      <revision>' + package.pkg_rev + '</revision>\n')

  # print subpkg entries for existing entry
  if pkgdbs:
    subpkgs = old_pkgdb_entry_ver.subpkg
    for subpkg_type in old_pkgdb_entry_ver.subpkg.keys ():
      subpkg = old_pkgdb_entry_ver.subpkg[subpkg_type]
      for pkgname_ext in subpkg['names']:
        # write subpkg entry if it does not match current subpkg. we
        # always replace existing subpkg with subkpg from package
        # being installed.
        if pkgname[len (pkgdb_entry_ver.pkgname_base):] != pkgname_ext:
          fh.write ('\n\
      <subpkg type="' + subpkg_type + '" pkgname-ext="' + pkgname_ext + '"\n')
          fh.write ('\
      size="' + str(subpkg['size']) + '"')
          if subpkg['order']:
            fh.write (' order="' + str (subpkg['order']) + '"')
          fh.write ('>\n')

          fh.write ('\
        <version>' + subpkg['version'] + '</version>\n\
        <revision>' + subpkg['revision'] + '</revision>\n')

          if subpkg.has_key ('depend'):
            # print sub-package dependencies
            depend = subpkg['depend']
            for pkgname_base in depend.keys ():
              for depend_subpkg in depend[pkgname_base].keys ():
                fh.write ('\
        <depend pkgname-base="' + pkgname_base + '"\n\
          subpkg="' + depend_subpkg + '">' +
                package.get_pkgdb_depend_str \
                (depend[pkgname_base][depend_subpkg]['test']) + '</depend>\n')

          fh.write ('\
      </subpkg>\n')

  subpkg_type = package.get_subpkg_type (pkgname)

  fh.write ('\n\
      <subpkg type="' + subpkg_type + '" pkgname-ext="' +
            pkgname[len (pkgdb_entry_ver.pkgname_base):] + '"\n')
  fh.write ('      size="' +
            str (pkgdb_entry_ver.subpkg[subpkg_type]['size']) + '"')
  if pkgdb_entry_ver.subpkg[subpkg_type]['order']:
    fh.write (' order="' +
              str (pkgdb_entry_ver.subpkg[subpkg_type]['order']) + '"')
  fh.write ('>\n')

  fh.write ('\
        <version>' + version + '</version>\n\
        <revision>' + package.pkg_rev + '</revision>\n')

  subpkg = pkgdb_entry_ver.subpkg[subpkg_type]
  if subpkg.has_key ('depend'):
    # print sub-package dependencies
    depend = subpkg['depend']
    for pkgname_base in depend.keys ():
      for depend_subpkg in depend[pkgname_base].keys ():
        fh.write ('\
        <depend pkgname-base="' + pkgname_base + '"\n\
          subpkg="' + depend_subpkg + '">' +
        package.get_pkgdb_depend_str \
        (depend[pkgname_base][depend_subpkg]['test']) + '</depend>\n')

  fh.write ('\
      </subpkg>\n')

  # write <changelog> entry
  write_changelog_entry (fh, pkgdb_entry_ver.changelog[:])

  fh.write ('\
    </package-manager>\n\
  </package>\n\
</packages>\n')
  fh.close ()

  apply (output, ['end', 'done'])

# remove entry from package database
def remove (package, pkgmgr, pkgname, dry_run, pkg_db, dtd_dirs,
            output, msg = None):
  pkgdbs = []
  filename = package.name + '-' + package.prog_ver
  pkgdb_path = os.path.join (pkg_db, filename)
  pkgdb_entry_ver = package.pkgdb_entry_ver

  if msg:
    apply (output, ['begin', msg])
  else:
    apply (output, ['begin', 'removing ' +
    package.get_subpkg_type (pkgname) + ' fileset package database entry'])
  if dry_run:
    apply (output, ['end', 'done'])
    return

  # read existing database entry
  if os.path.exists (pkgdb_path):
    pkgdbs = read_pkg_db (pkgdb_path, package.pkg_type, pkgmgr, dtd_dirs)
    pkgdb_entry = pkgdbs[0]

    pkgdb_ver = pkgdb_entry.version.keys ()[0]
    old_pkgdb_entry_ver = pkgdb_entry.version[pkgdb_ver]

    # remove subpkg entry
    subpkgs = old_pkgdb_entry_ver.subpkg
    for subpkg_type in old_pkgdb_entry_ver.subpkg.keys ():
      new_subpkg_names = []

      subpkg = old_pkgdb_entry_ver.subpkg[subpkg_type]
      for pkgname_ext in subpkg['names']:
        if not pkgname[len (pkgdb_entry_ver.pkgname_base):] == pkgname_ext:
          new_subpkg_names.append (pkgname_ext)
      subpkg['names'] = new_subpkg_names

      # if all packages remove for sub-package, remove sub-package
      if not subpkg['names']:
        del subpkgs[subpkg_type]

    # update subpkg entries
    pkgdb_entry_ver.subpkg = subpkgs

    if not old_pkgdb_entry_ver.subpkg:
      os.remove (os.path.join (pkg_db, package.name + '-' + package.prog_ver))

      apply (output, ['end', 'done'])
      return

  fh = open (pkgdb_path, 'w')
  fh.write ('<?xml version="%s"?>\n' % XML_VERSION)
  fh.write ('<!DOCTYPE packages SYSTEM "%s">\n' % XML_DTD)
  fh.write ('\
<packages>\n\
  <package name="' + package.name + '" version="' + package.prog_ver + '">\n')

  if pkgdb_entry_ver.description:
    fh.write ('\
    <description')
    if pkgdb_entry_ver.description['attribution']:
      fh.write (' attribution="' + 
      pkgdb_entry_ver.description['attribution'] + '">\n')
    else:
      fh.write ('>\n')
    fh.write (cgi.escape (pkgdb_entry_ver.description['text']) + '\n\
    </description>\n\n')

  # write <licenses> entry
  write_licenses_entry (fh, pkgdb_entry_ver.licenses[:])

  # write <notes> entry
  write_notes_entry (fh, pkgdb_entry_ver.notes[:])

  version = package.make_version (package.pkg_ver, package.pkg_rev)

  fh.write ('\
    <package-manager name="' + package.pkg_type + '">\n')
  if pkgdb_entry_ver.install_name:
    fh.write ('\
      <install-name>' + pkgdb_entry_ver.install_name + '</install-name>\n')
  else:
    fh.write ('\
      <install-name/>\n')
  fh.write ('\
      <pkgname-base>' + pkgdb_entry_ver.pkgname_base + '</pkgname-base>\n\
      <version>' + version + '</version>\n\
      <revision>' + package.pkg_rev + '</revision>\n')

  # print subpkg entries for existing entry
  if pkgdbs:
    subpkgs = old_pkgdb_entry_ver.subpkg
    for subpkg_type in old_pkgdb_entry_ver.subpkg.keys ():
      subpkg = old_pkgdb_entry_ver.subpkg[subpkg_type]
      for pkgname_ext in subpkg['names']:
        fh.write ('\n\
      <subpkg type="' + subpkg_type + '" pkgname-ext="' + pkgname_ext + '"\n')
        fh.write ('\
      size="' + str(subpkg['size']) + '"')
        if subpkg['order']:
          fh.write (' order="' + str (subpkg['order']) + '"')
        fh.write ('>\n')

        fh.write ('\
        <version>' + subpkg['version'] + '</version>\n\
        <revision>' + subpkg['revision'] + '</revision>\n')

        if subpkg.has_key ('depend'):
          # print sub-package dependencies
          depend = subpkg['depend']
          for pkgname_base in depend.keys ():
            for depend_subpkg in depend[pkgname_base].keys ():
              fh.write ('\
        <depend pkgname-base="' + pkgname_base + '"\n\
          subpkg="' + depend_subpkg + '">' +
              package.get_pkgdb_depend_str \
              (depend[pkgname_base][depend_subpkg]['test']) + '</depend>\n')

        fh.write ('\
      </subpkg>\n')

  # write <changelog> entry
  write_changelog_entry (fh, pkgdb_entry_ver.changelog[:])

  fh.write ('\
    </package-manager>\n\
  </package>\n\
</packages>\n')
  fh.close ()

  apply (output, ['end', 'done'])

# write <licenses> entry
def write_licenses_entry (fh, licenses):
  if not len (licenses):
    return

  fh.write ('\
    <licenses>\n')

  while licenses:
    license = licenses.pop (0)
    print_license_entry (fh, license)

    # separate <license> entries with blank line
    if licenses:
      fh.write ('\n')

  fh.write ('\
    </licenses>\n\n')

# write <license> entry
def print_license_entry (fh, license):
  items = license.items[:]

  fh.write ('\
      <license')
  if license.module:
    fh.write (' module="' + license.module + '"')

  if license.path:
    fh.write (' path="' + license.path + '"/>\n')
  elif not items:
    fh.write ('/>\n')
  else:
    fh.write ('>\n')
    while items:
      skip = 1
      item = items.pop (0)

      if item.type == 'indent':
        fh.write ('\
        <indent>\n')

        write_indent_entry (fh, item.items[:], 10)

        fh.write ('\
        </indent>\n')
      elif item.type == 'orderedlist':
        fh.write ('\
        <orderedlist')
        if item.list_type != 'decimal':
          fh.write (' type="' + item.list_type + '"')
        if item.start_num != '1':
          fh.write (' start-num="' + item.start_num + '"')
        fh.write ('>\n')

        for list_item in item.items:
          write_item_entry (fh, list_item.items[:], 10)

        fh.write ('\
        </orderedlist>\n')
      elif item.type == 'para':
        skip = item.skip

        fh.write ('\
        <para')
        if item.align != 'left':
          fh.write (' align="%s"' % item.align)
        if item.skip != 1:
          fh.write (' skip="%d"' % item.skip)
        fh.write ('>' + build_item_text (item) + '</para>\n')
      elif item.type == 'screen':
        text = cgi.escape (item.text)
        fh.write ('\
        <screen>\n' + text + '\n' + '\
        </screen>\n')
      elif item.type == 'unorderedlist':
        fh.write ('\
        <unorderedlist>\n')

        for list_item in item.items:
          write_item_entry (fh, list_item.items[:], 10)

        fh.write ('\
        </unorderedlist>\n')

      if items:
        if item.type != 'para':
          fh.write ('\n')
        if item.type == 'para' and items[0].type == 'para' and skip:
          fh.write ('\n')

    fh.write ('\
      </license>\n')

# write <notes> entry
def write_notes_entry (fh, notes):
  while notes:
    notes_entry = notes.pop (0)

    fh.write ('\
    <notes')
    if notes_entry.name:
      fh.write (' name="' + notes_entry.name + '"')
    fh.write ('>\n')

    notes_entry_items = notes_entry[:]
    while notes_entry_items:
      notes_entry_item = notes_entry_items.pop (0)

      if notes_entry_item.type == 'change':
        fh.write ('\
      <change from="' + notes_entry_item.from_version + '" to="' + \
          notes_entry_item.to_version + '">\n')

        write_change_entries (fh, notes_entry_item.items[:])

        fh.write ('\
      </change>\n')
      else:
        write_note_entry (fh, notes_entry_item)

      if notes_entry_items:
        fh.write ('\n')

    fh.write ('\
    </notes>\n\n')

# write <note> entry
def write_note_entry (fh, note):
  if note.type == 'platform':
    fh.write ('\
      <platform name="' + cgi.escape (note.name) + '">\n')
  else:
    fh.write ('\
      <note type="' + cgi.escape (note.type) + '">\n')

  items = note.items[:]
  while items:
    skip = 1
    item = items.pop (0)

    if item.type == 'indent':
      fh.write ('\
        <indent>\n')

      write_indent_entry (fh, item.items[:], 10)

      fh.write ('\
        </indent>\n')
    elif item.type == 'orderedlist':
      fh.write ('\
        <orderedlist')
      if item.list_type != 'decimal':
        fh.write (' type="' + item.list_type + '"')
      if item.start_num != '1':
        fh.write (' start-num="' + item.start_num + '"')
      fh.write ('>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], 10)

      fh.write ('\
        </orderedlist>\n')
    elif item.type == 'para':
      skip = item.skip

      fh.write ('\
        <para')
      if item.align != 'left':
        fh.write (' align="%s"' % item.align)
      if item.skip != 1:
        fh.write (' skip="%d"' % item.skip)
      fh.write ('>' + build_item_text (item) + '</para>\n')
    elif item.type == 'screen':
      text = cgi.escape (item.text)
      fh.write ('\
        <screen>\n' + text + '\n' + '\
        </screen>\n')
    elif item.type == 'unorderedlist':
      fh.write ('\
        <unorderedlist>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], 10)

      fh.write ('\
        </unorderedlist>\n')

    if items:
      if item.type != 'para':
        fh.write ('\n')
      if item.type == 'para' and items[0].type == 'para' and skip:
        fh.write ('\n')

  if note.type == 'platform':
    fh.write ('\
      </platform>\n')
  else:
    fh.write ('\
      </note>\n')

# write <changelog> entry
def write_changelog_entry (fh, changelog):
  if not changelog:
    return

  fh.write ('\n\
      <changelog>\n')

  while changelog:
    change = changelog.pop (0)

    fh.write ('\
        <change date="' + change.date + '" revision="' + \
    change.revision + '"\n\
        author="' + cgi.escape (change.author) + '" email="' +
    cgi.escape (change.email) + '">\n')

    write_change_entries (fh, change.items[:], 2)

    fh.write ('\
        </change>\n')

    if changelog:
      fh.write ('\n')

  fh.write ('\
      </changelog>\n')

# write <change> entries
def write_change_entries (fh, items, indent = 0):
  indent_space = ' ' * indent

  while items:
    item = items.pop (0)

    if item.type == 'item':
      write_item_entry (fh, item.items[:], indent + 8)
    elif item.type == 'item group':
      fh.write (indent_space + '\
        <items name="' + cgi.escape (item.group) + '">\n')

      write_change_entries (fh, item.items[:], indent + 2)

      fh.write (indent_space + '\
        </items>\n')
    elif item.type == 'orderedlist':
      fh.write (indent_space + '<orderedlist')
      if item.list_type != 'decimal':
        fh.write (' type="' + item.list_type + '"')
      if item.start_num != '1':
        fh.write (' start-num="' + item.start_num + '"')
      fh.write ('>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</orderedlist>\n')
    elif item.type == 'para':
      skip = item.skip

      fh.write (indent_space + '<para')
      if item.align != 'left':
        fh.write (' align="%s"' % item.align)
      if item.skip != 1:
        fh.write (' skip="%d"' % item.skip)
      fh.write ('>' + build_item_text (item) + '</para>\n')
    elif item.type == 'screen':
      text = cgi.escape (item.text)
      fh.write (indent_space + '<screen>\n' + text + '\n' +
                indent_space + '</screen>')
    elif item.type == 'unorderedlist':
      fh.write (indent_space + '<unorderedlist>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</unorderedlist>\n')

# write <item> entry
def write_item_entry (fh, items, indent):
  indent_space = ' ' * indent

  fh.write (indent_space + '<item>')

  while items:
    skip = 1
    item = items.pop (0)

    if item.type == 'indent':
      fh.write ('<indent>\n')

      write_indent_entry (fh, item.items[:], indent + 2)

      fh.write (indent_space + '</indent>')
    elif item.type == 'orderedlist':
      fh.write ('<orderedlist')
      if item.list_type != 'decimal':
        fh.write (' type="' + item.list_type + '"')
      if item.start_num != '1':
        fh.write (' start-num="' + item.start_num + '"')
      fh.write ('>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</orderedlist>')
    elif item.type == 'para':
      skip = item.skip

      fh.write ('<para')
      if item.align != 'left':
        fh.write (' align="%s"' % item.align)
      if item.skip != 1:
        fh.write (' skip="%d"' % item.skip)
      fh.write ('>' + build_item_text (item) + '</para>')
    elif item.type == 'screen':
      text = cgi.escape (item.text)
      fh.write ('<screen>\n' + text + '\n' +
      indent_space + '</screen>')
    elif item.type == 'unorderedlist':
      fh.write ('<unorderedlist>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</unorderedlist>')

    if items:
      fh.write ('\n' + indent_space)

  fh.write ('</item>\n')

# write <indent> entry
def write_indent_entry (fh, items, indent):
  indent_space = ' ' * indent

  while items:
    skip = 1
    item = items.pop (0)

    if item.type == 'indent':
      fh.write (indent_space + '<indent>\n')

      write_indent_entry (fh, item.items[:], indent + 2)

      fh.write (indent_space + '</indent>')
    elif item.type == 'orderedlist':
      fh.write (indent_space + '<orderedlist')
      if item.list_type != 'decimal':
        fh.write (' type="' + item.list_type + '"')
      if item.start_num != '1':
        fh.write (' start-num="' + item.start_num + '"')
      fh.write ('>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</orderedlist>')
    elif item.type == 'para':
      skip = item.skip

      fh.write (indent_space + '<para')
      if item.align != 'left':
        fh.write (' align="%s"' % item.align)
      if item.skip != 1:
        fh.write (' skip="%d"' % item.skip)
      fh.write ('>' + build_item_text (item) + '</para>')
    elif item.type == 'screen':
      text = cgi.escape (item.text)
      fh.write (indent_space + '<screen>\n' + text + '\n' +
                indent_space + '</screen>')
    elif item.type == 'unorderedlist':
      fh.write (indent_space + '<unorderedlist>\n')

      for list_item in item.items:
        write_item_entry (fh, list_item.items[:], indent + 2)

      fh.write (indent_space + '</unorderedlist>')

    fh.write ('\n')

    # separate some items by newline
    if items:
      if item.type != 'para':
        fh.write ('\n')
      if item.type == 'para' and items[0].type == 'para' and skip:
        fh.write ('\n')

# build text for <item> or <para>
def build_item_text (items):
  text = ''

  for item_entry in items:
    if item_entry.type == 'command':
      text += '<command>' + cgi.escape (item_entry.text) + '</command>'
    elif item_entry.type == 'link':
      text += '<ulink url="' + cgi.escape (item_entry.url) + \
      '">' + cgi.escape (item_entry.value) + '</ulink>'
    elif item_entry.type == 'text':
      text += cgi.escape (item_entry.text)

  return text

# read pkg-db.xml from a depot object
def read_pkgdb_xml (dep, filename, pkg_type, pkgmgr, dtd_dirs):
  pkgdb_handle = dep.open (filename, 'r')

  return read_pkg_db (pkgdb_handle, pkg_type, pkgmgr, dtd_dirs)

# read pkg-db.xml file
def read_pkg_db (pkgdb_path, pkg_type, pkgmgr, dtd_dirs):
  # Find a suitable DTD early, incase pkgdb_path is an open file handle
  # (which may be an unlinked file after objectify.parse() closes it).
  dtd = dtd_search (dtd_dirs, pkgdb_path, 'pkg-db.dtd')
  
  if type (pkgdb_path) in (types.StringType, types.UnicodeType):
    filename = pkgdb_path
  else:
    filename = pkgdb_path.name
    
  try:
    xml = objectify.parse (pkgdb_path)
  except etree.XMLSyntaxError, msg:
    raise XMLParseError ('Syntax error while parsing ' + filename +
                         ', the following error was encountered: ' +
                         str (msg))

  # license paths need to be resolved and substituted before validation
  for xlicenses in xml.iterfind ('.//licenses'):
    children = [xchild for xchild in xlicenses.iterfind ('.//license')]
    for i in range (len (children)):
      path = xlicenses.license[i].get ('path')
      if path is not None:
        fpath = os.path.join (datadir, 'licenses', path)

        try:
          flicense = open (fpath, 'r')
        except (IOError, OSError), msg:
          raise XMLParseError ('the license file could not be read, the ' +
                               'following error was encountered: ' +
                               str (msg))

        try:
          xlicense = objectify.parse (flicense).getroot ()
        except etree.XMLSyntaxError, msg:
          raise XMLParseError ('Syntax error while parsing ' + fpath +
                               ', the following error was encountered: ' +
                               str (msg))

        module = xlicenses.license[i].get ('module')
        xlicenses.license[i] = xlicense

        # retain the attributes
        if module is not None:
          xlicenses.license[i].set ('module', module)
        if path is not None:
          xlicenses.license[i].set ('path', path)

  # Use DTD to validate pkg-db.xml
  if dtd:
    pkgdtd = DTD (dtd)
    if not pkgdtd.validate (xml):
      raise XMLDTDError ('DTD validation error while parsing ' + pkgdb_path +
                         ', the following error was encountered: ' +
                         pkgdtd.error_log().filter_from_errors()[0].message)

  pkgdb_entries = PkgDBEntries (xml, pkgdb_path, pkg_type, pkgmgr)

  return pkgdb_entries.pkgdbs


class PkgDBEntries:
  def __init__ (self, xpackages, filename, pkg_type, pkgmgr):
    self.xpackages = xpackages
    self.filename = filename
    self.pkg_type_to_save = pkg_type
    self.pkgmgr = pkgmgr
    self.pkgdbs = []

    self.package = pkgsys.init (pkg_type, pkgmgr)

    for xpackage in xpackages.findall ('package'):
      self.prog_name = xpackage.get ('name')
      self.prog_ver = xpackage.get ('version')
      self.prog_rev = int(xpackage.get ('revision', '0'))
      
      self.description = self._description (xpackage, {})
      self.licenses = self._licenses (xpackage, [])
      self.notes = self._notes (xpackage, [])

      # This method updates self.pkgdbs
      self._package_managers (xpackage)

  def _change (self, xchange):
    change = PkgDBChange (xchange.get ('date'),
                          xchange.get ('revision'),
                          xchange.get ('author'),
                          xchange.get ('email'),
                          xchange.get ('from'),
                          xchange.get ('to'))

    for xchild in xchange.iterchildren ():
      if xchild.tag == 'item':
        change.append (self._item (xchild))
      elif xchild.tag == 'items':
        change.append (self._items (xchild))
      elif xchild.tag == 'orderedlist':
        change.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        change.append (self._para (xchild))
      elif xchild.tag == 'screen':
        change.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        change.append (self._unorderedlist (xchild))

    return change

  def _changelog (self, xparent, default):
    xchangelog = xparent.find ('changelog')
    if xchangelog is None:
      return default

    allxchange = xchangelog.findall ('change')
    if allxchange is None:
      return default

    changelogs = []
    for xchange in allxchange:
      changelogs.append (self._change (xchange))

    return changelogs

  def _command (self, xcommand):
    return PkgDBNoteItemCommand (xcommand.text)

  def _data (self, xparent, default):
    xdata = xparent.find ('data')
    if xdata is None:
      return default

    data = {}
    data['checksum'] = xdata.get ('checksum')
    data['checksum type'] = xdata.get ('checksum-type')
    data['encoding'] = xdata.get ('encoding')
    data['filename'] = xdata.get ('filename')
    data['size'] = xdata.get ('size')

    return data

  def _depends (self, xparent, default):
    allxdepend = xparent.findall ('depend')
    if allxdepend is None:
      return default

    depends = {}
    for xdepend in allxdepend:
      depend_pkg = xdepend.get ('pkgname-base')
      depend_subpkg = xdepend.get ('subpkg')
      depend_val = xdepend.text

      if not depends.has_key (depend_pkg):
        depends[depend_pkg] = {}
      if not depends[depend_pkg].has_key (depend_subpkg):
        depends[depend_pkg][depend_subpkg] = {}
      depends[depend_pkg][depend_subpkg]['test'] = depend_val

    return depends

  def _description (self, xparent, default):
    xdescription = xparent.find ('description')
    if xdescription is None:
      return default

    description = {}
    description['attribution'] = xdescription.get ('attribution', '')
    description['text'] = xdescription.text.strip ()
    return description

  def _indent (self, xindent):
    indent = PkgDBNoteIndent ()

    for xchild in xindent.iterchildren ():
      if xchild.tag == 'indent':
        indent.append (self._indent (xchild))
      elif xchild.tag == 'orderedlist':
        indent.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        indent.append (self._para (xchild))
      elif xchild.tag == 'screen':
        indent.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        indent.append (self._unorderedlist (xchild))

    return indent

  def _install_name (self, xparent, default):
    xinstall_name = xparent.find ('install-name')
    if xinstall_name is None:
      return default

    return xinstall_name.text

  def _item (self, xitem):
    item = PkgDBNoteItem ()

    for xchild in xitem.iterchildren ():
      if xchild.tag == 'indent':
        item.append (self._indent (xchild))
      elif xchild.tag == 'para':
        item.append (self._para (xchild))
      elif xchild.tag == 'screen':
        item.append (self._screen (xchild))
      elif xchild.tag == 'orderedlist':
        item.append (self._orderedlist (xchild))
      elif xchild.tag == 'unorderedlist':
        item.append (self._unorderedlist (xchild))

    return item

  def _items (self, xitems):
    items = PkgDBNoteItems (xitems.get ('name'))

    for xchild in xitems.iterchildren ():
      if xchild.tag == 'item':
        items.append (self._item (xchild))
      elif xchild.tag == 'orderedlist':
        items.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        items.append (self._para (xchild))
      elif xchild.tag == 'screen':
        items.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        items.append (self._unorderedlist (xchild))

    return items

  def _license (self, xlicense):
    license = PkgDBLicense (xlicense.get ('module'),
                            xlicense.get ('path'))

    for xchild in xlicense.iterchildren ():
      if xchild.tag == 'indent':
        license.append (self._indent (xchild))
      elif xchild.tag == 'orderedlist':
        license.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        license.append (self._para (xchild))
      elif xchild.tag == 'screen':
        license.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        license.append (self._unorderedlist (xchild))

    return license

  def _licenses (self, xparent, default):
    xlicenses = xparent.find ('licenses')
    if xlicenses is None:
      return default

    licenses = PkgDBLicenses ()
    for xlicense in xlicenses.iterchildren ():
      licenses.append (self._license (xlicense))

    return licenses

  def _note (self, xnote):
    note = PkgDBNote (xnote.get ('type'))

    for xchild in xnote.iterchildren ():
      if xchild.tag == 'indent':
        note.append (self._indent (xchild))
      elif xchild.tag == 'orderedlist':
        note.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        note.append (self._para (xchild))
      elif xchild.tag == 'screen':
        note.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        note.append (self._unorderedlist (xchild))

    return note

  def _notes (self, xparent, default):
    allxnotes = xparent.find ('notes')
    if allxnotes is None:
      return default

    notes = PkgDBNotes ()
    for xnotes in allxnotes:
      name = xnotes.get ('name', '')
      notes.append (name)

      for xchild in xnotes.iterchildren ():
        if xchild.tag == 'change':
          notes[name].append (self._change (xchild))
        elif xchild.tag == 'note':
          notes[name].append (self._note (xchild))
        elif xchild.tag == 'platform':
          notes[name].append (self._platform (xchild))

    return notes

  def _orderedlist (self, xorderedlist):
    start_num = xorderedlist.get ('start-num', '1')
    type = xorderedlist.get ('type', 'decimal')

    orderedlist = PkgDBNoteList ('orderedlist', type, start_num)

    for xitem in xorderedlist.findall ('item'):
      orderedlist.append (self._item (xitem))

    return orderedlist

  def _para (self, xpara):
    para = PkgDBNotePara ()
    para.align = xpara.get ('align', 'left')
    para.skip = int (xpara.get ('skip', '1'))

    # First add the element's text to para
    text = xpara.text
    if text is not None and len (text) > 0:
      para.append (PkgDBNoteItemText (text))

    # Then, the remaining children, in order, followed by their tails
    for xchild in xpara.iterchildren ():
      if xchild.tag == 'command':
        para.append (self._command (xchild))
      elif xchild.tag == 'ulink':
        para.append (self._ulink (xchild))

      text = xchild.tail
      if text is not None and len (text) > 0:
        para.append (PkgDBNoteItemText (text))

    return para

  def _package_managers (self, xparent):
    for xmanager in xparent.findall ('package-manager'):
      pkg_type = xmanager.get ('name', None)

      if self.pkg_type_to_save == pkg_type:
        # datetime?
        self.pkg = {}
      
        self.install_name = self._install_name (xmanager, '')
        self.pkgname_base = self._pkgname_base (xmanager, '')
        self.version = self.package.get_version (self._version (xmanager, ''))
        self.revision = Version (self._revision (xmanager, ''))
        self.subpkg = self._subpkgs (xmanager, {})
        self.replace = self._replace (xmanager, {})
        self.changelog = self._changelog (xmanager, [])
        self.data = self._data (xmanager, {})

        # if package name same as another in our list, append
        # version-specific entry if applicable (if different
        # version or newer than the previous)
        for pkgdb_entry in self.pkgdbs:
          if pkgdb_entry.prog_name == self.prog_name:
            break
        else:
          pkgdb_entry = PkgDB (pkg_type)
          pkgdb_entry.prog_name = self.prog_name
          self.pkgdbs.append (pkgdb_entry)

        package = pkgsys.init (pkg_type, self.pkgmgr)

        # ignore duplicate entry
        pkgdb_ver = PkgDBVersion (self.prog_ver, self.revision)
        if pkgdb_entry.version.has_key (pkgdb_ver):
          if pkgdb_entry.version[pkgdb_ver].revision == self.revision:
            return

        pkgdb_entry_ver = pkgdb_entry.version[pkgdb_ver] = PkgDBEntry ()

        pkgdb_entry_ver.install_name = self.install_name
        pkgdb_entry_ver.pkgname_base = self.pkgname_base
        pkgdb_entry_ver.version = self.version
        pkgdb_entry_ver.revision = self.revision
        pkgdb_entry_ver.subpkg = self.subpkg
        pkgdb_entry_ver.data = self.data
        pkgdb_entry_ver.description = self.description
        pkgdb_entry_ver.changelog = self.changelog
        pkgdb_entry_ver.notes = self.notes
        pkgdb_entry_ver.licenses = self.licenses

        pkgdb_entry_ver.replace = {}
        if self.replace:
          for pkgname_base in self.replace.keys ():
            pkgdb_entry_ver.replace[pkgname_base] = \
                self.replace[pkgname_base]

      return

  def _pkgname_base (self, xparent, default):
    xpkgname = xparent.find ('pkgname-base')
    if xpkgname is None:
      return default

    return xpkgname.text

  def _platform (self, xplatform):
    platform = PkgDBNotePlatform (xplatform.get ('name'))

    for xchild in xplatform.iterchildren ():
      if xchild.tag == 'orderedlist':
        platform.append (self._orderedlist (xchild))
      elif xchild.tag == 'para':
        platform.append (self._para (xchild))
      elif xchild.tag == 'screen':
        platform.append (self._screen (xchild))
      elif xchild.tag == 'unorderedlist':
        platform.append (self._unorderedlist (xchild))

    return platform

  def _replace (self, xparent, default):
    allxreplace = xparent.findall ('replace')
    if allxreplace is None:
      return default

    replaces = {}
    for xreplace in allxreplace:
      replaces[xreplace.get ('pkgname-base')] = xreplace.get ('install-name')

    return replaces

  def _revision (self, xparent, default):
    xrevision = xparent.find ('revision')
    if xrevision is None:
      return default

    return xrevision.text

  def _screen (self, xscreen):
    return PkgDBNoteScreen (xscreen.text.strip ())

  def _subpkgs (self, xmanager, default):
    allxsubpkg = xmanager.findall ('subpkg')
    if allxsubpkg is None:
      return default

    subpkgs = {}
    for xsubpkg in allxsubpkg:
      subpkg_type = xsubpkg.get ('type')
      if not subpkgs.has_key (subpkg_type):
        subpkgs[subpkg_type] = {}
      if not subpkgs[subpkg_type].has_key ('names'):
        subpkgs[subpkg_type]['names'] = []
      subpkgs[subpkg_type]['depend'] = self._depends (xsubpkg, {})
      subpkgs[subpkg_type]['names'].append (xsubpkg.get ('pkgname-ext'))
      subpkgs[subpkg_type]['order'] = int (xsubpkg.get ('order', '0'))
      subpkgs[subpkg_type]['size'] = int (xsubpkg.get ('size', '0'))
      subpkgs[subpkg_type]['version'] = Version (self._version (xsubpkg, ''))
      subpkgs[subpkg_type]['revision'] = Version (self._revision (xsubpkg, ''))

    return subpkgs

  def _ulink (self, xulink):
    return PkgDBNoteItemLink (xulink.get ('url'),
                              xulink.text.replace ('\n', ' '))

  def _unorderedlist (self, xunorderedlist):
    unorderedlist = PkgDBNoteList ('unorderedlist')

    for xitem in xunorderedlist.findall ('item'):
      unorderedlist.append (self._item (xitem))

    return unorderedlist

  def _version (self, xparent, default):
    xversion = xparent.find ('version')
    if xversion is None:
      return default

    return xversion.text

class PkgDB:
  def __init__ (self, pkg_type):
    self.pkg_type = pkg_type
    self.prog_name = ''
    self.version = PkgDBEntryVersion ()

class PkgDBEntry:
  def __init__ (self):
    self.install_name = ''
    self.pkgname_base = ''
    self.version = Version (0)
    self.revision = Version (0)
    self.subpkg = {}
    self.data = {}
    self.description = ''
    self.replace = {}
    self.changelog = []
    self.notes = PkgDBNotes ()
    self.licenses = PkgDBLicenses ()

class PkgDBEntryVersion:
  def __init__ (self):
    self.versions = {}

  def __getitem__ (self, key):
    if type (key) != types.InstanceType:
      raise KeyError, key

    for pkgdb_ver in self.versions.keys ():
      if pkgdb_ver.prog_ver == key.prog_ver and \
      pkgdb_ver.pkg_rev == key.pkg_rev:
        return self.versions[pkgdb_ver]
    raise KeyError, (key.prog_ver, str (key.pkg_rev))

  def __setitem__ (self, key, value):
    self.versions[key] = value

  def has_key (self, key):
    for pkgdb_ver in self.versions.keys ():
      if pkgdb_ver.prog_ver == key.prog_ver and \
      pkgdb_ver.pkg_rev == key.pkg_rev:
        return 1
    return 0

  def keys (self):
    return self.versions.keys ()

class PkgDBTextGroups:
  def __init__ (self, type, items = None):
    self.type = type
    if items is None:
      self.items = []
    else:
      self.items = items

  def __getitem__ (self, key):
    if not type (key) in (IntType, SliceType, StringType, UnicodeType):
      raise TypeError, key

    if type (key) == IntType:
      return self.items[key]
    elif type (key) == SliceType:
      return self.items[key.start:key.stop]
    else:
      for item in self.items:
        if item.name == key:
          return item
      else:
        raise IndexError, key

  def __iter__ (self):
    return self.items.__iter__ ()

  def append (self, name = ''):
    self.items.append (PkgDBNotesEntry (name))

class PkgDBTextGroupItem:
  def __init__ (self, type, items = None):
    self.type = type
    if items is None:
      self.items = []
    else:
      self.items = items

  def __getitem__ (self, key):
    if not type (key) in (IntType, SliceType):
      raise TypeError, key

    if type (key) == IntType:
      if key + 1 > len (self.items):
        raise IndexError, key
      return self.items[key]
    else:
      return self.items[key.start:key.stop]

  def __iter__ (self):
    return self.items.__iter__ ()

  def append (self, item):
    self.items.append (item)

class PkgDBNote (PkgDBTextGroupItem):
  def __init__ (self, type):
    PkgDBTextGroupItem.__init__ (self, type, PkgDBNoteItems ())

class PkgDBNotePlatform (PkgDBTextGroupItem):
  def __init__ (self, name):
    PkgDBTextGroupItem.__init__(self, 'platform', PkgDBNoteItems ())
    self.name = name

class PkgDBChange (PkgDBTextGroupItem):
  def __init__ (self, date = '', revision = None, author = '', email = '',
                from_version = '', to_version = ''):
    PkgDBTextGroupItem.__init__(self, 'change', PkgDBNoteItems ())
    
    self.from_version = from_version	# changes between from_version
    self.to_version = to_version	# and to_version
    self.date = date			# date of change
    self.revision = revision		# revision of .sb change
    self.author = author		# author of change
    self.email = email			# email address of change author
    
class PkgDBNotes (PkgDBTextGroups):
  def __init__ (self):
    PkgDBTextGroups.__init__ (self, 'notes')

class PkgDBNotesEntry (PkgDBTextGroupItem):
  def __init__ (self, name):
    PkgDBTextGroupItem.__init__ (self, 'note')
    self.name = name

class PkgDBNoteItems (PkgDBTextGroupItem):
  def __init__ (self, group = None):
    PkgDBTextGroupItem.__init__ (self, 'item group')
    self.group = group

class PkgDBNoteItem (PkgDBTextGroupItem):
  def __init__ (self):
    PkgDBTextGroupItem.__init__ (self, 'item')

class PkgDBNoteItemGroup (PkgDBTextGroupItem):
  def __init__ (self, name):
    PkgDBTextGroupItem.__init__ (self, 'item group')
    self.group = name

class PkgDBNoteIndent (PkgDBTextGroupItem):
  def __init__ (self):
    PkgDBTextGroupItem.__init__ (self, 'indent')

class PkgDBNotePara (PkgDBTextGroupItem):
  def __init__ (self):
    PkgDBTextGroupItem.__init__ (self, 'para')
    self.align = 'left'
    self.skip = 1

class PkgDBNoteList (PkgDBTextGroupItem):
  def __init__ (self, type, list_type = None, start_num = None):
    PkgDBTextGroupItem.__init__ (self, type)
    self.list_type = list_type
    self.start_num = start_num

class PkgDBNoteScreen:
  def __init__ (self, text):
    self.type = 'screen'
    self.text = text

class PkgDBNoteItemText:
  def __init__ (self, text):
    self.type = 'text'
    self.text = text

class PkgDBNoteItemCommand:
  def __init__ (self, text):
    self.type = 'command'
    self.text = text

class PkgDBNoteItemLink:
  def __init__ (self, url, value):
    self.type = 'link'
    self.url = url
    self.value = value

class PkgDBVersion:
  def __init__ (self, prog_ver, pkg_rev):
    self.prog_ver = prog_ver
    self.pkg_rev = pkg_rev

class PkgDBLicenses:
  def __init__ (self):
    self.items = []

  def __getitem__ (self, key):
    return self.items[key]

  def __setitem__ (self, key, value):
    self.items[key] = value

  def __len__ (self):
    return len (self.items)

  def append (self, item):
    self.items.append (item)

class PkgDBLicense (PkgDBTextGroupItem):
  def __init__ (self, module = '', path = ''):
    PkgDBTextGroupItem.__init__ (self, 'license')
    self.module = module
    self.path = path
