# Copyright (C) 2001-2004, 2010, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import pwd
import sys
import zipfile

import misclib
import pkginst.archive

class PkginstError (Exception):
  """base class for all exceptions raised by this module"""

  def __init__ (self, arg = None, retry = 0):
    self.msg = arg

  def __str__ (self):
    return self.msg

class PkginstArchiveError (PkginstError):
  """error processing archive or with files in archive"""

  def __init__ (self, msg):
    self.msg = msg

class PkginstArchiveMemberDoesNotExist (PkginstError):
  """member does not exist in archive"""

class PkginstExtractError (PkginstError):
  """error unpacking archive"""

  def __init__ (self, msg):
    self.msg = msg + '\n'

class PkginstDiskSpaceError (PkginstError):
  """out of disk space while unpacking archive"""

  def __init__ (self, msg):
    self.msg = msg + '\n'

class PkginstPackageNotFoundError (PkginstError):
  """package not found"""

class PkginstArchiveExtractError (PkginstError):
  """error extracting archive members"""

  def __init__ (self, archive, file):
    self.archive = archive
    self.file = file
    self.msg = 'problems extracting ' + file + ' from ' + archive

class PkginstArchiveFileNotFoundError (PkginstError):
  """file not found in archive"""

  def __init__ (self, archive, file):
    self.archive = archive
    self.file = file
    self.msg = file + ' not found in archive ' + archive

class PkginstArchiveProgNotFoundError (PkginstError):
  """program not found to process archive"""

class PkginstMD5ChecksumNotFoundError (PkginstError):
  """no MD5 checksum file for payload"""

class PkginstPackageNotFound (PkginstError):
  """package not found"""

class PkginstGPGSignatureNotFoundError (PkginstError):
  """no GPG signature file for payload"""

class PkginstGPGProgNotFoundError (PkginstError):
  """gpg program not found"""

class Pkginst:
  def __init__ (self, prog_name, prog_ver, pkg_type, path, tmpbasedir):
    self.prog_name = prog_name
    self.prog_ver = prog_ver
    self.pkg_type = pkg_type
    self.path = path
    self.pkgdir = None
    self.tmpbasedir = tmpbasedir

  def extract (self, local_depot_is_temporary, local_depot, callback):
    if not os.path.exists (self.path):
      raise PkginstPackageNotFoundError

    # temporary directory for extracting package files
    if not self.pkgdir:
      self.pkgdir = misclib.mkstempdir (self.tmpbasedir)

    archive = pkginst.archive.init (self.path, self.pkgdir)
    payload = archive.payload_file
    apply (callback, ['begin', 'extracting ' + payload + ' file'])
    if not os.path.exists (os.path.join (self.pkgdir, payload)):
      if not archive.extract (payload, local_depot_is_temporary,
                              local_depot,
                              os.path.join (self.pkgdir, payload)):
        misclib.rmdir (self.pkgdir)
        self.pkgdir = None
        raise PkginstArchiveExtractError (self.path, payload)
    apply (callback, ['end', 'done'])

    # test integrity of payload.
    apply (callback, ['begin', 'testing ' + payload])
    try:
      archive.verify_payload ()
    except PkginstExtractError, msg:
      raise
    apply (callback, ['end', 'done'])

    # extract archive
    apply (callback, ['begin', 'unpacking ' + payload + ' contents'])
    apply (callback, ['end', ''])
    archive.unpack_payload (local_depot_is_temporary, local_depot)
    return self.pkgdir

  # check MD5 checksum of payload
  def verify_md5_checksum (self, local_depot_is_temporary,
                           local_depot):
    if not os.path.exists (self.path):
      raise PkginstPackageNotFoundError

    # temporary directory for extracting package files
    if not self.pkgdir:
      self.pkgdir = misclib.mkstempdir (self.tmpbasedir)

    # make sure payload file exists in archive
    archive = pkginst.archive.init (self.path, self.pkgdir)
    payload = archive.payload_file
    if not archive.verify_md5_checksum (payload, local_depot_is_temporary,
                                        local_depot):
      misclib.rmdir (self.pkgdir)
      self.pkgdir = None

      return 0
    else:
      return 1

  # check GPG signature of payload
  def verify_gpg_sig (self, keyring_path, local_depot_is_temporary,
                      local_depot):
    if not os.path.exists (self.path):
      raise PkginstPackageNotFoundError

    # temporary directory for extracting package files
    if not self.pkgdir:
      self.pkgdir = misclib.mkstempdir (self.tmpbasedir)

    # make sure payload file exists in archive
    archive = pkginst.archive.init (self.path, self.pkgdir)
    payload = archive.payload_file
    if not archive.exists (payload):
      misclib.rmdir (self.pkgdir)
      self.pkgdir = None

      raise PkginstArchiveFileNotFoundError (self.path, payload)

    if not archive.exists (payload + '.gpg'):
      misclib.rmdir (self.pkgdir)
      self.pkgdir = None

      raise PkginstGPGSignatureNotFoundError

    # verify signature
    if not os.path.exists (os.path.join (self.pkgdir, payload)):
      if not archive.extract (payload, local_depot_is_temporary,
                              local_depot,
                              os.path.join (self.pkgdir, payload)):
        misclib.rmdir (self.pkgdir)
        self.pkgdir = None
        raise PkginstArchiveExtractError (self.path, payload)

    if not archive.extract (payload + '.gpg', local_depot_is_temporary,
                            local_depot,
                            os.path.join (self.pkgdir, payload + '.gpg')):
      misclib.rmdir (self.pkgdir)
      self.pkgdir = None
      raise PkginstArchiveExtractError (self.path, payload + '.gpg')

    try:
      if misclib.verify_gpg_sig (os.path.join (self.pkgdir, payload),
      os.path.join (self.pkgdir, payload + '.gpg'), keyring_path):
        return True
    except misclib.MisclibGPGProgNotFoundError:
      # remove package due to checksum failure
      os.remove (os.path.join (self.pkgdir, payload))
      os.remove (os.path.join (self.pkgdir, payload + '.gpg'))

      raise PkginstGPGProgNotFoundError

    # remove package due to checksum failure
    os.remove (os.path.join (self.pkgdir, payload))
    os.remove (os.path.join (self.pkgdir, payload + '.gpg'))

    return False

  def cleanup (self):
    if not self.pkgdir:
      return

    if self.pkgdir != self.path and \
    os.path.exists (self.pkgdir) and \
    os.path.isdir (self.pkgdir):
      misclib.rmdir (self.pkgdir)
    self.pkgdir = None
