# Copyright (C) 2002-2003, 2006, 2010 The Written Word, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import base64
import cStringIO
import os
import sys

from xml.sax import saxexts, saxlib

from misclib import md5sum
from pkginst.archive import Archive
from pkginst.archive.zip import ArchiveZip
from pkginst.archive.gcpio import ArchiveGcpio

def normalize_whitespace (text):
  "Remove redundant whitespace from a string"
  return (string.join (text.split (), ' '))

class read_pkg_db (saxlib.HandlerBase):
  def __init__ (self, data, extract = 0, output_file = None):
    self.in_data = 0
    self.data = data
    self.extract = extract		# whether or not to extract payload
    self.output_file = output_file	# output filename

  def startElement (self, name, attrs):
    if name == 'data':
      self.in_data = 1
      self.data_checksum = attrs.get ('checksum', None)
      self.data_checksum_type = attrs.get ('checksum-type', None)
      self.data_encoding = attrs.get ('encoding', None)
      self.data_filename = attrs.get ('filename', None)
      self.data_size = attrs.get ('size', None)
      self.payload = cStringIO.StringIO ()
      self.payload_size = 0
      if self.output_file:
        self.output = open (self.output_file, 'w')

  def characters (self, ch, start, length):
    if self.in_data and self.extract:
      self.payload.write (ch)
      self.payload_size += len (ch)

      # flush buffer every 77*128 lines
      if self.payload_size % (77*1024*128) == 0:
        self.output.write (base64.decodestring (self.payload.getvalue ()))
        self.payload.close ()
        self.payload = cStringIO.StringIO ()
        self.payload_size = 0

  def endElement (self, name):
    if name == 'data':
      self.data['checksum'] = self.data_checksum
      self.data['checksum type'] = self.data_checksum_type
      self.data['encoding'] = self.data_encoding
      self.data['filename'] = self.data_filename
      self.data['size'] = self.data_size
      self.in_data = 0

      # last incomplete line from payload data
      if self.payload_size:
        self.output.write (base64.decodestring (self.payload.getvalue ()))
      self.payload.close ()
      if self.output_file:
        self.output.close ()

class ArchiveXML (Archive):
  def __init__ (self, path, pkgdir):
    Archive.__init__ (self, path, pkgdir)
    self.data = {}

    # get payload filename
    fh = open (self.path, 'r')
    xmlh = read_pkg_db (self.data)
    p = saxexts.make_parser ()
    p.setDocumentHandler (xmlh)
    p.parseFile (fh)
    fh.close ()
    p.close ()

    self.payload_file = self.data['filename']

  def exists (self, member):
    fh = open (self.path, 'r')
    xmlh = read_pkg_db (self.data)
    p = saxexts.make_parser ()
    p.setDocumentHandler (xmlh)
    p.parseFile (fh)
    fh.close ()
    p.close ()

    if self.data['filename'] != member:
      return 0
    return 1

  def extract (self, member):
    # extract to package directory
    pwd = os.getcwd ()
    os.chdir (self.pkgdir)

    if not os.path.exists (member):
      fh = open (self.path, 'r')
      xmlh = read_pkg_db (self.data, extract = 1, output_file = member)
      p = saxexts.make_parser ()
      p.setDocumentHandler (xmlh)
      p.parseFile (fh)
      fh.close ()
      p.close ()

      if self.data['filename'] != member:
        os.remove (member)
        os.chdir (pwd)
        return 0

    os.chdir (pwd)
    return 1

  def verify_md5_sig (self, member):
    if not self.extract (member):
      return 0

    # calculate
    if self.data['checksum type'] == 'md5':
      member_cksum = md5sum (os.path.join (self.pkgdir, member))
    if member_cksum == self.data['checksum']:
      return 1

    # remove file as checksum didn't match
    os.remove (os.path.join (self.pkgdir, member))

    return 0

  def unpack_payload (self, local_depot_is_temporary, local_depot):
    if self.payload_file[-4:] == '.zip':
      archive = ArchiveZip (self.path, self.pkgdir)
    elif self.payload_file[-10:] == '.gcpio.bz2':
      archive = ArchiveGcpio (self.path, self.pkgdir)
    archive.unpack_payload (local_depot_is_temporary, local_depot)

  def verify_payload (self):
    if self.payload_file[-4:] == '.zip':
      archive = ArchiveZip (self.path, self.pkgdir)
    elif self.payload_file[-10:] == '.gcpio.bz2':
      archive = ArchiveGcpio (self.path, self.pkgdir)
    archive.verify_payload ()
