# Copyright (C) 2002-2003, 2006, 2009-2010, 2015 The Written Word, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import binascii
import os
import statvfs
import struct
import sys
import zipfile
import zlib

import misclib
import pkginst

from stat import ST_SIZE
from zipfile import BadZipfile, ZipInfo

from libarchive import _libarchive, SeekableArchive, is_archive
from libarchive.zip import ZipFile

from misclib import dir_size
from pkginst import PkginstArchiveError, \
                    PkginstArchiveMemberDoesNotExist, \
                    PkginstDiskSpaceError, PkginstExtractError
from pkginst.archive import Archive
from progress_meter import TextProgressMeter

class ProgressAdaptor (object):
  """Adaptor to allow use of `TextProgressMeter` with pylibarchive.

  pylibarchive's write progress callback will repeatedly call the
  named function between write chunks
  """

  def __init__ (self, progressmeter, zip, total_bytes):
    self.progressmeter = progressmeter
    self.total_bytes = total_bytes
    self.written_bytes = 0
    self.zip = zip

  def progress_cb (self, zip, additional_bytes):
    if self.written_bytes == 0:
      self.progressmeter.start (self.total_bytes)

    self.written_bytes += additional_bytes
    self.progressmeter.update (self.written_bytes)

class PkgZipFile (ZipFile):
  def __init__ (self, path, mode):
    self.entries = []

    if not is_archive (path, formats=('zip',)):
      raise BadZipfile

    super (PkgZipFile, self).__init__ (path, mode)

  def extract (self, name, path = None):
    if not path:
      path = os.getcwd()

    fh = open (os.path.join(path, name), 'w')
    return self.readpath (name, fh)
    fh.close ()

  def namelist (self):
    if not self.entries:
      for entry in self:
        self.entries.append (entry.pathname)

    return self.entries

  def testzip (self):
    """ Read all the files and check the CRC. """
    chunk_size = 2 ** 20

    for member in self.entries:
      with self.open (member, 'r') as fh:
        while self.read (member, chunk_size):
          pass
    else:
      return member

class RawZipEntry (object):
  def __init__ (self):
    self.compress_size = 0
    self.external_attr = 0
    self.file_size = 0
    self.filename = ''
    self.filetype = 0
    self.mode = 0
    self.perm = 0

    self._header_position = 0

  @classmethod
  def from_archive (cls):
    e = _libarchive.archive_entry_new ()

    ret = _libarchive.archive_read_next_header2 (self.fh, e)
    if ret == _libarchive.ARCHIVE_EOF:
      return
    if ret != _libarchive.ARCHIVE_OK:
      raise BadZipfile

    zipi = RawZipEntry ()
    zipi.file_size = _libarchive.archive_entry_size (e)
    zipi.filetype = _libarchive.archive_entry_filetype (e)
    zipi.mode = _libarchive.archive_entry_mode (e) & 0xFFFF
    zipi.perm = _libarchive.archive_entry_perm (e)
    zipi.filename = _libarchive.archive_entry_pathname(e).decode('utf-8')
    zipi._header_position = _libarchive.archive_read_header_position (
                              self.fh)

    zipi.external_attr = zipi.mode << 16L;

    self.entries.append (zipi)

    _libarchive.archive_entry_free (e)

    return zipi

class RawZipFile ():
  def __init__ (self, path, mode):
    self.entries = []
    self.mode = mode
    self.path = path
    self.fh = None
    self.progress_cb = None
    self.progress_data = None

    self._init ()

    self._populate_entries ()

  def __del__ (self):
    self.close ()

  def __exit__ (self):
    self.close ()

  def close (self):
    if self.fh != None:
      _libarchive.archive_read_close (self.fh)
      _libarchive.archive_read_free (self.fh)

      self.fh = None

  def extract (self, member, path = None):
    if not path:
      path = os.getcwd()

    self._seek (self.getinfo (member))

    fh = open (os.path.join (path, member), 'w')
    ret = _libarchive.archive_read_data_into_fd (self.fh, fh.fileno ())
    fh.close ()

    return ret

  def getinfo (self, member):
    for zipi in self.entries:
      if zipi.filename == member:
        return zipi

  def namelist (self):
    fl = []
    for zipi in self.entries:
      fl.append (zipi.filename)

    return fl

  def readpath (self, member, fh):
    self._seek (self.getinfo (member))

    ret = _libarchive.archive_read_data_into_fd (self.fh, fh.fileno ())

    return ret

  def set_write_progress_callback (self, cb, user_data):
    self.progress_cb = cb
    self.progress_data = user_data

    _libarchive.archive_write_data_set_progress_callback (self.fh,
      cb, user_data)

  def _get_header_position (self):
    return _libarchive.archive_read_header_position (self.fh)

  def _init (self):
    self.fh = _libarchive.archive_read_new ()

    _libarchive.archive_read_support_format_zip (self.fh)

    # Ensure file is a zip archive
    ret = _libarchive.archive_read_open_filename (
            self.fh, self.path, 10240)
    if ret == _libarchive.ARCHIVE_FATAL:
      raise BadZipfile

  def _populate_entries (self):
    while True:
      e = _libarchive.archive_entry_new ()

      ret = _libarchive.archive_read_next_header2 (self.fh, e)
      if ret == _libarchive.ARCHIVE_EOF:
        break
      if ret != _libarchive.ARCHIVE_OK:
        raise BadZipfile

      zipi = RawZipEntry ()
      zipi.file_size = _libarchive.archive_entry_size (e)
      zipi.filetype = _libarchive.archive_entry_filetype (e)
      zipi.mode = _libarchive.archive_entry_mode (e) & 0xFFFF
      zipi.perm = _libarchive.archive_entry_perm (e)
      zipi.filename = _libarchive.archive_entry_pathname(e).decode('utf-8')
      zipi._header_position = _libarchive.archive_read_header_position (
                                self.fh)

      zipi.external_attr = zipi.mode << 16L;

      self.entries.append (zipi)

      _libarchive.archive_entry_free (e)

  def _reopen (self):
    self.close ()
    self._init ()

    if self.progress_cb:
      self.set_write_progress_callback (self.progress_cb,
                                        self.progress_data)

  def _seek (self, member):
    move = member._header_position - self._get_header_position ()
    if move != 0:
      if move < 0:
        # can't move back, re-open archive:
        self._reopen()

      # move to proper position in stream
      while True:
        e = _libarchive.archive_entry_new ()

        ret = _libarchive.archive_read_next_header2 (self.fh, e)
        if ret == _libarchive.ARCHIVE_EOF:
          break
        if ret != _libarchive.ARCHIVE_OK:
          raise BadZipfile

        filename = _libarchive.archive_entry_pathname(e).decode('utf-8')

        _libarchive.archive_entry_free (e)

        if filename == member.filename:
          break

# get computer size representation of integer (B/K/M)
def comp_size (size):
  string = ''

  if size < 1024:
    string = str (size) + 'B'
  elif size < 1024 * 1024:
    string = str (int (size / 1024)) + 'K'
  else:
    string = str (int (size / (1024 * 1024))) + 'M'

  return string

class ArchiveZip (Archive):
  def __init__ (self, path, pkgdir):
    Archive.__init__ (self, path, pkgdir)
    self.payload_file = 'data.zip'

  def exists (self, member):
    try:
      zip = RawZipFile (self.path, 'r')
    except BadZipfile:
      raise PkginstArchiveError ('Errors encountered while \
verifying if ' + member + ' exists in ' + self.path + '. It is not \
a valid zip archive.')

    files = zip.namelist ()
    zip.close ()
    if not member in files:
      return False
    return True

  def extract (self, member, local_depot_is_temporary,
               local_depot, dest = None):
    if not self.exists (member):
      raise PkginstArchiveMemberDoesNotExist \
        ('archive member "' + member +
         '" does not exist in archive file ' + self.path + '.')

    try:
      zip = RawZipFile (self.path, 'r')
    except BadZipfile:
      raise PkginstArchiveError ('Errors encountered while \
extracting ' + member + ' from ' + self.path + '. It is not a \
valid zip archive.')

    if not dest:
      dest = os.path.join (os.getcwd (), member)

    fs = os.statvfs (os.path.dirname (dest))
    fs_free = long (fs[statvfs.F_BFREE]) * long (fs[statvfs.F_FRSIZE])
    member_size = zip.getinfo (member).file_size

    # make sure enough space on destination filesystem
    if member_size > fs_free:
      member_size_str = comp_size (member_size)
      fs_free_str = comp_size (fs_free)

      if local_depot_is_temporary:
        local_depot_size = dir_size (local_depot)
        local_depot_str = comp_size (local_depot_size)

        raise PkginstDiskSpaceError \
          ('not enough space to write ' + member + ' to ' +
           os.getcwd () + '. ' + member_size_str + ' (' +
           str (member_size) + ' bytes) needed but only ' +
           fs_free_str + ' (' + str (fs_free) +
           ' bytes) currently available (' +
           comp_size (fs_free + local_depot_size) +
           ' is really available but the temporary local depot, ' +
           local_depot + ', created because you did not specify \
--local-depot or set the local_depot variable in the config file, is ' +
           local_depot_str + ').')
      else:
        raise PkginstDiskSpaceError \
          ('not enough space to write ' + member + ' to ' +
           os.getcwd () + '. ' + member_size_str + ' (' +
           str (member_size) + ' bytes) needed but only ' +
           fs_free_str + ' (' + str (fs_free) + ' bytes) available.')

    fh = open (dest, 'w')
    try:
      zip.readpath (member, fh)
    except (IOError, OSError), msg:
      fh.close ()
      zip.close ()
      os.remove (member)

      raise PkginstExtractError (str (msg) + ' encountered \
while unpacking ' + member + ' from ' + self.path)

    fh.close ()
    zip.close ()

    return 1

  def open (self, member, mode):
    fh, filename = misclib.mkstempfile ()
    fh.close ()

    self.extract (member, False, None, filename)
    fh = open (filename, mode)
    os.remove (filename)

    return fh

  def unpack_payload (self, local_depot_is_temporary, local_depot):
    # extract to package directory
    pwd = os.getcwd ()
    os.chdir (self.pkgdir)

    try:
      zip = RawZipFile (self.payload_file, 'r')
    except BadZipfile:
      os.chdir (pwd)
      raise PkginstArchiveError ('Errors encountered while \
unpacking contents of ' + self.payload_file + '. It is not a valid \
zip archive.')

    # space of all members in archive
    zip_members = zip.namelist ()
    zip_members.sort ()
    zip_members_size = 0
    for zip_member in zip_members:
      zip_members_size += zip.getinfo (zip_member).file_size
    zip_members_size_str = comp_size (zip_members_size)

    fs = os.statvfs (os.getcwd ())
    fs_free = long (fs[statvfs.F_BFREE]) * long (fs[statvfs.F_FRSIZE])

    # make sure enough space on destination filesystem
    if zip_members_size > fs_free:
      fs_free_str = comp_size (fs_free)

      payload_size = os.lstat (self.payload_file)[ST_SIZE]
      payload_size_str = comp_size (payload_size)

      os.chdir (pwd)
      if local_depot_is_temporary:
        local_depot_size = dir_size (local_depot)
        local_depot_str = comp_size (local_depot_size)

        raise PkginstDiskSpaceError \
          ('not enough space to unpack ' + self.payload_file +
           ' payload to ' + pwd + '. ' + zip_members_size_str +
           ' (' + str (zip_members_size) + ' bytes) needed but only ' +
           fs_free_str + ' (' + str (fs_free) +
           ' bytes) currently available (' +
           comp_size (fs_free + local_depot_size + payload_size) +
           ' is really available but the temporary local depot, ' +
           local_depot + ', created because you did not specify \
--local-depot or set the local_depot variable in the config file, is ' +
           local_depot_str + ' and the ' + self.payload_file +
           ' payload file is ' + comp_size (payload_size) + ').')
      else:
        raise PkginstDiskSpaceError \
          ('not enough space to unpack ' + self.payload_file +
           ' payload to ' + pwd + '. ' + zip_members_size_str +
           ' (' + str (zip_members_size) + ' bytes) needed but only ' +
           fs_free_str + ' (' + str (fs_free) +
           ' bytes) currently available (' +
           comp_size (fs_free + payload_size) +
           ' is really available but the ' + self.payload_file +
           ' payload file is ' + comp_size (payload_size) + ').')

    adaptor = ProgressAdaptor (TextProgressMeter (6), zip,
                               zip_members_size)

    zip.set_write_progress_callback (adaptor.progress_cb, adaptor)

    for zip_member in zip_members:
      perm = (zip.getinfo (zip_member).external_attr >> 16L) & 0x1FF
      if zip_member[-1] == '/':
        os.mkdir (zip_member)
      else:
        try:
          zip.extract (zip_member)
        except (IOError, OSError), msg:
          zip.close ()
          if os.path.exists (zip_member):
            os.remove (zip_member)
          os.chdir (pwd)

          raise PkginstExtractError \
            (str (msg) + ' encountered while unpacking ' +
             self.payload_file + ' payload')

      os.chmod (zip_member, perm)

    adaptor.progressmeter.end ()

    zip.close ()
    os.chdir (pwd)

  def verify_payload (self):
    # extract to package directory
    pwd = os.getcwd ()
    os.chdir (self.pkgdir)

    try:
      zip = zipfile.ZipFile (self.payload_file, 'r')
    except BadZipfile:
      os.chdir (pwd)
      raise PkginstArchiveError ('Errors encountered while \
verifying contents of ' + self.payload_file + '. It is not a valid \
zip archive.')

    badfile = zip.testzip ()
    if badfile:
      zip.close ()
      os.chdir (pwd)
      raise PkginstExtractError ('Error encountered while \
unpacking ' + self.payload_file + ' payload. First file with problems \
is ' + badfile + '.')

    zip.close ()
    os.chdir (pwd)
