# Copyright (C) 2001-2003, 2006, 2008, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import re
import string
import sys

from types import IntType, SliceType

from version import Version

# cache of installed packages
pkgcache = None

# exceptions
class PkgsysError (Exception):
  """Base class for all exceptions raised by this module"""

  def __init__ (self, arg = None):
    self.msg = arg

  def __str__ (self):
    return self.msg

class PkgsysProgNotFound (PkgsysError):
  """package manager program not found"""

  def __init__ (self, prog, config_var):
    self.msg = 'Could not find "' + prog + '" in the search path. \
Either rerun and add the path where "' + prog + '" is located to \
$PATH or set the ' + config_var + ' variable in the config file.'

class PkgsysNeedRootError (PkgsysError):
  """package management systems wants user running as root"""

  def __init__ (self):
    self.msg = 'need root access to determine if packages installed'

class PkgsysNoSubpkgsError (PkgsysError):
  """package has no subpackages"""

class PkgsysPackageNotFoundError (PkgsysError):
  """package not found"""

  def __init__ (self, pkgnames = []):
    self.pkgnames = pkgnames

class PkgsysInvalidSubpkgTypeError (PkgsysError):
  """invalid sub-package type"""

  def __init__ (self, pkgname_base, subpkg_type):
    self.pkgname_base = pkgname_base
    self.subpkg_type = subpkg_type

class PkgsysExtractError (PkgsysError):
  """error copying raw package file"""

class PkgsysExtractDuplicate (PkgsysError):
  """duplicate file in destination directory"""

  def __init__ (self, pkgname):
    self.msg = pkgname

class Pkglist:
  def __init__ (self, entries = None):
    if not entries:
      self.entries = []
    else:
      self.entries = entries

  def __add__ (self, item):
    return Pkglist (self.entries + item)

  def __contains__ (self, item):
    for entry in self.entries:
      if item.name == entry.name and \
      item.pkg_ver == entry.pkg_ver:
        return True

    return False

  def __delitem__ (self, key):
    if type (key) != IntType:
      raise TypeError, key

    if type (key) == IntType:
      del self.entries[key]

  def __getitem__ (self, key):
    if type (key) != IntType and type (key) != SliceType:
      raise TypeError, key

    if type (key) == IntType:
      return self.entries[key]
    elif type (key) == SliceType:
      return Pkglist (self.entries[key])

  def __iter__ (self):
    self.entry_index = 0

    return iter (self.next, None)

  def __len__ (self):
    return len (self.entries)

  def __nonzero__ (self):
    if self.entries:
      return True

    return False

  def __str__ (self):
    return str (self.entries)

  def append (self, value):
    self.entries.append (value)

  def insert (self, pos, value):
    self.entries.insert (pos, value)

  def next (self):
    if len (self.entries) == 0 or \
    len (self.entries) == self.entry_index:
      return None

    self.entry_index += 1

    return self.entries[self.entry_index - 1]

  def pop (self, pos = 0):
    return self.entries.pop (pos)

  def remove (self, item):
    self.entries.remove (item)

  def sort (self, cmp = None, key = None, reverse = False):
    def _sort (x, y):
      if (x.pkg_ver < y.pkg_ver):
        return -1

      if (x.pkg_ver == y.pkg_ver) and (x.pkg_rev < y.pkg_rev):
        return -1
      if (x.pkg_ver == y.pkg_ver) and (x.pkg_rev == y.pkg_rev):
        return 0
      if (x.pkg_ver == y.pkg_ver) and (x.pkg_rev > y.pkg_rev):
        return 1

      if (x.pkg_ver > y.pkg_ver):
        return 1

    if (cmp):
      self.entries.sort (cmp, key, reverse)
    else:
      self.entries.sort (_sort, key, reverse)

class Pkgmgr:
  def __init__ (self):
    self.args = []		# arguments to package manager
    self.args_rm = []		# arguments to package manager when removing
				# (for pkg-inst)
    self.ignoredeps = None	# ignore dependencies
    self.npkg_arch = ''		# GNU system name
    self.npkg_db = None		# Npkg package database directory
    self.npkgadd_path = None	# path to npkgadd command
    self.npkginfo_path = None	# path to npkgadd command
    self.npkgrm_path = None	# path to npkgadd command
    self.pkgname_prefix = ''	# prefix for all package names
    self.reinstall = False	# reinstall package
    self.root = ''		# alternate root directory
    self.rpm_arch = ''		# RPM architecture type
    self.rpm_db = None		# RPM package database directory
    self.rpm_path = None	# path to rpm command
    self.systype = None		# GNU system name
    self.verbose = 0		# Enable verbosity?

class Pkgsys:
  def __init__ (self, output = None):
    self.attribs = None
    self.archive = None		# archive file containing package
    self.output = output
    self.pkgdb_entry = None	# entry in pkg-db.xml file
    self.pkgdb_entry_ver = None	# entry in pkg-db.xml file for this version
    self.name = None		# program name
    self.pkgname_base = None	# package manager base name
    self.prog_ver = None	# program version
    self.prog_rev = None	# program revision
    self.pkg_ver = Version (0)	# package version known by package manager
    self.pkg_rev = Version (0)	# package revision known by package manager
    self.pkg_type = ''		# package manager system
    self.is_dependency = False	# if package added to meet a dependency
    self.subpkgs = []		# subpackage types to install
    self.installed_subpkgs = [] # subpackage types already installed
    self.depot = None
    self.pkginst = None		# Pkginst instance for .pkg-inst files
    self.install_base = None	# installation base
    self.install_path = None	# installation path

  def __eq__ (self, item):
    if item:
      if self.name == item.name and \
      self.prog_ver == item.prog_ver:
        return True

    return False

  def get_install_path (self, install_name):
    return re.sub ('\${INSTALL_NAME}', install_name, self.install_path)

  # sort subpackages based on "order" attribute
  def sort_subpkgs (self, subpkg1, subpkg2):
    return (cmp (self.pkgdb_entry_ver.subpkg[subpkg2]['order'],
                 self.pkgdb_entry_ver.subpkg[subpkg1]['order']))

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type, list):
    if not list.has_key (subpkg_type):
      raise PkgsysInvalidSubpkgTypeError (self.pkgname_base, subpkg_type)

    pkgnames = []
    for ext in list[subpkg_type]:
      # add only those extensions for which we have a package for
      if self.pkgdb_entry_ver and \
      self.pkgdb_entry_ver.subpkg.has_key (subpkg_type) and \
      not ext in self.pkgdb_entry_ver.subpkg[subpkg_type]['names']:
        continue
      pkgnames.append (self.pkgname_base + ext)
    return pkgnames

  # get pkg-db.xml dependency string
  def get_pkgdb_depend_str (self, ver_tests):
    # add r>=1 for any test without a revision.
    for ver_test in string.split (ver_tests, ' '):
      var = re.sub ('^(\w)([>=<]+)(\d+)', '\\1', ver_test)
      if var == 'r':
        break
    else:
      return re.sub ('>=', '&gt;=', ver_tests) + ' r&gt;=1'

    return re.sub ('>=', '&gt;=', ver_tests)

  # make version from package version/revision specific to
  # package-management system
  def make_version (self, version, revision):
    return version
  def make_revision (self, version, revision):
    return revision

  # get package version specific to package-management system
  def get_version (self, version):
    return Version (version)

  # get package revision specific to package-management system
  def get_revision (self, version, revision = None):
    return Version (revision)

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, list):
    if not list.has_key (subpkg_type):
      return None
    return list[subpkg_type]

  # determine subpackage type depending on package name
  def get_subpkg_type (self, package_name, list):            
    for subpkg in list:
      if package_name.find (subpkg[0], len (package_name) -
                            len (subpkg[0]), len (package_name)) != -1:
        return subpkg[1]
    return None

  # package version
  def version_str (self, pkg_ver, pkg_rev):
    return pkg_ver + '-' + pkg_rev

def init (pkg_type, pkgmgr, output = None, cache = 0):
  exec 'import pkgsys.%s' % (pkg_type)
  exec 'package = pkgsys.%s.%s (pkgmgr, output, cache)' % \
       (pkg_type, pkg_type)

  return package

# string containing arg values separated by space
def arg_string (args):
  arg_str = ''

  args_to_print = args[:]
  while args_to_print:
    arg = args_to_print.pop (0)

    if arg.find (' ') != -1:
      arg_str += '"' + arg + '"'
    else:
      arg_str += arg

    if args_to_print:
      arg_str += ' '

  return arg_str
