# Copyright (C) 2001-2003, 2006, 2011, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import shutil
import string
import sys

import pkgsys

from pkgcache import PkgCacheEntry, PkgCache
from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# cache of installed packages
pkgcache = None

class depot (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = False):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'depot'

    # initialize package cache
    if not pkgcache or (cache and not pkgcache.cache):
      pkgcache = MyPkgCache (pkgmgr.pkgname_prefix, cache)

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []
    pkg_depot = ''

    # package either in depot directory or depot file
    pkg_depot = None
    if os.path.exists (os.path.join (self.pkgdir,
                                     re.sub ('\.', '/', pkgname))):
      pkg_depot = self.pkgdir
    elif os.path.exists (os.path.join (self.pkgdir,
                                       self.pkgname_base) + '.depot'):
      pkg_depot = os.path.join (self.pkgdir, self.pkgname_base) + \
                  '.depot'
      if not pkg_depot:
        raise pkgsys.PkgsysPackageNotFoundError ([re.sub ('\.', '/', pkgname),
                                                  self.pkgname_base + '.depot'])

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps:
      if not re.search ('enforce_dependencies=',
                        string.join (args, ' ')):
        args.append ('-xenforce_dependencies=false')
      if not re.search ('autoselect_dependencies=',
                        string.join (args, ' ')):
        args.append ('-xautoselect_dependencies=false')
    if pkgmgr.reinstall and not re.search ('reinstall=',
                                              string.join (args, ' ')):
      args.append ('-xreinstall=true')

    apply (self.output, ['msg', '$ swinstall ' + arg_string (args) +
                         ' -s ' + pkg_depot + ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'swinstall', ['swinstall'] +
                       args + ['-s', pkg_depot, pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps:
      if not re.search ('enforce_dependencies=',
                        string.join (args, ' ')):
        args.append ('-xenforce_dependencies=false')
      if not re.search ('autoselect_dependents=',
                        string.join (args, ' ')):
        args.append ('-xautoselect_dependents=false')

    apply (self.output, ['msg', '$ swremove ' + arg_string (args) +
                         ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'swremove', ['swremove'] +
                       args + [pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    # package either in depot directory or depot file
    pkg_depot = None
    if os.path.exists (os.path.join (self.pkgdir,
                                     re.sub ('\.', '/', pkgname))):
      pkg_depot = self.pkgdir
      dest_depot = os.path.join (destdir, pkgname)
    elif os.path.exists (os.path.join (self.pkgdir,
                                       self.pkgname_base) + '.depot'):
      pkg_depot = os.path.join (self.pkgdir, self.pkgname_base) + '.depot'
      dest_depot = os.path.join (destdir, self.pkgname_base + '.depot')
    if not pkg_depot:
      raise pkgsys.PkgsysPackageNotFoundError ([re.sub ('\.', '/', pkgname),
                                                self.pkgname_base + '.depot'])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and os.path.exists (dest_depot):
      raise pkgsys.PkgsysExtractDuplicate (dest_depot)

    # UNIX copy if depot is a file else swcopy
    if os.path.isfile (pkg_depot):
      apply (self.output, ['msg', '$ cp ' + pkg_depot + ' ' + destdir + '\n'])
      if not dry_run:
        try:
          shutil.copyfile (pkg_depot, os.path.join (destdir,
                                                    self.pkgname_base +
                                                    '.depot'))
        except (IOError, OSError), msg:
          raise pkgsys.PkgsysExtractError (str (msg))
    elif os.path.isdir (pkg_depot):
      if os.spawnv (os.P_WAIT, '/usr/sbin/swcopy',
                    ['/usr/sbin/swcopy', '-xwrite_remote_files=true',
                     '-s', pkg_depot, pkgname, '@', destdir]):
        raise pkgsys.PkgsysExtractError ('error copying package with swcopy')
    else:
      raise pkgsys.PkgsysExtractError (pkg_depot + ' not recognized')

    return dest_depot

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      # determine if version in package database is newer than
      # installed version
      for pkgname in self.get_pkgnames (subpkg):
        entry = pkgcache.get (pkgname)
        if entry:
          # get package version and revision
          pkg_ver = entry.version
          pkg_rev = entry.revision

          if self.pkg_ver < pkg_ver:
            pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_ver > pkg_ver:
            pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_rev is not None:
            if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
              pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

          pkg_status.installed.append (pkgname)
        else:
          pkg_status.not_installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    entry = pkgcache.get (pkgname)
    if entry:
      # get package version and revision
      pkg_ver = entry.version
      pkg_rev = entry.revision
    else:
      raise pkgsys.PkgsysPackageNotFoundError

    return [pkg_ver, pkg_rev]

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { 'conf': [ '.' + self.pkgname_base +
                                             '-CONF' ],
                                   'doc': [ '.' + self.pkgname_base +
                                            '-HELP' ],
                                   'libruntime': [ '.' + self.pkgname_base +
                                                   '-LIBR' ],
                                   'man': [ '.' + self.pkgname_base +
                                            '-MAN' ],
                                   'packages': [ '.' + self.pkgname_base +
                                                 '-PKG' ],
                                   'runtime': [ '.' + self.pkgname_base +
                                                '-RUN' ] }))

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    if pkgname:
      pkgname_base = self.get_basename (pkgname)
    else:
      pkgname_base = self.pkgname_base

    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { 'conf': [ '.' + pkgname_base +
                                                 '-CONF' ],
                                       'doc': [ '.' + pkgname_base +
                                                '-HELP' ],
                                       'libruntime': [ '.' + pkgname_base +
                                                       '-LIBR' ],
                                       'man': [ '.' + pkgname_base +
                                                '-MAN' ],
                                       'packages': [ '.' + pkgname_base +
                                                     '-PKG' ],
                                       'runtime': [ '.' + pkgname_base +
                                                    '-RUN' ] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    pkgname_base = self.get_basename (pkgname)

    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ [ '.' + pkgname_base + '-CONF',
                                        'conf'],
                                      [ '.' + pkgname_base + '-HELP',
                                        'doc'],
                                      [ '.' + pkgname_base + '-LIBR',
                                        'libruntime'],
                                      [ '.' + pkgname_base + '-MAN',
                                        'man'],
                                      [ '.' + pkgname_base + '-PKG',
                                        'packages'],
                                      [ '.' + pkgname_base + '-RUN',
                                        'runtime'] ]))

  # given package name, determine basename
  def get_basename (self, pkgname):
    return (re.sub ('^([^\.]+)\..*', '\\1', pkgname))

  def list_packages (self):
    pkgcache.get_entries ()

    return pkgcache.entries

class MyPkgCache (PkgCache):
  def __init__ (self, pkgname_prefix, cache = False):
    PkgCache.__init__ (self, pkgname_prefix, cache)

    if cache:
      self.get_entries ()

  def exists (self, pkgname):
    if self.cache:
      if self.has_pkgname (pkgname):
        return True
    else:
      for entry in os.popen ('swlist -l fileset -a state -a revision ' +
                             pkgname + ' 2>&1', 'r').readlines ():
        if entry.find ('corrupt') != -1:
          break
        if re.match ('^\s+' + pkgname + '\s+', entry.rstrip ()):
          return True

    return False

  # return entry
  def get (self, pkgname):
    if self.cache:
      if self.has_pkgname (pkgname):
        return (self.get (pkgname))
    else:
      for entry in os.popen ('swlist -l fileset -a state -a revision ' +
                             pkgname + ' 2>&1', 'r').readlines ():
        if entry.find ('corrupt') != -1:
          break
        if re.match ('^\s+' + pkgname + '\s+', entry.rstrip ()):
          version = re.sub ('^\s+([^\s]+)\s+T\.(.*)\.([0-9]+)\s.*',
                            '\\2', entry.rstrip ())
          revision = re.sub ('^\s+([^\s]+)\s+T\.(.*)\.([0-9]+)\s.*',
                            '\\3', entry.rstrip ())

          return (PkgCacheEntry (pkgname, version, revision))

    return None

  # get listing of all package entries
  def get_entries (self):
    for entry in os.popen ('swlist -l fileset -a state -a revision ' +
                           self.pkgname_prefix + '\* 2>&1',
                           'r').readlines ():
      if not re.match ('\s+' + self.pkgname_prefix, entry):
        continue
      if entry.find ('corrupt') != -1:
        continue

      pkgname = re.sub ('^\s+([^\s]+)\s+T\.(.*)\.([0-9]+)\s.*',
                        '\\1', entry.rstrip ())
      version = re.sub ('^\s+([^\s]+)\s+T\.(.*)\.([0-9]+)\s.*',
                        '\\2', entry.rstrip ())
      revision = re.sub ('^\s+([^\s]+)\s+T\.(.*)\.([0-9]+)\s.*',
                        '\\3', entry.rstrip ())

      self.entries.append (PkgCacheEntry (pkgname, version, revision))
