# Copyright (C) 2001-2003, 2006, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import shutil
import string
import sys

from fnmatch import fnmatch

import pkgsys

from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# cache of installed packages
pkgcache = None

# constants
VERSION_STRING_LENGTH = 10		# length of version string

class inst (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = 0):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'inst'

    # initialize package cache
    if not pkgcache or (cache and not pkgcache.cache):
      pkgcache = PkgCache (cache)

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []

    # package must exist in depot before installation
    if not os.path.exists (os.path.join (self.pkgdir, self.pkgname_base)):
      raise pkgsys.PkgsysPackageNotFoundError ([self.pkgname_base])

    # default root
    if pkgmgr.root:
      args.append ('-r')
      args.append (pkgmgr.root)

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps and not 'rulesoverride:on' in args:
      args.append ('-V')
      args.append ('rulesoverride:on')

    # don't beep when package installation ends (if installing
    # many packages, could be annoying)
    if not re.search ('beep:', string.join (args, ' ')):
      args.append ('-V')
      args.append ('beep:off')

    # show files being installed
    if not re.search ('show_files:', string.join (args, ' ')):
      args.append ('-V')
      args.append ('show_files:on')

    # print files being installed on a separate line
    if not re.search ('overprint:', string.join (args, ' ')):
      args.append ('-V')
      args.append ('overprint:off')

    # increase buffer size for exitops
    if not re.search ('exitop_limit:', string.join (args, ' ')):
      args.append ('-V')
      args.append ('exitop_limit:300000')

    apply (self.output, ['msg', '$ inst -a ' + arg_string (args) +
                         ' -f ' + self.pkgdir + ' -I ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'inst', ['inst', '-a'] + args +
                       ['-f', self.pkgdir, '-I', pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    # default root
    if pkgmgr.root:
      args.append ('-r')
      args.append (pkgmgr.root)

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps:
      args.append ('-F')

    apply (self.output, ['msg', '$ versions ' + arg_string (args) +
                         ' remove ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'versions', ['versions'] +
                       args + ['remove', pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    # package must exist in depot before installation
    if not os.path.exists (os.path.join (self.pkgdir, self.pkgname_base)):
      raise pkgsys.PkgsysPackageNotFoundError ([self.pkgname_base])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and \
    os.path.exists (os.path.join (destdir, self.pkgname_base)):
      raise pkgsys.PkgsysExtractDuplicate (os.path.join (destdir,
                                                         self.pkgname_base))

    apply (self.output, ['msg', '$ cp ' + os.path.join (self.pkgdir,
                                                        self.pkgname_base) +
                         ' ' + os.path.join (self.pkgdir, self.pkgname_base) +
                         '.* ' + destdir + '\n'])

    if not dry_run:
      try:
        shutil.copyfile (os.path.join (self.pkgdir, self.pkgname_base),
                         os.path.join (destdir, self.pkgname_base))
      except (IOError, OSError), msg:
        raise pkgsys.PkgsysExtractError (str (msg))

      files = os.listdir (self.pkgdir)
      for file in files:
        if fnmatch (file, self.pkgname_base + '.*'):
          try:
            shutil.copyfile (os.path.join (self.pkgdir, file),
                             os.path.join (destdir, file))
          except (IOError, OSError), msg:
            raise pkgsys.PkgsysExtractError (str (msg))

    return os.path.join (destdir, self.pkgname_base)

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      # determine if version in package database is newer than
      # installed version
      for pkgname in self.get_pkgnames (subpkg):
        entry = pkgcache.version (pkgname)
        if entry:
          # get package version. the version contains the package
          # revision embedded in it and the pkg-db.xml file will
          # never have a test or entry for the revision component
          pkg_ver = Version (entry)

          if self.pkg_ver < pkg_ver:
            pkg_status.downgrade[pkgname] = [pkg_ver, Version (0)]
            continue

          if self.pkg_ver > pkg_ver:
            pkg_status.upgrade[pkgname] = [pkg_ver, Version (0)]
            continue

          pkg_status.installed.append (pkgname)
        else:
          pkg_status.not_installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # given package name, determine basename
  def get_basename (self, pkgname):
    return (re.sub ('^([^\.]+).*', '\\1', pkgname))

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    # determine which subpackages are installed. only check installed
    entry = pkgcache.version (pkgname)
    if entry:
      # get package version. the version contains the package
      # revision embedded in it and the pkg-db.xml file will
      # never have a test or entry for the revision component
      pkg_ver = Version (entry)
    else:
      raise pkgsys.PkgsysPackageNotFoundError

    return [pkg_ver, Version (0)]

  def get_pkgdb_depend_str (self, ver_tests):
    return re.sub ('>=', '&gt;=', ver_tests)

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { 'conf': [ '.sw.config' ],
                                   'doc': [ '.man.doc' ],
                                   'libruntime': [ '.sw.librun' ],
                                   'man': [ '.man.info', '.man.pages' ],
                                   'packages': [ '.sw.packages' ],
                                   'runtime': [ '.sw.base' ] }))

  # get revision specific to package-management system
  def get_pkgsys_version (self, version, revision):
    # pkgutils releases prior to 1.5 didn't contain a revision
    # element for IRIX packages. compensate by splitting the
    # two-character revision from the end of the version string.
    if revision and revision != '0':
      return Version (version), Version (revision)
    if len (version) == VERSION_STRING_LENGTH:
      new_version = Version (version)
      new_revision_str = str (version)[-2:]
      if new_revision_str[0:1] == '0':
        new_revision = Version (str (version)[-1:])
      else:
        new_revision = Version (str (version)[-2:])
      return new_version, new_revision

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { 'conf': [ '.sw.config' ],
                                       'doc': [ '.man.doc' ],
                                       'libruntime': [ '.sw.librun' ],
                                       'man': [ '.man.info', '.man.pages' ],
                                       'packages': [ '.sw.packages' ],
                                       'runtime': [ '.sw.base' ] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ [ '.sw.config', 'conf' ],
                                      [ '.man.doc' , 'doc' ],
                                      [ '.sw.librun', 'libruntime' ],
                                      [ '.man.info', 'man' ],
                                      [ '.man.pages', 'man' ],
                                      [ '.sw.packages', 'packages' ],
                                      [ '.sw.base', 'runtime' ] ]))

  # package version string. no package revision.
  def version_str (self, pkg_ver, pkg_rev):
    return pkg_ver

class PkgCache:
  def __init__ (self, cache = 0):
    self.cache = cache
    self.neg_entries = []			# negative hits
    self.pos_entries = {}			# positive hits

  # add entries to cache
  def add (self, pkgname):
    if self.cache:
      for entry in self.get_pkg_entry (pkgname):
        if re.match ('\w+\s+' + pkgname + '\s+', entry.rstrip ()):
          self.pos_entries[pkgname] = {}

          self.pos_entries[pkgname]['version'] = \
            re.sub ('^\w+\s+[\w.]+\s+(\d+).*', '\\1', entry.rstrip ())

          break
      else:
        self.neg_entries.append (pkgname)
        return 0

      return 1

  # get listing for package
  def get_pkg_entry (self, pkgname):
      return os.popen ('versions -I -n ' + pkgname +
                       ' 2>&1', 'r').readlines ()

  def exists (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return 0
      if self.pos_entries.has_key (pkgname):
        return 1

      # populate cache with entry
      if self.add (pkgname):
        return 1
    else:
      for entry in self.get_pkg_entry (pkgname):
        if re.match ('\w+\s+' + pkgname + '\s+', entry.rstrip ()):
          return 1

    return 0

  def version (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return None
      if self.pos_entries.has_key (pkgname):
        return self.pos_entries[pkgname]['version']

      # populate cache with entry
      if self.add (pkgname):
        return self.pos_entries[pkgname]['version']

    else:
      for entry in self.get_pkg_entry (pkgname):
        if re.match ('\w+\s+' + pkgname + '\s+', entry.rstrip ()):
          return re.sub ('^\w+\s+[\w.]+\s+(\d+).*', '\\1',
                         entry.rstrip ())

    return None
