# Copyright (C) 2001-2003, 2006, 2015 The Written Word, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import shutil
import string
import sys

import pkgsys

from pkgcache import PkgCacheEntry, PkgCache
from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# cache of installed packages
pkgcache = None

# constants
VERSION_STRING_LENGTH = 4		# length of version string

class lpp (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = False):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'lpp'

    # initialize package cache
    if not pkgcache or (cache and not pkgcache.cache):
      pkgcache = MyPkgCache (pkgmgr.pkgname_prefix, cache)

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.reinstall and not '-F' in args:
      args.append ('-F')

    # show files being installed
    if pkgmgr.verbose and not re.search ('-V', string.join (args, ' ')):
      args.append ('-V2')

    # package must exist in depot before installation
    pkg_depot = os.path.join (self.pkgdir, self.pkgname_base + '.bff')
    if not os.path.exists (pkg_depot):
      raise pkgsys.PkgsysPackageNotFoundError ([pkg_depot])

    # append arg string with space to separate from next argument
    args_str = arg_string (args)
    if args_str:
      args_str += ' '
    apply (self.output, ['msg', '$ installp -acqQ ' +
                         args_str + '-d ' + pkg_depot +
                         ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'installp', ['installp', '-acqQ'] +
                       args + ['-d', pkg_depot, pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    # show files being removed
    if not re.search ('-V', string.join (args, ' ')):
      args.append ('-V2')

    apply (self.output, ['msg', '$ installp -u ' + arg_string (args) +
                         ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'installp', ['installp', '-u'] +
                       args + [pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    # package must exist in depot before installation
    pkg_depot = os.path.join (self.pkgdir, self.pkgname_base + '.bff')
    if not os.path.exists (pkg_depot):
      raise pkgsys.PkgsysPackageNotFoundError ([pkg_depot])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and \
    os.path.exists (os.path.join (destdir, os.path.basename (pkg_depot))):
      raise pkgsys.PkgsysExtractDuplicate (os.path.join (destdir,
                                           os.path.basename (pkg_depot)))

    apply (self.output, ['msg', '$ cp ' +
                         os.path.join (self.pkgdir,
                                       self.pkgname_base + '.bff') + ' ' +
                         os.path.join (destdir,
                                       self.pkgname_base + '.bff') + '\n'])

    if not dry_run:
      try:
        shutil.copyfile (os.path.join (self.pkgdir,
                                       self.pkgname_base + '.bff'),
                         os.path.join (destdir,
                                       self.pkgname_base + '.bff'))
      except (IOError, OSError), msg:
        raise pkgsys.PkgsysExtractError (str (msg))

      apply (self.output, ['msg', '$ /usr/sbin/inutoc ' + destdir + '\n'])
      if os.spawnv (os.P_WAIT, '/usr/sbin/inutoc',
                    ['/usr/sbin/inutoc', destdir]):
        raise pkgsys.PkgsysExtractError ('error creating TOC file with \
inutoc in ' + destdir)

    return os.path.join (destdir, os.path.basename (pkg_depot))

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      # determine if version in package database is newer than
      # installed version
      for pkgname in self.get_pkgnames (subpkg):
        entry = pkgcache.get (pkgname)
        if entry:
          # get package version and revision
          pkg_ver = entry.version
          pkg_rev = entry.revision

          if self.pkg_ver < pkg_ver:
            pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_ver > pkg_ver:
            pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_rev is not None:
            if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
              pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

          pkg_status.installed.append (pkgname)
        else:
          pkg_status.not_installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # given package name, determine basename
  def get_basename (self, pkgname):
    return (re.sub ('^(.*)\.[^\.]+$', '\\1', pkgname))

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    # determine which subpackages are installed. only check installed
    entry = pkgcache.get (pkgname)
    if entry:
      return [entry.version, entry.revision]
    else:
      raise pkgsys.PkgsysPackageNotFoundError

    return [None, None]

  def get_pkgdb_depend_str (self, ver_tests):
    return re.sub ('>=', '&gt;=', ver_tests)

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { 'conf': ['.conf'], 'doc': ['.doc'],
                                   'libruntime': ['.librun'],
                                   'man': ['.man'],
                                   'packages': ['.pkg'],
                                   'runtime': ['.rte'] }))

  # make version from package version/revision specific to
  # package-management system
  def make_version (self, version, revision):
    return Version (str (version) + '.' + str (revision))
  def make_revision (self, version, revision):
    return Version (0)

  # get package version specific to package-management system
  def get_version (self, version):
    return Version ('.'.join (version.split('.')[:3]))

  # get package revision specific to package-management system
  def get_revision (self, version, revision = None):
    return Version (version.split('.')[3])

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { 'conf': ['.conf'], 'doc': ['.doc'],
                                       'libruntime': ['.librun'],
                                       'man': ['.man'],
                                       'packages': ['.pkg'],
                                       'runtime': ['.rte'] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ ['.conf', 'conf'], ['.doc' , 'doc'],
                                      ['.librun', 'libruntime'],
                                      ['.man', 'man'],
                                      ['.pkg', 'packages'],
                                      ['.rte', 'runtime'] ]))

  def list_packages (self):
    pkgcache.get_entries ()

    return pkgcache.entries

class MyPkgCache (PkgCache):
  # add entries to cache
  def add (self, pkgname):
    if self.cache:
      for entry in self.get_pkg_entry (pkgname):
        if re.match (pkgname, entry.split (':')[1]):
          if entry.split (':')[5] == 'C':
            lpp_version = entry.split (':')[2].split ('.')
            version = lpp_version[0] + '.' + lpp_version[1] + '.' + \
                      lpp_version[2]
            revision = lpp_version[3]

            self.entries.append (PkgCacheEntry (pkgname, version,
                                                revision))

            break
      else:
        self.neg_entries.append (pkgname)
        return False

      return True

  def exists (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return False
      if self.has_pkgname (pkgname):
        return True

      # populate cache with entry
      if self.add (pkgname):
        return True
    else:
      for entry in self.get_pkg_entry (pkgname):
        if re.match (pkgname, entry.split (':')[1]):
          if entry.split (':')[5] == 'C':
            return True

    return False

  # return entry
  def get (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return None
      if self.has_pkgname (pkgname):
        return (self.get (pkgname))

      # populate cache with entry
      if self.add (pkgname):
        return (self.get (pkgname))
    else:
      for entry in self.get_pkg_entry (pkgname):
        if re.match (pkgname, entry.split (':')[1]):
          if entry.split (':')[5] == 'C':
            lpp_version = entry.split (':')[2].split ('.')
            version = lpp_version[0] + '.' + lpp_version[1] + '.' + \
                      lpp_version[2]
            revision = lpp_version[3]

            return (PkgCacheEntry (pkgname, version, revision))

    return None

  # get listing of all package entries
  def get_entries (self):
    for entry in os.popen ('lslpp -q -c -L ' + self.pkgname_prefix +
                           '\* 2>&1', 'r').readlines ():
      if not re.match (".*:" + self.pkgname_prefix + "\..*",
                       entry.strip ()):
        continue

      pkgname = entry.split (':')[1]

      lpp_version = entry.split (':')[2].split ('.')
      version = lpp_version[0] + '.' + lpp_version[1] + '.' + \
                lpp_version[2]
      revision = lpp_version[3]

      if not self.has_pkgname (pkgname):
        self.entries.append (PkgCacheEntry (pkgname, version, revision))

  # get listing for package
  def get_pkg_entry (self, pkgname):
    # packages not in the root fileset need to be queried
    # using the /usr/lib/objrepos ODM repository
    if pkgname[-4:] != 'conf':
      os.environ['ODMDIR'] = '/usr/lib/objrepos'

    entries = os.popen ('lslpp -q -c -L ' + pkgname +
                        ' 2>&1', 'r').readlines ()

    if pkgname[-4:] != 'conf':
      del os.environ['ODMDIR']

    return entries
