# Copyright (C) 2001-2003, 2006, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import string
import sys

import misclib
import pkgsys

from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# cache of installed packages
pkgcache = None

class npkg (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = 0):
    Pkgsys.__init__ (self, output)
    self.pkg_type = 'npkg'
    self.npkg = {}

    if pkgmgr:
      if pkgmgr.npkg_db:
        self.npkg['db'] = pkgmgr.npkg_db
      if pkgmgr.npkg_arch:
        self.npkg['systype'] = pkgmgr.npkg_arch

      # make sure npkgadd is in $PATH
      if not pkgmgr.npkgadd_path:
        self.npkg['npkgadd path'] = misclib.find_path_prog ('npkgadd')
        if not self.npkg['npkgadd path']:
          raise pkgsys.PkgsysProgNotFound ('npkgadd', 'npkgadd-path')
      else:
         self.npkg['npkgadd path'] = pkgmgr.npkgadd_path

      # make sure npkginfo is in $PATH
      if not pkgmgr.npkginfo_path:
        self.npkg['npkginfo path'] = misclib.find_path_prog ('npkginfo')
        if not self.npkg['npkginfo path']:
          raise pkgsys.PkgsysProgNotFound ('npkginfo', 'npkginfo-path')
      else:
        self.npkg['npkginfo path'] = pkgmgr.npkginfo_path

      # make sure npkgrm is in $PATH
      if not pkgmgr.npkgrm_path:
        self.npkg['npkgrm path'] = misclib.find_path_prog ('npkgrm')
        if not self.npkg['npkgrm path']:
          raise pkgsys.PkgsysProgNotFound ('npkgrm', 'npkgrm-path')
      else:
        self.npkg['npkgrm path'] = pkgmgr.npkgrm_path

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []
    root = pkgmgr.root

    # default root
    if root:
      args.append ('-R')
      args.append (root)

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.reinstall and not '--reinstall' in args:
      args.append ('--reinstall')
    if pkgmgr.ignoredeps and not '--ignoredeps' in args:
      args.append ('--ignoredeps')
    if self.npkg['db']:
      args.append ('--db-path=' + self.npkg['db'])

    # convert `.' to / as `.' is subpackage divider
    pkg = re.sub ('\.', '/', pkgname)

    # package must exist in depot before installation
    if os.path.isdir (os.path.join (self.pkgdir, pkg)):
      pkg_depot = self.pkgdir
    else:
      raise pkgsys.PkgsysPackageNotFoundError ([pkg])

    apply (self.output, ['msg', '$ ' + self.npkg['npkgadd path'] +
                         ' ' + arg_string (args) + ' --depot=' +
                         pkg_depot + ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, self.npkg['npkgadd path'],
                       [self.npkg['npkgadd path']] + args +
                       ['--depot=' + pkg_depot, pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []
    root = pkgmgr.root

    # default root
    if root:
      args.append ('-R')
      args.append (root)

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps and not '--ignoredeps' in args:
      args.append ('--ignoredeps ')
    if self.npkg['db']:
      args.append ('--db-path=' + self.npkg['db'])

    # convert `.' to / as `.' is subpackage divider
    pkg = re.sub ('\.', '/', pkgname)

    apply (self.output, ['msg', '$ ' + self.npkg['npkgrm path'] +
                         ' ' + arg_string (args) + ' --systype=' +
                         self.npkg['systype'] + ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, self.npkg['npkgrm path'],
                       [self.npkg['npkgrm path']] + args +
                       ['--systype=' + self.npkg['systype'],
                       pkgname]) == 0

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = None			# if subpackages found
    npkg_db = []

    if self.npkg['db']:
      npkg_db.append ('--db-path=' + self.npkg['db'])

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      for pkgname in self.get_pkgnames (subpkg):
        if os.spawnvp (os.P_WAIT, self.npkg['npkginfo path'],
                       [self.npkg['npkginfo path']] + npkg_db +
                       ['--systype=' + self.npkg['systype'],
                       '--query=package-installed', pkgname]):
          pkg_status.not_installed.append (pkgname)
        else:
          pkg_ver = Version (0)
          pkg_rev = Version (0)

          # determine if version in package database is newer than
          # installed version
          for entry in misclib.popen (self.npkg['npkginfo path'],
                                      npkg_db + ['--systype=' +
                                                 self.npkg['systype'],
                                                 '--display-long', pkgname],
                                      'r').readlines ():
            if entry.find ('Version: ') != -1:
              # get package version and revision
              pkg_ver = Version (re.sub ('.*Version: (.*)$', '\\1',
                                         entry.rstrip ()))
            elif entry.find ('Revision: ') != -1:
              pkg_rev = Version (re.sub ('.*Revision: (.*)$', '\\1',
                                         entry.rstrip ()))
            else:
              continue

            if pkg_ver and pkg_rev:
              if self.pkg_ver < pkg_ver:
        	pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
        	break

              if self.pkg_ver > pkg_ver:
                pkg_status.upgrade[pkgname] = [ pkg_ver, pkg_rev ]
                break

              if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
                pkg_status.upgrade[pkgname] = [ pkg_ver, pkg_rev ]
                break

              pkg_status.installed.append (pkgname)
              break

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    npkg_db = []

    if self.npkg['db']:
      npkg_db.append ('--db-path=' + self.npkg['db'])

    if os.spawnvp (os.P_WAIT, self.npkg['npkginfo path'],
                   [self.npkg['npkginfo path']] + npkg_db +
                   ['--systype=' + self.npkg['systype'],
                   '--query=package-installed', pkgname]):
      raise pkgsys.PkgsysPackageNotFoundError
    else:
      pkg_ver = Version (0)
      pkg_rev = Version (0)

      # determine if version in package database is newer than
      # installed version
      for entry in misclib.popen (self.npkg['npkginfo path'],
                                  npkg_db + ['--systype=' +
                                             self.npkg['systype'],
                                             '--display-long', pkgname],
                                  'r').readlines ():
        if entry.find ('Version: ') != -1:
          # get package version and revision
          pkg_ver = Version (re.sub ('.*Version: (.*)$', '\\1',
                                     entry.rstrip ()))
        elif entry.find ('Revision: ') != -1:
          pkg_rev = Version (re.sub ('.*Revision: (.*)$', '\\1',
                                     entry.rstrip ()))
        else:
          continue

        if pkg_ver and pkg_rev:
          return [pkg_ver, pkg_rev]

    return [None, None]

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { 'conf': ['.conf'], 'doc': ['.doc'],
                                   'libruntime': ['.librun'],
                                   'man': ['.man'], 'packages': ['.pkg'],
                                   'runtime': ['.run'] }))

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { 'conf': ['.conf'], 'doc': ['.doc'],
                                       'libruntime': ['.librun'],
                                       'man': ['.man'],'packages': ['.pkg'],
                                       'runtime': ['.run'] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ ['.conf', 'conf'], ['.doc', 'doc'],
                                      ['.librun', 'libruntime'],
                                      ['.man', 'man'], ['.pkg', 'packages'],
                                      ['.run', 'runtime'] ]))

  # given package name, determine basename
  def get_basename (self, pkgname):
    return (re.sub ('^(.*)\..*$', '\\1', pkgname))
