# Copyright (C) 2001-2003, 2005-2006, 2015, 2019 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import string
import sys

import pkgsys

from pkgcache import PkgCacheEntry, PkgCache
from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# constants
datadir = '/opt/TWWfsw/pkgutils16/share'

# cache of installed packages
pkgcache = None

class pkgadd (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = False):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'pkgadd'

    # initialize package cache
    if not pkgcache or (cache and not pkgcache.cache):
      pkgcache = MyPkgCache (pkgmgr.pkgname_prefix, cache)
      if pkgmgr.root:
        pkgcache.root = pkgmgr.root

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []

    # default root
    if pkgmgr.root:
      args.append ('-R')
      args.append (pkgmgr.root)

    # if no admin file specified as an argument, specify our custom
    # admin file
    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if not '-a' in args:
      args.append ('-a')
      args.append (datadir + '/pkgadd/admin')

    # package must exist in depot before installation
    pkg_depot = None
    if os.path.isdir (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = self.pkgdir
    if os.path.isfile (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = os.path.join (self.pkgdir, pkgname)
    if not pkg_depot:
      raise pkgsys.PkgsysPackageNotFoundError ([pkgname])

    apply (self.output, ['msg', '$ pkgadd ' + arg_string (args) +
                         ' -d ' + pkg_depot + ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'pkgadd', ['pkgadd'] +
                       args + ['-d', pkg_depot, pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    # default root
    if pkgmgr.root:
      args.append ('-R')
      args.append (pkgmgr.root)

    # if no admin file specified as an argument, specify our custom
    # admin file
    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if not '-a' in args:
      args.append ('-a')
      args.append (datadir + '/pkgadd/admin')

    apply (self.output, ['msg', '$ pkgrm ' + arg_string (args) +
                         ' -n ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'pkgrm', ['pkgrm'] +
                       args + ['-n', pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    # package must exist in depot before installation
    pkg_depot = None
    if os.path.isdir (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = self.pkgdir
    if os.path.isfile (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = os.path.join (self.pkgdir, pkgname)
    if not pkg_depot:
      raise pkgsys.PkgsysPackageNotFoundError ([pkgname])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and \
    os.path.exists (os.path.join (destdir, pkgname)):
      raise pkgsys.PkgsysExtractDuplicate (os.path.join (destdir, pkgname))

    apply (self.output, ['msg', '$ pkgtrans -o ' + pkg_depot + ' ' +
                         destdir + ' ' + pkgname + '\n'])

    if not dry_run:
      if os.spawnv (os.P_WAIT, '/usr/bin/pkgtrans',
                    ['/usr/bin/pkgtrans', '-o', pkg_depot,
                     destdir, pkgname]):
        raise pkgsys.PkgsysExtractError ('error copying package to ' +
                                         destdir + ' with pkgtrans')

    return os.path.join (destdir, pkgname)

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      for pkgname in self.get_pkgnames (subpkg):
        if not pkgcache.exists (pkgname):
          pkg_status.not_installed.append (pkgname)
        else:
          # determine if version in package database is newer than
          # installed version
          entry = pkgcache.get (pkgname)
          if entry:
            # get package version and revision
            pkg_ver = entry.version
            pkg_rev = entry.revision

            if self.pkg_ver < pkg_ver:
              pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

            if self.pkg_ver > pkg_ver:
              pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

            if self.pkg_rev is not None:
              if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
                pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
                continue

            pkg_status.installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return pkg_status

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    if not pkgcache.exists (pkgname):
      raise pkgsys.PkgsysPackageNotFoundError
    else:
      # determine if version in package database is newer than
      # installed version
      entry = pkgcache.get (pkgname)
      if entry:
        return [entry.version, entry.revision]

    return [None, None]

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return Pkgsys.get_pkgnames (self, subpkg_type,
                                { "conf": ["c"], "doc": ["d"],
                                  "libruntime": ["r"], "man": ["m"],
                                  "packages": ["p"], "runtime": ["u"] })

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return Pkgsys.get_subpkg_names (self, subpkg_type,
                                    { "conf": ["c"], "doc": ["d"],
                                      "libruntime": ["r"], "man": ["m"],
                                      "packages": ["p"], "runtime": ["u"] })

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    return Pkgsys.get_subpkg_type (self, pkgname,
                                   [ ["c", "conf"], ["d", "doc"],
                                     ["r", "libruntime"], ["m", "man"],
                                     ["p", "packages"], ["u", "runtime"] ])

  # given package name, determine basename
  def get_basename (self, pkgname):
    return re.sub ('^(.*).$', '\\1', pkgname)

  def list_packages (self):
    pkgcache.get_entries ()

    return pkgcache.entries

class MyPkgCache (PkgCache):
  def add (self, pkgname):
    if self.cache:
      if self.root:
        entries = os.popen ("pkginfo -l -R " + self.root + " " +
                            pkgname + " 2>&1", "r").readlines ()
      else:
        entries = os.popen ("pkginfo -l " + pkgname + " 2>&1",
                            "r").readlines ()

      index = 0
      while index < len (entries):
        if entries[index].find ('PKGINST:  ' + self.pkgname_prefix) == -1:
          index += 1
          continue

        pkgname = re.sub ('.*PKGINST:\s+(.*)', '\\1',
                          entries[index].rstrip ())

        index += 1
        while index < len (entries):
          if entries[index].rstrip () == '':
            break
          if entries[index].find ('STATUS:  partially') != -1:
            break
          if entries[index].find ('VERSION:  ') != -1:
            version = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\1',
                              entries[index].rstrip ())
            revision = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\2',
                              entries[index].rstrip ())

            self.entries.append (PkgCacheEntry (pkgname, version,
                                                revision))

            break
          index += 1
        index += 1
      else:
        self.neg_entries.append (pkgname)
        return False

      return True

  def exists (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return False
      if self.has_pkgname (pkgname):
        return True

      # populate cache with entry
      if self.add (pkgname):
        return True
    else:
      if self.root:
        if not os.system ("pkginfo -i -R " + self.root + " " +
                          pkgname + ">/dev/null 2>&1"):
          return True
      else:
        if not os.system ("pkginfo -i " + pkgname + ">/dev/null 2>&1"):
          return True

    return False

  # return entry
  def get (self, pkgname):
    if self.cache:
      if pkgname in self.neg_entries:
        return None
      if self.has_pkgname (pkgname):
        return (self.get (pkgname))

      # populate cache with entry
      if self.add (pkgname):
        return (self.get (pkgname))
    else:
      if self.root:
        entries = os.popen ("pkginfo -l -R " + self.root + " " +
                            pkgname + " 2>&1", "r").readlines ()
      else:
        entries = os.popen ("pkginfo -l " + pkgname + " 2>&1",
                            "r").readlines ()

      for entry in entries:
        if entry.find ('STATUS:  partially') != -1:
          break
        if entry.find ('VERSION:  ') != -1:
          version = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\1',
                            entry.rstrip ())
          revision = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\2',
                            entry.rstrip ())

          return (PkgCacheEntry (pkgname, version, revision))

    return None

  # get listing of all package entries
  def get_entries (self):
    if self.root:
      entries = os.popen ("pkginfo -l -R " + self.root + " 2>&1",
                          "r").readlines ()
    else:
      entries = os.popen ("pkginfo -l 2>&1", "r").readlines ()

    index = 0
    while index < len (entries):
      if entries[index].find ('PKGINST:  ' + self.pkgname_prefix) == -1:
        index += 1
        continue

      pkgname = re.sub ('.*PKGINST:\s+(.*)', '\\1',
                        entries[index].rstrip ())

      index += 1
      while index < len (entries):
        if entries[index].rstrip () == '':
          break
        if entries[index].find ('STATUS:  partially') != -1:
          break
        if entries[index].find ('VERSION:  ') != -1:
          version = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\1',
                            entries[index].rstrip ())
          revision = re.sub ('.*VERSION:\s+([^,][^,]*),REV=(.*)', '\\2',
                            entries[index].rstrip ())

          self.entries.append (PkgCacheEntry (pkgname, version, revision))

          break
        index += 1
      index += 1
