# Copyright (C) 2001-2003, 2006, 2008, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import shutil
import string
import sys

from fnmatch import fnmatch

import misclib
import pkgsys

from pkgcache import PkgCacheEntry, PkgCache
from pkgstatus import PkgStatus
from pkgsys import Pkgsys
from pkgsys import arg_string
from version import Version

# cache of installed packages
pkgcache = None

# RPM architecture type
rpm_arch = None

# get RPM architecture type identifier for OS
def get_rpm_arch (gnu_systype):
  global rpm_arch

  if rpm_arch:
    return rpm_arch

  if fnmatch (gnu_systype, '*-aix*'):
    rpm_arch = 'ppc'
  elif fnmatch (gnu_systype, 'hppa*-hpux*'):
    rpm_arch = 'parisc'
  elif fnmatch (gnu_systype, 'ia64*-hpux*'):
    rpm_arch = 'ia64'
  elif fnmatch (gnu_systype, 'x86_64*-linux*'):
    rpm_arch = 'x86_64'
  elif fnmatch (gnu_systype, '*-irix*'):
    rpm_arch = 'mipseb'
  elif fnmatch (gnu_systype, '*-linux*') or \
       fnmatch (gnu_systype, '*-pc-solaris*'):
    rpm_arch = 'i386'
  elif fnmatch (gnu_systype, '*-osf*'):
    rpm_arch = 'alpha'
  elif fnmatch (gnu_systype, '*-solaris*'):
    rpm_arch = 'sparc'
  else:
    return None

  return rpm_arch

class rpm (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = False):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'rpm'
    self.rpm = {}
    if pkgmgr:
      self.rpm['db'] = pkgmgr.rpm_db

    # make sure rpm is in $PATH
    if pkgmgr:
      if not pkgmgr.rpm_path:
        self.rpm['rpm path'] = misclib.find_path_prog ('rpm')
        if not self.rpm['rpm path']:
          raise pkgsys.PkgsysProgNotFound ('rpm', 'rpm-path')
      else:
        self.rpm['rpm path'] = pkgmgr.rpm_path

    # initialize package cache
    if pkgmgr:
      if not pkgcache or (cache and not pkgcache.cache):
        pkgcache = MyPkgCache (pkgmgr.pkgname_prefix, self.rpm['db'],
                               self.rpm['rpm path'], cache)

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps and not '--nodeps' in args:
      args.append ('--nodeps')
    if pkgmgr.reinstall and not '--replacepkgs' in args:
      args.append ('--replacepkgs')
    if self.rpm['db']:
      args.append ('--dbpath')
      args.append (self.rpm['db'])

    # determine RPM architecture type
    if not pkgmgr.rpm_arch:
      arch = get_rpm_arch (pkgmgr.systype)
    else:
      arch = pkgmgr.rpm_arch

    # full path to rpm package
    rpm_file = pkgname + '-' + self.pkg_ver + '-' + \
               self.pkg_rev + '.' + arch + '.rpm'

    # package must exist in depot before installation
    if not os.path.exists (os.path.join (self.pkgdir, rpm_file)):
      raise pkgsys.PkgsysPackageNotFoundError ([rpm_file])

    apply (self.output, ['msg', '$ ' + self.rpm['rpm path'] +
                         ' -Uv ' + arg_string (args) + ' ' +
                         os.path.join (self.pkgdir, rpm_file) + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, self.rpm['rpm path'],
                       [self.rpm['rpm path'], '-Uv'] + args +
                       [os.path.join (self.pkgdir, rpm_file)]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)
    if pkgmgr.ignoredeps and not '--nodeps' in args:
      args.append ('--nodeps')
    if self.rpm['db']:
      args.append ('--dbpath')
      args.append (self.rpm['db'])

    apply (self.output, ['msg', '$ ' + self.rpm['rpm path'] +
                         ' -ev ' + arg_string (args) + ' ' +
                         pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, self.rpm['rpm path'],
                       [self.rpm['rpm path'], '-ev'] +
                       args + [pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    # determine RPM architecture type
    if not pkgmgr.rpm_arch:
      arch = get_rpm_arch (pkgmgr.systype)
    else:
      arch = pkgmgr.rpm_arch

    # full path to rpm package
    rpm_file = pkgname + '-' + self.pkg_ver + '-' + \
               self.pkg_rev + '.' + arch + '.rpm'

    # package must exist in depot before installation
    if not os.path.exists (os.path.join (self.pkgdir, rpm_file)):
      raise pkgsys.PkgsysPackageNotFoundError ([rpm_file])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and \
    os.path.exists (os.path.join (destdir, rpm_file)):
      raise pkgsys.PkgsysExtractDuplicate (os.path.join (destdir, rpm_file))

    apply (self.output, ['msg', '$ cp ' + os.path.join (self.pkgdir,
                                                        rpm_file) + ' ' +
                         destdir + '\n'])

    if not dry_run:
      try:
        shutil.copyfile (os.path.join (self.pkgdir, rpm_file),
                         os.path.join (destdir, rpm_file))
      except (IOError, OSError), msg:
        raise pkgsys.PkgsysExtractError (str (msg))

    return os.path.join (destdir, rpm_file)

  # remove package
  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      for pkgname in self.get_pkgnames (subpkg):
        entry = pkgcache.get (pkgname)
        if entry:
          # get package version and revision
          pkg_ver = entry.version
          pkg_rev = entry.revision

          if self.pkg_ver < pkg_ver:
            pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_ver > pkg_ver:
            pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_rev is not None:
            if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
              pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

          pkg_status.installed.append (pkgname)
        else:
          pkg_status.not_installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    # determine if version in package database is newer than
    # installed version
    entry = pkgcache.get (pkgname)
    if entry:
      return [entry.version, entry.revision]
    else:
      raise pkgsys.PkgsysPackageNotFoundError

    return [None, None]

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { "conf": ["-conf"], "doc": ["-doc"],
                                   "libruntime": ["-librun"],
                                   "man": ["-man"], "packages": ["-pkg"],
                                   "runtime": [""] }))

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { "conf": ["-conf"], "doc": ["-doc"],
                                       "libruntime": ["-librun"],
                                       "man": ["-man"], "packages": ["-pkg"],
                                       "runtime": [""] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ ["-conf", "conf"], ["-doc", "doc"],
                                      ["-librun", "libruntime"],
                                      ["-man", "man"], ["-pkg", "packages"],
                                      ["", "runtime"] ]))

  # given package name, determine basename
  def basename (self, pkgname):
    return (re.sub ('^([^-]+).*$', '\\1', pkgname))

  def list_packages (self):
    pkgcache.get_entries ()

    return pkgcache.entries

class MyPkgCache (PkgCache):
  def __init__ (self, pkgname_prefix, rpm_db, rpm_path, cache = False):
    PkgCache.__init__ (self, pkgname_prefix, cache)

    self.pos_entries = {}
    self.rpm_db = rpm_db
    self.rpm_path = rpm_path

    if cache:
      self.get_entries ()

  def exists (self, pkgname):
    if self.cache:
      if self.has_pkgname (pkgname):
        return True
    else:
      if self.rpm_db:
        rpm_db = '--dbpath ' + self.rpm_db + ' '
      else:
        rpm_db = ''

      if not os.system (self.rpm_path + ' -q ' + rpm_db + ' ' +
                        pkgname + ' >/dev/null 2>&1'):
        return True

    return False

  # return entry
  def get (self, pkgname):
    if self.cache:
      if self.has_pkgname (pkgname):
        return (self.get (pkgname))
    else:
      if self.rpm_db:
        rpm_db = '--dbpath ' + self.rpm_db + ' '
      else:
        rpm_db = ''

      for entry in os.popen (self.rpm_path + ' -q \
--qf "%{name} %{version} %{release}\n" ' + rpm_db +
                             pkgname + ' 2>&1', 'r').readlines ():
        entry = entry.rstrip ()

        if pkgname == entry.split ()[0]:
          version = entry.split ()[1]
          revision = entry.split ()[2]

          return (PkgCacheEntry (pkgname, version, revision))

    return None

  # get listing of all package entries
  def get_entries (self):
    if self.rpm_db:
      rpm_db = '--dbpath ' + self.rpm_db + ' '
    else:
      rpm_db = ''

    for entry in os.popen (self.rpm_path + ' -qa \
--qf "%{name} %{version} %{release}\n" ' +
                           rpm_db, 'r').readlines ():
      entry = entry.rstrip ()

      if not re.match (self.pkgname_prefix, entry):
        continue

      # skip OS stub package
      if re.match ('TWWosstub10', entry):
        continue

      pkgname = entry.split ()[0]
      version = entry.split ()[1]
      revision = entry.split ()[2]

      self.entries.append (PkgCacheEntry (pkgname, version, revision))
