# Copyright (C) 2001-2003, 2006, 2015 The Written Word, Inc.
# Copyright (C) 2000-2001 The Written Word, LLC
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id$

import os
import re
import shutil
import string
import sys

import misclib
import pkgsys

from pkgstatus import PkgStatus
from pkgsys import Pkgsys, arg_string
from version import Version

# cache of installed packages
pkgcache = None

# cached values for later use (expensive to recalculate)
os_ident = None

# get 3-digit identifier for OS. each package has this identifier
# appended to the package name
def get_os_ident ():
  global os_ident
  if os_ident: return (os_ident)

  uname = os.uname ()
  if uname[2] == 'V4.0':
    os_ident = '425'
  elif uname[2] == 'V5.0':
    os_ident = '505'
  elif uname[2] == 'V5.1':
    os_ident = '510'
  else:
    return (None)

  return (os_ident)

class setld (Pkgsys):
  def __init__ (self, pkgmgr, output, cache = 0):
    global pkgcache

    Pkgsys.__init__ (self, output)
    self.pkg_type = 'setld'

    # initialize package cache
    if not pkgcache or (cache and not pkgcache.cache):
      pkgcache = PkgCache (pkgmgr.pkgname_prefix, cache)

  # install package
  def install (self, pkgname, dry_run, pkgmgr):
    args = []
    pkg_depot = ''

    # package either in subdirectory of package basename or
    # in kit directory
    pkg_depot = None
    if os.path.exists (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = self.pkgdir
    elif os.path.exists (os.path.join (self.pkgdir, self.pkgname_base)):
      pkg_depot = self.pkgdir + '/' + self.pkgname_base
    if not pkg_depot:
      raise pkgsys.PkgsysPackageNotFoundError ([pkgname, self.pkgname_base])

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)

    apply (self.output, ['msg', '$ setld ' + arg_string (args) +
                         ' -l ' + pkg_depot + ' ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'setld', ['setld'] + args +
                       ['-l', pkg_depot, pkgname]) == 0

  # remove package
  def remove (self, pkgname, dry_run, pkgmgr):
    args = []

    if pkgmgr.args:
      for arg in pkgmgr.args:
        args.append (arg)

    apply (self.output, ['msg', '$ setld ' + arg_string (args) +
                         ' -d ' + pkgname + '\n'])

    if dry_run:
      return True

    return os.spawnvp (os.P_WAIT, 'setld', ['setld'] + args +
                       ['-d', pkgname]) == 0

  # extract raw package
  def extract (self, pkgname, dry_run, pkgmgr, destdir):
    pkg_depot = ''
    dest_depot = ''

    # package either in subdirectory of package basename or
    # in kit directory
    pkg_depot = None
    if os.path.exists (os.path.join (self.pkgdir, pkgname)):
      pkg_depot = self.pkgdir
      dest_depot = os.path.join (destdir, pkgname)
    elif os.path.exists (os.path.join (self.pkgdir, self.pkgname_base)):
      pkg_depot = os.path.join (self.pkgdir, self.pkgname_base)
      dest_depot = os.path.join (destdir, self.pkgname_base)
    if not pkg_depot:
      raise pkgsys.PkgsysPackageNotFoundError ([pkgname, self.pkgname_base])

    # soft error if duplicate file in destination directory and
    # --reinstall not given
    if not pkgmgr.reinstall and os.path.exists (dest_depot):
      raise pkgsys.PkgsysExtractDuplicate (dest_depot)

    if os.path.exists (dest_depot):
      apply (self.output, ['msg', '$ rm -rf ' + dest_depot + '\n'])
      if not dry_run:
        misclib.rmdir (dest_depot)

    apply (self.output, ['msg', '$ cp -pr ' + pkg_depot + ' ' +
                         dest_depot + '\n'])

    # copy package to destination directory
    if not dry_run:
      shutil.copytree (pkg_depot, dest_depot)

    return dest_depot

  # determine which packages are installed
  def installed (self):
    pkg_status = PkgStatus ()
    subpackages = 0			# if subpackages found

    # determine which subpackages are installed. only check installed
    # subpkgs for those which we are going to install
    for subpkg in self.subpkgs:
      if not self.pkgdb_entry_ver.subpkg.has_key (subpkg):
        continue

      subpackages = 1

      # determine if version in package database is newer than
      # installed version
      for pkgname in self.get_pkgnames (subpkg):
        entry = pkgcache.version (pkgname)
        if entry:
          # get package version and revision. earlier releases didn't
          # have the package revision as part of the package
          # description so default to rev = 1.
          pkg_ver = Version (re.sub ('^v([^,]+).*', '\\1', entry))
          if re.search ('^v[^,]+, r', entry):
            pkg_rev = Version (re.sub ('^v[^,]+, r(\d+)', '\\1', entry))
          else:
            pkg_rev = Version ('1')

          if self.pkg_ver < pkg_ver:
            pkg_status.downgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_ver > pkg_ver:
            pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
            continue

          if self.pkg_rev is not None:
            if pkg_ver == self.pkg_ver and pkg_rev < self.pkg_rev:
              pkg_status.upgrade[pkgname] = [pkg_ver, pkg_rev]
              continue

          pkg_status.installed.append (pkgname)
        else:
          pkg_status.not_installed.append (pkgname)

    if not subpackages:
      raise pkgsys.PkgsysNoSubpkgsError
    return (pkg_status)

  # get version and release of an installed package
  def get_installed_pkg_ver (self, pkgname):
    entry = pkgcache.version (pkgname)
    if entry:
      # get package version and revision. early releases didn't
      # have the package revision as part of the package
      # description so default to rev = 1.
      pkg_ver = Version (re.sub ('^v([^,]+).*', '\\1', entry))
      if re.search ('^v[^,]+, r', entry):
        pkg_rev = Version (re.sub ('^v[^,]+, r(\d+)', '\\1', entry))
      else:
        pkg_rev = Version ('1')
    else:
      raise pkgsys.PkgsysPackageNotFoundError

    return [pkg_ver, pkg_rev]

  # determine package manager name based on package name base
  # and subpackage type
  def get_pkgnames (self, subpkg_type):
    ident = get_os_ident ()

    return (Pkgsys.get_pkgnames (self, subpkg_type,
                                 { 'conf': ['CONF' + ident],
                                   'doc': ['DOC' + ident],
                                   'libruntime': ['LIBR' + ident],
                                   'man': ['MAN' + ident],
                                   'packages': ['PKG' + ident],
                                   'runtime': ['BIN' + ident] }))

  # determine package manager subpackage extension based on
  # subpackage type
  def get_subpkg_names (self, subpkg_type, pkgname = None):
    return (Pkgsys.get_subpkg_names (self, subpkg_type,
                                     { 'conf': [self.pkgname_base + 'CONF'],
                                       'doc': [self.pkgname_base + 'DOC'],
                                       'libruntime': [self.pkgname_base + 'LIBR'],
                                       'man': [self.pkgname_base + 'MAN'],
                                       'packages': [self.pkgname_base + 'PKG'],
                                       'runtime': [self.pkgname_base + 'BIN'] }))

  # determine subpackage type depending on package name
  def get_subpkg_type (self, pkgname):
    ident = get_os_ident ()

    return (Pkgsys.get_subpkg_type (self, pkgname,
                                    [ [self.pkgname_base + 'CONF' + ident,
                                       'conf'],
                                      [self.pkgname_base + 'DOC' + ident,
                                       'doc'],
                                      [self.pkgname_base + 'LIBR' + ident,
                                       'libruntime'],
                                      [self.pkgname_base + 'MAN' + ident,
                                       'man'],
                                      [self.pkgname_base + 'PKG' + ident,
                                       'packages'],
                                      [self.pkgname_base + 'BIN' + ident,
                                       'runtime'] ]))

  # given package name, determine basename
  def get_basename (self, pkgname):
    return (re.sub ('^(.*)(?:BIN|CONF|DOC|LIBR|MAN|PKG).*', '\\1',
                    pkgname))

class PkgCache:
  def __init__ (self, pkgname_prefix, cache = 0):
    self.cache = cache
    self.pkgname_prefix = pkgname_prefix
    self.entries = {}

    if cache:
      for entry in os.popen ('setld -i 2>&1', 'r').readlines ():
        if not re.match (self.pkgname_prefix + '[^\s]+\s+installed\s+',
                         entry):
          continue

        pkgname = re.sub ('^([^\s]+).*', '\\1', entry.rstrip ())
        self.entries[pkgname] = {}
        self.entries[pkgname]['version'] = re.sub ('^.*, (v[^,]+(?:, r\d+))',
                                                   '\\1', entry.rstrip ())

  def exists (self, pkgname):
    if self.cache:
      if self.entries.has_key (pkgname):
        return 1
    else:
      for entry in os.popen ('setld -v ' + pkgname + ' 2>&1',
                             'r').readlines ():
        if re.search ('by super-user only', entry):
          raise pkgsys.PkgsysNeedRootError
        if re.search ('\(' + pkgname + '\)', entry):
          return 1

    return 0

  def version (self, pkgname):
    if self.cache:
      if self.entries.has_key (pkgname):
        return self.entries[pkgname]['version']
    else:
      for entry in os.popen ('setld -v ' + pkgname + ' 2>&1',
                             'r').readlines ():
        if re.search ('by super-user only', entry):
          raise pkgsys.PkgsysNeedRootError
        if re.search ('\(' + pkgname + '\)', entry):
          return re.sub ('^.*, (v[^ ,]+(?:, r\d+)) .*', '\\1',
                         entry.rstrip ())

    return None
