# Copyright (C) 2015 The Written Word, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id: pkgstatus.py 218 2006-09-17 17:13:44Z china $

import string

import depend
import misclib
import pkgsys

from version import Version

def print_changelog (io, package, indent = 0, restrict_rev = False):
  changelog = package.pkgdb_entry_ver.changelog[:]
  if changelog:
    while changelog:
      change = changelog.pop (0)

      if restrict_rev:
        if package.pkg_rev != Version (change.revision):
          continue

      io.stdout (indent + 2, '', 'Changelog: rev ' + \
                 change.revision + ' -- ' + change.date + '\n')
      io.stdout (indent + 2, '', '-' * len ('Changelog: rev ' + \
                 change.revision + ' -- ' + change.date) + '\n')

      print_change_entry (io, change.items[:], indent)

      if changelog and not restrict_rev:
        io.stdout (0, '', '\n')

# print package dependencies
def print_dependencies (io, package, pkgmgr, depots, pkgdbs,
                        output_callback, indent = 0):
  io.stdout (indent + 2, '', 'dependencies ... ')

  # list of dependencies for package
  try:
    pkg_depends = depend.get_dependencies (package, pkgmgr,
                                           output_callback,
                                           find_installed = 0)
  except pkgsys.PkgsysInvalidSubpkgTypeError, msg:
    io.stdout (indent + 2, '', 'failed\n')
    io.stdout (indent + 4, 'error', 'invalid subpackage type detected in \
the dependency list for this package. The basename of the invalid \
package is "' + msg.pkgname_base + '" and the invalid subpackage \
type is "' + msg.subpkg_type + '". Apparently the entry in the \
pkg-db.xml file for this package is wrong.\n');
    sys.exit (1)

  if not pkg_depends:
    io.stdout (indent + 2, '', 'none\n')
  else:
    io.stdout (indent + 2, '', '\n')

    for pkg_depend in pkg_depends:
      depend_pkg = pkg_depend.pkg

      subpkg  = pkg_depend.subpkg_name

      try:
        if depots:
          depend_pkg.name = misclib.get_name_from_depots (depots,
                                                          depend_pkg)
        else:
          depend_pkg.name = misclib.get_name_from_pkgdb (pkgdbs,
                                                         depend_pkg)
      except misclib.MisclibPackageNotFoundError:
        depend_pkg.name = None

      io.stdout (indent + 4, '',
                 string.join (depend_pkg.get_pkgnames (subpkg), ', '))
      if depend_pkg.name:
        io.stdout (indent + 4, '', ' [' +
                   misclib.get_name_from_depots (depots, depend_pkg) + ']')
      io.stdout (indent + 4, '', ' (' + pkg_depend.test_ver)
      if pkg_depend.test_rev:
        io.stdout (indent + 4, '', ', ' + pkg_depend.test_rev)
      io.stdout (indent + 4, '', ')\n')

# print package description
def print_description (io, package, indent = 2):
  description = package.pkgdb_entry_ver.description
  if description:
    io.stdout (indent, '', 'description ...\n')
    text = package.pkgdb_entry_ver.description['text']
    paragraphs = text.split ('\n\n')
    while paragraphs:
      paragraph = paragraphs.pop (0)
      io.stdout (indent + 2, '', string.join (paragraph.split ()) + '\n')

      # separate successive paragraphs with newline
      if paragraphs:
        io.stdout (0, '', '\n')

# print package installation path component
def print_install_path (io, package, indent = 2):
  io.stdout (indent, '', 'installation path component for v' +
             package.prog_ver + ' ... ' +
             package.pkgdb_entry_ver.install_name + '\n')

# print package licenses
def print_licenses (io, package, indent = 0):
  licenses = package.pkgdb_entry_ver.licenses[:]
  while licenses:
    license = licenses.pop (0)
    items = license.items[:]

    if license.module:
      io.stdout (indent, '', 'program license (' + license.module + \
                 ') ...\n')
    else:
      io.stdout (indent, '', 'program license ...\n')

    while items:
      item = items.pop (0)
      skip = 1

      if item.type == 'indent':
        print_indent_items (io, item.items[:], indent + 4)
      elif item.type == 'orderedlist':
        item_num = int (item.start_num)
        list_type = item.list_type
        for list_item in item.items:
          io.stdout (indent, '', '\n')
          print_orderedlist_item_entry (io, list_item.items[:],
                                        indent + 2,
                                        list_type = list_type,
                                        item_no = item_num)
          item_num += 1
      elif item.type == 'para':
        skip = item.skip
        io.stdout (indent + 2, '', build_item_text (item) + '\n',
                   align = item.align)
      elif item.type == 'screen':
        for text in item.text.split ('\n'):
          io.stdout (indent + 4, '', text + '\n')
      elif item.type == 'unorderedlist':
        for list_item in item.items:
          print_unorderedlist_item_entry (io, list_item.items[:],
                                          indent + 2)

      # separate current text from next if paragraph next, however, if
      # paragraph begins with lowercase, then don't insert newline.
      if skip and items and items[0].type == 'para':
        para_text = build_item_text (items[0])
        if not para_text[0].islower ():
          io.stdout (0, '', '\n')

    if licenses:
        io.stdout (0, '', '\n')

# print package notes
def print_notes (io, package, indent = 0):
  notes = package.pkgdb_entry_ver.notes[:]
  while notes:
    notes_entry = notes.pop (0)

    if notes_entry.name:
      io.stdout (indent, '', 'program notes (' + notes_entry.name + \
                 ') ...\n')
    else:
      io.stdout (indent, '', 'program notes ...\n')

    notes_entry_items = notes_entry[:]
    while notes_entry_items:
      notes_entry_item = notes_entry_items.pop (0)

      if notes_entry_item.type == 'change':
        io.stdout (indent + 2, '', 'Changes: v' + \
                   notes_entry_item.from_version + \
                   ' -> v' + notes_entry_item.to_version + '\n')
        io.stdout (indent + 2, '', '-' * \
                   len ('Changes: v' + notes_entry_item.from_version + \
                        ' -> v' + notes_entry_item.to_version) + '\n')

        print_change_entry (io, notes_entry_item.items[:], indent)
      else:
        print_note_entry (io, notes_entry_item, indent)

      if notes_entry_items:
        io.stdout (0, '', '\n')

    if notes:
        io.stdout (0, '', '\n')

# build text for <item> or <para>
def build_item_text (items):
  text = ''

  for item_entry in items:
    if item_entry.type == 'command':
      text += item_entry.text
    elif item_entry.type == 'link':
      text += item_entry.value
    elif item_entry.type == 'text':
      text += item_entry.text

  return text.replace ('\n', ' ')

# print <indent>ed elements
def print_indent_items (io, items, indent):
  while items:
    skip = 1
    item = items.pop (0)

    if item.type == 'indent':
      print_indent_items (io, item.items[:], indent + 2)
    elif item.type == 'orderedlist':
      item_num = int (item.start_num)
      list_type = item.list_type
      for list_item in item.items:
        io.stdout (indent, '', '\n')
        print_orderedlist_item_entry (io, list_item.items[:],
                                      indent + 2,
                                      list_type = list_type,
                                      item_no = item_num)
        item_num += 1
    elif item.type == 'para':
      skip = item.skip
      io.stdout (indent + 2, '', build_item_text (item) + '\n',
                 align = item.align)
    elif item.type == 'screen':
      for text in item.text.split ('\n'):
        io.stdout (indent + 4, '', text + '\n')
    elif item.type == 'unorderedlist':
      for list_item in item.items:
        print_unorderedlist_item_entry (io, list_item.items[:],
                                        indent + 2)

    # separate current text from next if paragraph next, however, if
    # paragraph begins with lowercase, then don't insert newline.
    if skip and items and items[0].type == 'para':
      para_text = build_item_text (items[0])
      if not para_text[0].islower ():
        io.stdout (0, '', '\n')

def print_change_entry (io, items, indent):
  item_no = 1

  prev_item = None
  while items:
    item = items.pop (0)
    skip = 1

    if item.type == 'item':
      print_item_no = True

      change_items = item.items[:]
      while change_items:
        change_item = change_items.pop (0)
        change_skip = 1

        if change_item.type == 'orderedlist':
           item_num = int (change_item.start_num)
           list_type = change_item.list_type
           for list_item in change_item.items:
             io.stdout (indent, '', '\n')
             print_orderedlist_item_entry (io, list_item.items[:],
                                           indent + 2,
                                           list_type = list_type,
                                           item_no = item_num)
             item_num += 1
        elif change_item.type == 'indent':
          print_indent_items (io, change_item.items[:], indent + 7)
        elif change_item.type == 'para':
          change_skip = change_item.skip
          if print_item_no:
            io.stdout (indent + 2, str (item_no).rjust (3) + '. ',
                       build_item_text (change_item) + '\n', 0,
                       hangindent = 1, align = change_item.align)
          else:
            io.stdout (indent + 7, '', build_item_text (change_item) + '\n',
                       align = change_item.align)
        elif change_item.type == 'screen':
          for text in change_item.text.split ('\n'):
            io.stdout (indent + 9, '', text + '\n')
        elif change_item.type == 'unorderedlist':
          for list_item in change_item.items:
            print_unorderedlist_item_entry (io, list_item.items[:],
                                            indent + 7)

        # separate current text from next if paragraph next, however, if
        # paragraph befins with lowercase, then don't insert newline.
        if change_skip and change_items and change_items[0].type == 'para':
          para_text = build_item_text (change_items[0])
          if not para_text[0].islower ():
            io.stdout (0, '', '\n')

        print_item_no = False

      item_no += 1
    elif item.type == 'item group':
      if prev_item:
        io.stdout (0, '', '\n')

      io.stdout (indent + 2, '', '(' + item.group + ')\n')
      print_change_entry (io, item.items, indent + 2)
      if items and items[0].type != 'item group':
        io.stdout (0, '', '\n')
    elif item.type == 'orderedlist':
      item_num = int (item.start_num)
      list_type = item.list_type
      for list_item in item.items:
        io.stdout (indent, '', '\n')
        print_orderedlist_item_entry (io, list_item.items[:],
                                      indent + 2,
                                      list_type = list_type,
                                      item_no = item_num)
        item_num += 1
    elif item.type == 'para':
      skip = item.skip
      io.stdout (indent + 2, '', build_item_text (item) + '\n',
                 align = item.align)
    elif item.type == 'screen':
      for text in item.text.split ('\n'):
        io.stdout (indent + 4, '', text + '\n')
    elif item.type == 'unorderedlist':
      for list_item in item.items:
        print_unorderedlist_item_entry (io, list_item.items[:],
                                        indent + 2)

    # separate current text from next if paragraph next, however, if
    # paragraph begins with lowercase, then don't insert newline.
    if skip and items and items[0].type == 'para':
      para_text = build_item_text (items[0])
      if not para_text[0].islower ():
        io.stdout (0, '', '\n')

    prev_item = item.type

# print <note> and <platform> entries
def print_note_entry (io, note, indent):
  if note.type == 'platform':
    io.stdout (indent + 2, '', 'Platform notes: ' + note.name + '\n')
    io.stdout (indent + 2, '', '-' * len ('Platform notes: ' +
                                          note.name) + '\n')
  else:
    io.stdout (indent + 2, '', note.type.capitalize () + ' notes:\n')
    io.stdout (indent + 2, '', '-' * len (note.type + ' notes:') + '\n')

  items = note.items[:]
  while items:
    item = items.pop (0)
    skip = 1

    if item.type == 'indent':
      print_indent_items (io, item.items[:], indent + 2)
    elif item.type == 'orderedlist':
      item_num = int (item.start_num)
      list_type = item.list_type
      for list_item in item.items:
        io.stdout (indent, '', '\n')
        print_orderedlist_item_entry (io, list_item.items[:],
                                      indent + 2,
                                      list_type = list_type,
                                      item_no = item_num)
        item_num += 1
    elif item.type == 'para':
      skip = item.skip
      io.stdout (indent + 2, '', build_item_text (item) + '\n',
                 align = item.align)
    elif item.type == 'screen':
      for text in item.text.split ('\n'):
        io.stdout (indent + 4, '', text + '\n')
    elif item.type == 'unorderedlist':
      for list_item in item.items:
        print_unorderedlist_item_entry (io, list_item.items[:],
                                        indent + 2)

    # separate current text from next if paragraph next. however, if
    # paragraph begins with lowercase, then don't insert newline.
    if skip and items and items[0].type == 'para':
      para_text = build_item_text (items[0])
      if not para_text[0].islower ():
        io.stdout (0, '', '\n')


# print <item> entry for <orderdlist> entries
def print_orderedlist_item_entry (io, items, indent, item_no = 1,
                                  list_type = 'decimal'):
  just_width = 3	# how much space to reserve for item_no_string

  if list_type == 'decimal':
    item_no_string = str (item_no)
  elif list_type == 'lower-alpha':
    item_no_string = misclib.convert_ordnum (item_no, string.ascii_lowercase)
  elif list_type == 'upper-alpha':
    item_no_string = misclib.convert_ordnum (item_no, string.ascii_uppercase)
  elif list_type == 'lower-roman':
    item_no_string = misclib.convert_roman (item_no)
    just_width = 5
  elif list_type == 'upper-roman':
    item_no_string = string.upper (misclib.convert_roman (item_no))
    just_width = 5

  print_item_no = True
  while items:
    item = items.pop (0)
    skip = 1

    if item.type == 'orderedlist':
      item_num = int (item.start_num)
      list_type = item.list_type
      for list_item in item.items:
        io.stdout (indent, '', '\n')
        print_orderedlist_item_entry (io, list_item.items[:],
                                      indent + just_width + 2,
                                      list_type = list_type,
                                      item_no = item_num)
        item_num += 1
    elif item.type == 'indent':
      print_indent_items (io, item.items[:], indent + just_width + 4)
    elif item.type == 'para':
      skip = item.skip
      if print_item_no:
        io.stdout (indent + 2, item_no_string.rjust (just_width) + '. ',
                   build_item_text (item) + '\n', 0, hangindent = 1,
                   align = item.align)
      else:
        io.stdout (indent + 2 + just_width + 2, '',
                   build_item_text (item) + '\n', 0, hangindent = 1,
                   align = item.align)
      print_item_no = False
    elif item.type == 'screen':
      for text in item.text.split ('\n'):
        io.stdout (indent + just_width + 6, '', text + '\n')
    elif item.type == 'unorderedlist':
      unordered_level = level + 1
      for list_item in item.items:
        print_unorderedlist_item_entry (io, list_item.items[:],
                                        indent + unordered_level,
                                        unordered_level)

    # separate current text from next if paragraph next, however, if
    # paragraph begins with lowercase, then don't insert newline.
    if skip and items and items[0].type == 'para':
      para_text = build_item_text (items[0])
      if not para_text[0].islower ():
        io.stdout (0, '', '\n')


# print <item> entry for <unorderdlist> entries
def print_unorderedlist_item_entry (io, items, indent, level = 1):
  level_string = '*' * level

  print_level_string = True
  while items:
    item = items.pop (0)
    skip = 1

    if item.type == 'orderedlist':
      item_num = int (item.start_num)
      list_type = item.list_type
      for list_item in item.items:
        io.stdout (indent, '', '\n')
        print_orderedlist_item_entry (io, list_item.items[:],
                                      indent + 2,
                                      list_type = list_type,
                                      item_no = item_num)
        item_num += 1
    elif item.type == 'para':
      skip = item.skip
      if print_level_string:
        io.stdout (indent, level_string + ' ',
                   build_item_text (item) + '\n', 0, hangindent = 1,
                   align = item.align)
      else:
        io.stdout (indent, ' ' * len (level_string + ' '),
                   build_item_text (item) + '\n', 0, hangindent = 1,
                   align = item.align)
      print_level_string = False
    elif item.type == 'screen':
      for text in item.text.split ('\n'):
        io.stdout (indent + 4, '', text + '\n')
    elif item.type == 'unorderedlist':
      unordered_level = level + 1
      for list_item in item.items:
        print_unorderedlist_item_entry (io, list_item.items[:],
                                        indent + unordered_level,
                                        unordered_level)

    # separate current text from next if paragraph next, however, if
    # paragraph begins with lowercase, then don't insert newline.
    if skip and items and items[0].type == 'para':
      para_text = build_item_text (items[0])
      if not para_text[0].islower ():
        io.stdout (0, '', '\n')
