# Copyright (C) 2008-2009 The Written Word, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# $Id: xmlutil.py 497 2009-04-13 18:47:38Z china $

import os
import re
import sys

from lxml import etree
from types import StringType, UnicodeType

class XMLError (Exception):
  """Base class for all exceptions raised by this module"""

  def __init__ (self, msg = None):
    self.msg = msg

  def __str__ (self):
    return self.msg

class XMLDTDError (XMLError):
  """DTD validation error"""

class XMLParseError (XMLError):
  """XML parsing error"""


# When xml_path has a DOCTYPE declaration, extract a DTD path from it.
# Return an absolute path as is, or else search dtd_dirs for the first
# matching relative path and return that.
def dtd_search (dtd_dirs, xml_path, default):
  dtd_regex = re.compile ('<!DOCTYPE .* SYSTEM "(file://)?([^"]+)"')
  dtd_maybe = None

  if type (xml_path) in (StringType, UnicodeType):    
    # Normalize, and strip leading 'file://'
    match = re.match ('file://(.*)$', xml_path)
    if match:
      xml_path = match.group (1)
    xml_path = os.path.normpath (xml_path)
    
    fh = open (xml_path, 'r')
  else:
    # xml_path must be an open file handle
    fh = xml_path

    # save the current read position, but scan from the beginning
    filepos = fh.tell ()
    fh.seek (0)

  # Optimise for well-formed db, where all files have a
  # DOCTYPE declaration in the first few lines
  while 1:
    aline = fh.readline ()
    if not aline:
      # no DOCTYPE thru EOF
      dtd_maybe = default
      break
    match = dtd_regex.search (aline)
    if match:
      # exit loop with first match
      dtd_maybe = match.group (2)
      break

  if type (xml_path) in (StringType, UnicodeType):
    # close the handle opened by this function
    fh.close ()
  else:
    # reset the read position back to where it was originally
    fh.seek (filepos)

  # when dtd is a relative path, search for first match in dtd_dirs
  if not dtd_maybe.startswith ('/'):
    for dir in dtd_dirs:
      path = os.path.join (dir, dtd_maybe)
      if os.path.exists (path):
        dtd_maybe = path
        break

  return dtd_maybe

class DTD:
  """All instances of this class share a cache of parsed DTD documents.
  Whenever a DTD validation is performed, the class instance uses the
  previously cached DTD object in preference to parsing that DTD
  again."""

  cache = {}

  def __init__ (self, path):
    self.path = path
    if not path in DTD.cache:
      DTD.cache[path] = etree.DTD (path)

  def error_log (self):
    return DTD.cache[self.path].error_log

  def validate (self, xml):
    return DTD.cache[self.path].validate (xml)

